grammar silver:analysis:agnosis:core;
import silver:base;
import silver:utils:command;
import core;


aspect production run
top::RunUnit ::= iIn::IO args::String p::Production(Root ::= String)
{
  ops <- [printAgnosis(a'', grammars)]; 
}


abstract production printAgnosis
top::Unit ::= a::Command specs::[GrammarSpec]
{
  local attribute r :: Decorated RootSpec;
  r = head(specs).rootSpec;

  local attribute has_start :: Boolean;
  has_start = r.hasStartSymbol;

  local attribute shn :: String;
  shn = makeName(r.startName);

  local attribute ags :: String;
  ags = buildAgnosis(r.agnosis, shn);

  local attribute i :: IO;
  i = if !has_start
      then top.ioIn 
      else writeFile(a.generatedPath ++ "Agnosis_" ++ makeName(head(specs).name), ags ++ "\n\n", top.ioIn);

  local attribute recurse :: Unit;
  recurse = printAgnosis(a'', tail(specs));
  recurse.ioIn = i;

  top.io = if null(specs) then top.ioIn else recurse.io;

  top.code = 0;

  top.order = 8;
}

function buildAgnosis
String ::= ags::[AgData] s::String
{
  
  return  "AgnInvoke\n" ++ 
          " [AgnCmdValidateGAG]\n" ++
	  " (AgnInvokeGrammar\n" ++ 
	  "  (AG\n" ++ 
 	  makeCFG(getType("NT", ags), getType("T", ags), getType("Prod", ags), s) ++ "\n\n" ++ 
	  "   [ " ++ makeAttrDcls(getType("AttrDecl", ags)) ++ " ]\n\n" ++ 
	  "   [ " ++ makeLocalAttrDcls(getType("LocalAttrDecl", ags)) ++ " ]\n\n" ++ 
	  "   [ " ++ makeAttributionRules(getType("AttributionRule", ags)) ++ " ]\n  )\n )\n";
}

function makeCFG
String ::= nts::[AgData] ts::[AgData] ps::[AgData] s::String
{
  return "   (CFG\n" ++
	 "    [ " ++ makeNTList(nts) ++ " ]\n" ++ 
	 "    [ " ++ makeTList(ts) ++ " ]\n" ++ 
         "    [ " ++ makeProdList(ps) ++ " ]\n" ++
	 "    (NT \"" ++ s ++ "\")\n   )\n";
}


function makeNTList
String ::= ags::[AgData]
{
  return if null(ags) 
	 then ""
	 else "(NT \"" ++ head(ags).ntName ++ "\")" ++
	      (if null(tail(ags)) then "" else ",\n      " ++ makeNTList(tail(ags)));
}

function makeTList
String ::= ags::[AgData]
{
  return if null(ags) 
	 then ""
	 else "(T \"" ++ head(ags).tName ++ "\")" ++
	      (if null(tail(ags)) then "" else ",\n      " ++ makeTList(tail(ags)));
}

function makeProdList
String ::= ags::[AgData]
{
  return if null(ags) 
	 then ""
	 else "(Prod (ProdName \"" ++ head(ags).pName ++ "\") (NT \"" ++ head(ags).pOutName ++ "\") [" ++ makeProdSig(head(ags).pInNames) ++ "])" ++
	      (if null(tail(ags)) then "" else ",\n      " ++ makeProdList(tail(ags)));
}

function makeProdSig
String ::= ags::[AgData]
{
  return if null(ags) 
	 then ""
	 else (if head(ags).type == "NT" then "(VNT (NT \"" ++ head(ags).ntName ++ "\"))" else "(VT (T \"" ++ head(ags).tName ++ "\"))") ++
	      (if null(tail(ags)) then "" else ", " ++ makeProdSig(tail(ags)));
}

function makeAttrDcls
String ::= ags::[AgData]
{
  return if null(ags) 
	 then ""
	 else "(AttributeDeclaration " ++ 
	      "(NT \"" ++ head(ags).ntName ++ "\") " ++
	      "(AttrDecl \"" ++ head(ags).attrName ++ "\" (" ++ head(ags).attrType ++ ") (" ++ head(ags).attrDirection ++ ")))" ++
	      (if null(tail(ags)) then "" else ",\n     " ++ makeAttrDcls(tail(ags)));
}

function makeLocalAttrDcls
String ::= ags::[AgData]
{
  return if null(ags) 
	 then ""
	 else "(LocalAttributeDeclaration " ++ 
	      "(ProdName \"" ++ head(ags).pName ++ "\") " ++
	      "(AttrDecl \"" ++ head(ags).attrName ++ "\" (" ++ head(ags).attrType ++ ") (" ++ head(ags).attrDirection ++ ")))" ++
	      (if null(tail(ags)) then "" else ",\n     " ++ makeLocalAttrDcls(tail(ags)));
}

function makeAttributionRules
String ::= ags::[AgData]
{
  return if null(ags)
	 then ""
	 else "(AttributionRule (ProdName \"" ++ head(ags).pName ++ "\") " ++ makeAttrDef(head(ags).attribution) ++ ")" ++
	      (if null(tail(ags)) then "" else ",\n     " ++ makeAttributionRules(tail(ags)));
}

function makeAttrAccess
String ::= ag::AgData
{
  return if      ag.type == "AttrDepElt"
	 then "(AttrDepElt " ++ toString(ag.preDotIndex) ++ " (AttrName \"" ++ ag.postDotName ++ "\"))"
	 else if ag.type == "AttrDepEltLocal"
	 then "(AttrDepEltLocal (AttrName \"" ++ ag.localName ++ "\"))"
	 else if ag.type == "AttrDepEltRef"
	 then "(AttrDepEltRef (AttrName \"" ++ ag.preDotName ++ "\") (AttrName \"" ++ ag.postDotName ++ "\"))"
	 else if ag.type == "AttrDepEltTree"
	 then "(AttrDepEltTree (AttrName \"" ++ ag.preDotName ++ "\") (AttrName \"" ++ ag.postDotName ++ "\"))"
	 else error("Unknown Access: " ++ ag.type);
}

function makeAttrDef
String ::= ag::AgData
{
  return if      ag.type == "AttrDep"
	 then "(AttrDep " ++ makeAttrAccess(ag.lhsRule) ++ " " ++ makeAttrAccess(ag.rhsRule) ++ ")"
	 else if ag.type == "AttrDepHigher"
	 then "(AttrDepHigher " ++ makeAttrAccess(ag.lhsRule) ++ " " ++ toString(ag.rhsIndex) ++ ")"
	 else if ag.type == "AttrDepRef"
	 then "(AttrDepRef " ++ makeAttrAccess(ag.lhsRule) ++ " " ++ toString(ag.rhsIndex) ++ ")"	 
	 else if ag.type == "AttrDepRefTree"
	 then "(AttrDepRefTree " ++ makeAttrAccess(ag.lhsRule) ++ " " ++ makeAttrAccess(ag.rhsRule) ++ ")"	 
	 else if ag.type == "AttrDepTreeRef"
	 then "(AttrDepTreeRef " ++ makeAttrAccess(ag.lhsRule) ++ " " ++ makeAttrAccess(ag.rhsRule) ++ ")"	 
	 else if ag.type == "AttrDepConst"
	 then "(AttrDepConst " ++ makeAttrAccess(ag.lhsRule) ++ ")"
	 else error("Unknown Def: " ++ ag.type);
}

function getType
[AgData] ::= s::String ags::[AgData]
{
  return if null(ags) 
	 then [] 
	 else if head(ags).type == s 
	      then cons(head(ags), getType(s, tail(ags)))
	      else getType(s, tail(ags));
}