grammar silver:analysis:agnosis:core;
import silver:base:core;

synthesized attribute isConstant :: Boolean;
attribute isConstant occurs on Expr;

synthesized attribute isChild :: Boolean;
attribute isChild occurs on Expr;

synthesized attribute dataName :: String;
attribute dataName occurs on Expr;

synthesized attribute possibleAgnosis :: [AgData];
attribute possibleAgnosis occurs on Expr;

attribute agnosis occurs on Expr, Exprs, ExprInh, ExprLHSExpr;

inherited attribute idDown :: Integer;
synthesized attribute idUp :: Integer;

synthesized attribute lastAccess :: [AgData];
attribute lastAccess occurs on Expr;

synthesized attribute baseType :: Decorated TypeRep;
attribute baseType occurs on Expr;


--Base Expressions
aspect production nestedExpr
top::Expr ::= l::LParen_t e::Expr r::RParen_t
{
  top.isConstant = e.isConstant;
  top.isChild = e.isChild;
  top.dataName = e.dataName;
  
  top.baseType = top.typerep;
 
  top.possibleAgnosis = e.possibleAgnosis;
  top.agnosis = e.agnosis;
  top.lastAccess = e.lastAccess;
}

aspect production dontDecorateExpr
top::Expr ::= q::QName
{
  top.isConstant = false;
  top.isChild = in_sig;

  top.dataName = makeName(fName);

  local attribute i :: Integer;
  i = getIndex(fName, top.fullSignatureNames);

  top.baseType = top.typerep;

  top.possibleAgnosis = [];
  top.agnosis = [];
  top.lastAccess = if in_sig 
		   then [attrAccessDclData6(i, top.typerep)] 
		   else [attrAccessDclData2(fName, top.typerep)];
}

aspect production decorateExpr
top::Expr ::= q::QName
{
  top.isConstant = false;
  top.isChild = in_sig;

  top.dataName = makeName(fName);

  local attribute i :: Integer;
  i = getIndex(fName, top.fullSignatureNames);

  top.baseType = top.typerep;

  top.possibleAgnosis = [];
  top.agnosis = [];
  top.lastAccess = if in_sig 
		   then [attrAccessDclData6(i, top.typerep)] 
		   else [attrAccessDclData2(fName, top.typerep)];
}

aspect production productionReference
top::Expr ::= q::QName
{
  top.isConstant = false;
  top.isChild = false;

  top.dataName = "";

  top.baseType = top.typerep;

  top.agnosis = [attrAccessDclData5()];
  top.possibleAgnosis = [];
  top.lastAccess = [];
}


--no one can take the dot off of a production call there would need to be a decorate ...with {...} inserted.
aspect production genericApplicationDispatcher
top::Expr ::= e::Expr es::Exprs
{
  local attribute lDcl :: AgData;
  lDcl = localAttrDclData(top.thisName, makeName(top.dataName),
			  "AttrHigher (NT \"" ++ makeName(top.typerep.typeName) ++ "\")", "Synthesized");
  
  local attribute depDcl :: AgData;
  depDcl = attributionData(top.thisName, attrDefDclData6(attrAccessDclData2(makeName(top.dataName), top.typerep)));

  top.baseType = top.typerep;

  top.possibleAgnosis = [lDcl'', depDcl''];
  top.lastAccess = [];

  top.isChild = false;
  top.isConstant = false;

  top.dataName = "NEW_LOCAL_ATTRIBUTE_" ++ toString(genint()); 

  top.agnosis = e.agnosis +++ e.lastAccess +++ es.agnosis;
}

aspect production atAccess
top::Expr ::= e::Expr d::At_t q::QName
{
  -- make a new local declaration
  local attribute pN :: String;
  pN = top.thisName;

  local attribute atN :: String;
  atN = makeName(top.dataName);

  local attribute atT :: String;
  atT = if top.typerep.isNonTerminal 
	then "AttrHigher (NT \"" ++ makeName(top.typerep.typeName) ++ "\")"
	else if top.typerep.isDecorated
	then "AttrRef (NT \"" ++ makeName(top.typerep.typeName) ++ "\")"
	else "AttrScaler";

  local attribute atD :: String;
  atD = "Synthesized";

  local attribute lDcl :: AgData;
  lDcl = localAttrDclData(pN, atN, atT, atD);
  
  -- make it depend on e
  local attribute depDcl :: AgData;
  depDcl = attributionData(pN, attrDefDclData1(attrAccessDclData2(top.dataName, top.typerep), new(ag)));

  top.possibleAgnosis = [new(lDcl), new(depDcl)];

  top.baseType = top.typerep;

  top.isChild = false;
  top.isConstant = false;
  top.dataName = "NEW_LOCAL_ATTRIBUTE_" ++ toString(genint()); 

  local attribute i :: Integer;
  i = getIndex(e.dataName, top.fullSignatureNames);

  --make the current dependancy  
  local attribute ag :: AgData;
  ag = 	if e.isChild 
        then attrAccessDclData1(i, makeName(fName), top.typerep)
        else if e.typerep.isDecorated
 	     then attrAccessDclData3(e.dataName, makeName(fName), top.typerep)
	     else attrAccessDclData4(e.dataName, makeName(fName), top.typerep);

  top.agnosis = e.possibleAgnosis +++ e.agnosis;

  top.lastAccess = [new(ag)];
}

aspect production trueConst
top::Expr ::= t::True_kwd
{
  top.isConstant = true;
  top.isChild = false;
  top.dataName = "";
  
  top.baseType = top.typerep;
  top.possibleAgnosis = [];
  top.agnosis = [];
  top.lastAccess = [];
}

aspect production falseConst
top::Expr ::= f::False_kwd
{
  top.isConstant = true;
  top.isChild = false;
  top.dataName = "";
  
  top.baseType = top.typerep;
  top.possibleAgnosis = [];
  top.agnosis = [];
  top.lastAccess = [];
}

aspect production and
top::Expr ::= e1::Expr a::And_t e2::Expr
{
  top.isConstant = e1.isConstant && e2.isConstant;
  top.isChild = false;
  top.dataName = "";
  
  top.baseType = top.typerep;
  top.possibleAgnosis = [];
  top.agnosis = e1.agnosis +++ e2.agnosis +++ e1.lastAccess +++ e2.lastAccess;
  top.lastAccess = [];
}

aspect production or
top::Expr ::= e1::Expr o::Or_t e2::Expr
{
  top.isConstant = e1.isConstant && e2.isConstant;
  top.isChild = false;
  top.dataName = "";
  
  top.baseType = top.typerep;
  top.possibleAgnosis = [];
  top.agnosis = e1.agnosis +++ e2.agnosis +++ e1.lastAccess +++ e2.lastAccess;
  top.lastAccess = [];
}

aspect production not
top::Expr ::= n::Not_t e::Expr
{
  top.isConstant = e.isConstant;
  top.isChild = false;
  top.dataName = "";
  
  top.baseType = top.typerep;
  top.possibleAgnosis = [];
  top.agnosis = e.agnosis +++ e.lastAccess;
  top.lastAccess = [];
}

aspect production gt
top::Expr ::= e1::Expr g::GT_t e2::Expr
{
  top.isConstant = e1.isConstant && e2.isConstant;
  top.isChild = false;
  top.dataName = "";
  
  top.baseType = top.typerep;
  top.possibleAgnosis = [];
  top.agnosis = e1.agnosis +++ e2.agnosis +++ e1.lastAccess +++ e2.lastAccess;
  top.lastAccess = [];
}

aspect production lt
top::Expr ::= e1::Expr l::LT_t e2::Expr
{
  top.isConstant = e1.isConstant && e2.isConstant;
  top.isChild = false;
  top.dataName = "";
  
  top.baseType = top.typerep;
  top.possibleAgnosis = [];
  top.agnosis = e1.agnosis +++ e2.agnosis +++ e1.lastAccess +++ e2.lastAccess;
  top.lastAccess = [];
}


aspect production gteq
top::Expr ::= e1::Expr g::GTEQ_t e2::Expr
{
  top.isConstant = e1.isConstant && e2.isConstant;
  top.isChild = false;
  top.dataName = "";
  
  top.baseType = top.typerep;
  top.possibleAgnosis = [];
  top.agnosis = e1.agnosis +++ e2.agnosis +++ e1.lastAccess +++ e2.lastAccess;
  top.lastAccess = [];
}


aspect production lteq
top::Expr ::= e1::Expr l::LTEQ_t e2::Expr
{
  top.isConstant = e1.isConstant && e2.isConstant;
  top.isChild = false;
  top.dataName = "";
  
  top.baseType = top.typerep;
  top.possibleAgnosis = [];
  top.agnosis = e1.agnosis +++ e2.agnosis +++ e1.lastAccess +++ e2.lastAccess;
  top.lastAccess = [];
}


aspect production eqeq
top::Expr ::= e1::Expr eq::EQEQ_t e2::Expr
{
  top.isConstant = e1.isConstant && e2.isConstant;
  top.isChild = false;
  top.dataName = "";
  
  top.baseType = top.typerep;
  top.possibleAgnosis = [];
  top.agnosis = e1.agnosis +++ e2.agnosis +++ e1.lastAccess +++ e2.lastAccess;
  top.lastAccess = [];
}


aspect production neq
top::Expr ::= e1::Expr n::NEQ_t e2::Expr
{
  top.isConstant = e1.isConstant && e2.isConstant;
  top.isChild = false;
  top.dataName = "";
  
  top.baseType = top.typerep;
  top.possibleAgnosis = [];
  top.agnosis = e1.agnosis +++ e2.agnosis +++ e1.lastAccess +++ e2.lastAccess;
  top.lastAccess = [];
}


aspect production ifThenElse
top::Expr ::= i::If_kwd e1::Expr t::Then_kwd e2::Expr e::Else_kwd e3::Expr
{
  top.isConstant = e2.isConstant && e3.isConstant;
  top.isChild = false;
  top.dataName = "";
  
  top.baseType = top.typerep;
  top.possibleAgnosis = e2.possibleAgnosis +++ e3.possibleAgnosis;
  top.agnosis = e1.agnosis +++ e2.agnosis +++ e3.agnosis;
  top.lastAccess = e1.lastAccess +++ e2.lastAccess +++ e3.lastAccess;
}

aspect production intConst
top::Expr ::= i::Int_t
{
  top.isConstant = true;
  top.isChild = false;
  top.dataName = "";
  
  top.baseType = top.typerep;
  top.possibleAgnosis = [];
  top.agnosis = [];
  top.lastAccess = [];
}

aspect production floatConst
top::Expr ::= f::Float_t
{
  top.isConstant = true;
  top.isChild = false;
  top.dataName = "";
  
  top.baseType = top.typerep;
  top.possibleAgnosis = [];
  top.agnosis = [];
  top.lastAccess = [];
} 

aspect production plus
top::Expr ::= e1::Expr p::Plus_t e2::Expr
{
  top.isConstant = e1.isConstant && e2.isConstant;
  top.isChild = false;
  top.dataName = "";
  
  top.baseType = top.typerep;
  top.possibleAgnosis = [];
  top.agnosis = e1.agnosis +++ e2.agnosis +++ e1.lastAccess +++ e2.lastAccess;
  top.lastAccess = [];
}

aspect production minus
top::Expr ::= e1::Expr m::Minus_t e2::Expr
{
  top.isConstant = e1.isConstant && e2.isConstant;
  top.isChild = false;
  top.dataName = "";
  
  top.baseType = top.typerep;
  top.possibleAgnosis = [];
  top.agnosis = e1.agnosis +++ e2.agnosis +++ e1.lastAccess +++ e2.lastAccess;
  top.lastAccess = [];
}
aspect production multiply
top::Expr ::= e1::Expr m::Multiply_t e2::Expr
{
  top.isConstant = e1.isConstant && e2.isConstant;
  top.isChild = false;
  top.dataName = "";
  
  top.lastAccess = [];
  top.baseType = top.typerep;
  top.possibleAgnosis = [];
  top.agnosis = e1.agnosis +++ e2.agnosis +++ e1.lastAccess +++ e2.lastAccess;
}
aspect production divide
top::Expr ::= e1::Expr d::Divide_t e2::Expr
{
  top.isConstant = e1.isConstant && e2.isConstant;
  top.isChild = false;
  top.dataName = "";
  
  top.baseType = top.typerep;
  top.possibleAgnosis = [];
  top.agnosis = e1.agnosis +++ e2.agnosis +++ e1.lastAccess +++ e2.lastAccess;
  top.lastAccess = [];
}

aspect production neg
top::Expr ::= m::Minus_t e::Expr
{
  top.isConstant = e.isConstant;
  top.isChild = false;
  top.dataName = "";
  
  top.baseType = top.typerep;
  top.possibleAgnosis = [];
  top.agnosis = e.agnosis +++ e.lastAccess;
  top.lastAccess = [];
}

aspect production stringConst
top::Expr ::= s::String_t
{
  top.isConstant = true;
  top.isChild = false;
  top.dataName = "";
  
  top.baseType = top.typerep;
  top.possibleAgnosis = [];
  top.agnosis = [];
  top.lastAccess = [];
}

aspect production plusPlus
top::Expr ::= e1::Expr p::PlusPlus_t e2::Expr
{
  top.isConstant = e1.isConstant && e2.isConstant;
  top.isChild = false;
  top.dataName = "";
  
  top.baseType = top.typerep;
  top.possibleAgnosis = [];
  top.agnosis = e1.agnosis +++ e2.agnosis +++ e1.lastAccess +++ e2.lastAccess;
  top.lastAccess = [];
}

aspect production exprsEmpty
top::Exprs ::=
{
  top.agnosis = [];
}

aspect production exprsSingle
top::Exprs ::= e::Expr
{
  top.agnosis = e.agnosis +++ e.lastAccess;
}

aspect production exprsCons
top::Exprs ::= e1::Expr c::Comma_t e2::Exprs
{
  top.agnosis = e1.agnosis +++ e2.agnosis +++ e1.lastAccess;
}

aspect production exprWith
top::Expr ::= ev::Decorate_kwd e::Expr w::With_kwd l::LCurly_t inh::ExprInh r::RCurly_t
{
  local attribute lDcl1 :: AgData;
  lDcl1 = localAttrDclData(top.thisName, makeName(tN),
			  "AttrHigher (NT \"" ++ makeName(top.typerep.typeName) ++ "\")", "Synthesized");

  local attribute lDcl2 :: AgData;
  lDcl2 = localAttrDclData(top.thisName, makeName(top.dataName),
			  "AttrRef (NT \"" ++ makeName(top.typerep.typeName) ++ "\")", "Synthesized");
  
  --lhs of Dcl (e or childNum)
  local attribute lhsDcl :: AgData;
  lhsDcl = if e.isChild
	   then head(e.lastAccess)
	   else attrAccessDclData2(e.dataName, e.typerep);

  -- make it depend on foo = e (or foo = childNum)
  local attribute depDcl1 :: AgData;
  depDcl1 = if e.isChild 
	   then attributionData(top.thisName, attrDefDclData2(attrAccessDclData2(tN, top.typerep), lhsDcl.index))
	   else attributionData(top.thisName, attrDefDclData1(attrAccessDclData2(tN, top.typerep), new(lhsDcl)));

  --assign a bunch of inherited attributes to foo


  --convert the tree to a reference
  local attribute depDcl2 :: AgData;
  depDcl2 = attributionData(top.thisName, attrDefDclData4(attrAccessDclData2(top.dataName, top.typerep),
							  attrAccessDclData2(tN, top.typerep)));

  top.possibleAgnosis = [];
  top.lastAccess = [attrAccessDclData2(top.dataName, top.typerep)];

  top.isChild = false;
  top.isConstant = false;
  top.baseType = top.typerep;

  local attribute tN :: String;
  tN = "NEW_LOCAL_ATTRIBUTE_" ++ toString(genint()); 

  top.dataName = "NEW_LOCAL_ATTRIBUTE_" ++ toString(genint()); 

  top.agnosis = e.agnosis +++ e.possibleAgnosis +++ [new(lDcl1), new(depDcl1), new(lDcl2), new(depDcl2)] +++ inh.agnosis;

  inh.iDataName = tN;
}

autocopy attribute iDataName :: String;
attribute iDataName occurs on ExprInh, ExprLHSExpr;

aspect production exprStmtsEmpty
top::ExprInh ::= 
{
  top.agnosis = [];
}

aspect production exprStmtsCons
top::ExprInh ::= lhs::ExprLHSExpr ekwd::Equal_t e::Expr s::Semi_t inh::ExprInh
{ 
  local attribute allAg ::[AgData];
  allAg = e.agnosis +++ e.lastAccess;
  
  local attribute ags :: [AgData];
  ags = if e.isConstant 
	then [attributionData(top.thisName, attrDefDclData6(head(lhs.agnosis)))]
	else makeAgData(top.thisName, head(lhs.agnosis), allAg);

  top.agnosis = ags +++ inh.agnosis;
}

aspect production exprLhsExpr
top::ExprLHSExpr ::= q::QName
{
  top.agnosis = if head(vals).typerep.isDecorated
		then [attrAccessDclData3(makeName(top.iDataName), makeName(fName), top.typerep)]
		else [attrAccessDclData4(makeName(top.iDataName), makeName(fName), top.typerep)];
}