grammar silver:analysis:typechecking:core;
import silver:base;

aspect production stringLength
top::Expr ::= e::Expr
{
  top.typeErrors = e.typeErrors;      
}

aspect production unknownLength
top::Expr ::= e::Expr
{
  local attribute er :: String;
  er = top.location ++ "Error: Opperand to 'length(..)' is not compatible.\n";

  top.typeErrors = er ++ e.typeErrors;      
}

aspect production indexOfFunction
top::Expr ::= i::IndexOf_kwd l::LParen_t e1::Expr c::Comma_t e2::Expr r::RParen_t
{
  top.typeErrors = er1 ++ er2 ++ e1.typeErrors ++ e2.typeErrors;

  local attribute er1 :: String;
  er1 = if (e1.typerep.typeName == "String") 
	then ""
	else top.location ++ "Error: First opperand to 'indexof(...)' must be of type String.\n";

  local attribute er2 :: String;
  er2 = if (e2.typerep.typeName == "String") 
	then ""
	else top.location ++ "Error: Second opperand to 'indexof(...)' must be of type String.\n";
}
aspect production subStringFunction
top::Expr ::= s::SubString_kwd l::LParen_t e1::Expr c1::Comma_t e2::Expr c2::Comma_t e3::Expr r::RParen_t
{
  top.typeErrors = er1 ++ er2 ++ er3 ++ e1.typeErrors ++ e2.typeErrors ++ e3.typeErrors;

  local attribute er1 :: String;
  er1 = if (e1.typerep.typeName == "Integer") 
	then ""
	else top.location ++ "Error: First opperand to 'substring(...)' must be of type Integer.\n";

  local attribute er2 :: String;
  er2 = if (e2.typerep.typeName == "Integer") 
	then ""
	else top.location ++ "Error: Second opperand to 'substring(...)' must be of type Integer.\n";

  local attribute er3 :: String;
  er3 = if (e3.typerep.typeName == "String") 
	then ""
	else top.location ++ "Error: Third opperand to 'substring(...)' must be of type String.\n";


}
aspect production errorFunction
top::Expr ::= er::Error_kwd l::LParen_t e::Expr r::RParen_t
{
  local attribute err :: String;
  err = if (e.typerep.typeName == "String") 
       then ""
       else top.location ++ "Error: Opperand to 'error(..)' must be of type String.\n";

  top.typeErrors = err ++ e.typeErrors;      
}
aspect production stringToIntFunction
top::Expr ::= t::ToInt_kwd l::LParen_t e::Expr r::RParen_t
{
  local attribute er :: String;
  er = if (e.typerep.typeName == "String") 
       then ""
       else top.location ++ "Error: Opperand to 'toInt(..)' must be of type String.\n";

  top.typeErrors = er ++ e.typeErrors;      
}
aspect production stringToFloatFunction
top::Expr ::= t::ToFloat_kwd l::LParen_t e::Expr r::RParen_t
{
  local attribute er :: String;
  er = if (e.typerep.typeName == "String") 
       then ""
       else top.location ++ "Error: Opperand to 'toFloat(..)' must be of type String.\n";

  top.typeErrors = er ++ e.typeErrors;      
}
aspect production numToStringFunction
top::Expr ::= t::ToString_kwd l::LParen_t e::Expr r::RParen_t
{
  local attribute er :: String;
  er = if (e.typerep.typeName == "Integer") ||
	  (e.typerep.typeName == "Float")  ||
          (e.typerep.typeName == "Boolean") 
       then ""
       else top.location ++ "Error: Opperand to 'toString(..)' must be of type Integer, Float or Boolean.\n";

  top.typeErrors = er ++ e.typeErrors;      
}

aspect production ftoiFunction
top::Expr ::= t::FTOI_kwd l::LParen_t e::Expr r::RParen_t
{
  local attribute er :: String;
  er = if e.typerep.typeName == "Float"
       then ""
       else top.location ++ "Error: Opperand to 'ftoi(..)' must be of type Float.\n";

  top.typeErrors = er ++ e.typeErrors;      

}

aspect production itofFunction
top::Expr ::= t::ITOF_kwd l::LParen_t e::Expr r::RParen_t
{
  local attribute er :: String;
  er = if e.typerep.typeName == "Integer"
       then ""
       else top.location ++ "Error: Opperand to 'itof(..)' must be of type Integer.\n";

  top.typeErrors = er ++ e.typeErrors;      
}

aspect production isDigitFunction
top::Expr ::= t::IsDigit_kwd l::LParen_t e::Expr r::RParen_t
{
  local attribute er :: String;
  er = if e.typerep.typeName == "String"
       then ""
       else top.location ++ "Error: Opperand to 'isDigit(..)' must be of type String.\n";

  top.typeErrors = er ++ e.typeErrors;      
}

aspect production isAlphaFunction
top::Expr ::= t::IsAlpha_kwd l::LParen_t e::Expr r::RParen_t
{
  local attribute er :: String;
  er = if e.typerep.typeName == "String"
       then ""
       else top.location ++ "Error: Opperand to 'isAlpha(..)' must be of type String.\n";

  top.typeErrors = er ++ e.typeErrors;      
}

aspect production isSpaceFunction
top::Expr ::= t::IsSpace_kwd l::LParen_t e::Expr r::RParen_t
{
  local attribute er :: String;
  er = if e.typerep.typeName == "String"
       then ""
       else top.location ++ "Error: Opperand to 'isSpace(..)' must be of type String.\n";

  top.typeErrors = er ++ e.typeErrors;      
}

aspect production isLowerFunction
top::Expr ::= t::IsLower_kwd l::LParen_t e::Expr r::RParen_t
{
  local attribute er :: String;
  er = if e.typerep.typeName == "String"
       then ""
       else top.location ++ "Error: Opperand to 'isLower(..)' must be of type String.\n";

  top.typeErrors = er ++ e.typeErrors;      
}

aspect production isUpperFunction
top::Expr ::= t::IsUpper_kwd l::LParen_t e::Expr r::RParen_t
{
  local attribute er :: String;
  er = if e.typerep.typeName == "String"
       then ""
       else top.location ++ "Error: Opperand to 'isUpper(..)' must be of type String.\n";

  top.typeErrors = er ++ e.typeErrors;      
}


aspect production newFunction
top::Expr ::= n::New_kwd l::LParen_t e::Expr r::RParen_t
{
  local attribute er :: String;
  er = if (e.typerep.isDecorated) 
       then ""
       else top.location ++ "Error: Opperand to 'new(..)' must be a Decorated Nonterminal.\n";

  top.typeErrors = er ++ e.typeErrors;      
}
aspect production terminalFunction
top::Expr ::= term::Terminal_kwd l::LParen_t t::Type c::Comma_t e::Expr r::RParen_t
{
  local attribute er :: String;
  er = if (e.typerep.typeName == "String") 
       then ""
       else top.location ++ "Error: Opperand to 'terminal(..)' must be of type String.\n";

  top.typeErrors = er ++ e.typeErrors;      
}
aspect production terminalFunction2
top::Expr ::= term::Terminal_kwd l::LParen_t t::Type c1::Comma_t e1::Expr c2::Comma_t e2::Expr c3::Comma_t e3::Expr r::RParen_t
{
  local attribute er1 :: String;
  er1 = if (t.typerep.isTerminal) 
        then ""
        else top.location ++ "Error: First Opperand to 'terminal(..)' must be a Terminal.\n";

  local attribute er2 :: String;
  er2 = if (e1.typerep.typeName == "String") 
        then ""
        else top.location ++ "Error: Second Opperand to 'terminal(..)' must be of type String.\n";

  local attribute er3 :: String;
  er3 = if (e2.typerep.typeName == "Integer") 
        then ""
        else top.location ++ "Error: Third Opperand to 'terminal(..)' must be of type Integer.\n";

  local attribute er4 :: String;
  er4 = if (e3.typerep.typeName == "Integer") 
        then ""
        else top.location ++ "Error: Fourth Opperand to 'terminal(..)' must be of type Integer.\n";

  top.typeErrors = er1 ++ er2 ++ er3 ++ er4 ++ e1.typeErrors ++ e2.typeErrors ++ e3.typeErrors;      
}
