grammar silver:analysis:typechecking:core;
import silver:base;
import core;

attribute typeErrors occurs on Expr;

--Base Expressions
aspect production nestedExpr
top::Expr ::= l::LParen_t e::Expr r::RParen_t
{
  top.typeErrors = e.typeErrors;
}

aspect production productionReference
top::Expr ::= q::QName
{
  top.typeErrors = "";
}

aspect production dontDecorateExpr
top::Expr ::= q::QName
{
  top.typeErrors = "";
}

aspect production decorateExpr
top::Expr ::= q::QName
{
  local attribute er2 :: String;
  er2 = if !null(vals) && !head(vals).typerep.isNonTerminal
	then top.location ++ "Error: Ticked value '" ++ q.name ++ "' must be an undecorated nonterminal.\n" 
	else "";

  local attribute er3 :: String;
  er3 = if !in_sig && !in_locals
	then top.location ++ "Error: Ticked name '" ++ q.name ++ "'must be in the signature or declared local.\n" 
	else "";

  top.typeErrors = er2 ++ er3;
}

aspect production genericApplicationDispatcher
top::Expr ::= e::Expr es::Exprs
{
  local attribute list_error :: String;
  list_error = if  e.typerep.isProduction && genListEquals(e.typerep.inputTypes, es.inputTypes) then ""
	       else top.location ++ "Error: Parameters given do not match the signature of the production.\n" ++
		 	"(expected)\n   " ++ printTypes(e.typerep.inputTypes) ++ "(actual)\n   " ++ printTypes(es.inputTypes) ++ "\n";
  
  local attribute notProdError :: String;
  notProdError = if e.typerep.isProduction then ""
   	         else top.location ++ "Error: " ++ e.pp ++ " does not have type Production.\n";

  top.typeErrors = notProdError ++ list_error ++ e.typeErrors ++ es.typeErrors;
}

function printTypes
String ::= ts::[Decorated TypeRep]
{
  return if null(ts) then ""
	 else head(ts).pp ++ 
	 (if null(tail(ts)) then "\n"
	  else ",\n   " ++ printTypes(tail(ts)));
}

aspect production atAccess
top::Expr ::= e::Expr d::At_t q::QName
{
  local attribute er1 :: String;
  er1 = if occursOn then ""
       else top.location ++ "Error: Attribute '" ++ q.name ++ "' does not decorate type '" ++ e.typerep.typeName ++ "'.\n";

  local attribute er2 :: String;
  er2 = if e.typerep.isDecorated || e.typerep.isTerminal then ""
	else top.location ++ "Error: @ value must be a decorated nonterminal or a terminal.\n" ;

  top.typeErrors = er2 ++ er1 ++ e.typeErrors;
}

aspect production hashAccess
top::Expr ::= e::Expr d::Hash_t q::QName
{}
aspect production trueConst
top::Expr ::= t::True_kwd
{
  top.typeErrors = "";
}

aspect production falseConst
top::Expr ::= f::False_kwd
{
  top.typeErrors = "";
}


aspect production and
top::Expr ::= e1::Expr a::And_t e2::Expr
{
  local attribute er :: String;
  er = if (e1.typerep.typeName == "Boolean") &&
	  (e2.typerep.typeName == "Boolean")
       then ""
       else top.location ++ "Error: Operands to && must be of type Boolean.\n";

  top.typeErrors = er ++ e1.typeErrors ++ e2.typeErrors;
}

aspect production or
top::Expr ::= e1::Expr o::Or_t e2::Expr
{
  local attribute er :: String;
  er = if (e1.typerep.typeName == "Boolean") &&
	  (e2.typerep.typeName == "Boolean")
       then ""
       else top.location ++ "Error: Operands to || must be of type Boolean.\n";

  top.typeErrors = er ++ e1.typeErrors ++ e2.typeErrors;
}

aspect production not
top::Expr ::= n::Not_t e::Expr
{
  local attribute er :: String;
  er = if (e.typerep.typeName == "Boolean")
       then ""
       else top.location ++ "Error: Opperand to ! must be of type Boolean.\n";

  top.typeErrors = er ++ e.typeErrors;
}

aspect production gt
top::Expr ::= e1::Expr g::GT_t e2::Expr
{
  local attribute er :: String;
  er = if ((e1.typerep.typeName == "Integer") &&
	   (e2.typerep.typeName == "Integer"))
	   ||
	  ((e1.typerep.typeName == "String") &&	
	   (e2.typerep.typeName == "String"))	
       then ""
       else top.location ++ "Error: Operands to > must be of type Integer or String.\n";

  top.typeErrors = er ++ e1.typeErrors ++ e2.typeErrors;
}

aspect production lt
top::Expr ::= e1::Expr l::LT_t e2::Expr
{
  local attribute er :: String;
  er = if ((e1.typerep.typeName == "Integer") &&
	   (e2.typerep.typeName == "Integer"))
	   ||
	  ((e1.typerep.typeName == "String") &&	
	   (e2.typerep.typeName == "String"))	
       then ""
       else top.location ++ "Error: Operands to < must be of type Integer or String.\n";

  top.typeErrors = er ++ e1.typeErrors ++ e2.typeErrors;
}


aspect production gteq
top::Expr ::= e1::Expr g::GTEQ_t e2::Expr
{
  local attribute er :: String;
  er = if ((e1.typerep.typeName == "Integer") &&
	   (e2.typerep.typeName == "Integer"))
	   ||
	  ((e1.typerep.typeName == "String") &&	
	   (e2.typerep.typeName == "String"))	
       then ""
       else top.location ++ "Error: Operands to >= must be of type Integer or String.\n";

  top.typeErrors = er ++ e1.typeErrors ++ e2.typeErrors;
}


aspect production lteq
top::Expr ::= e1::Expr l::LTEQ_t e2::Expr
{
  local attribute er :: String;
  er = if ((e1.typerep.typeName == "Integer") &&
	   (e2.typerep.typeName == "Integer"))
	   ||
	  ((e1.typerep.typeName == "String") &&	
	   (e2.typerep.typeName == "String"))	
       then ""
       else top.location ++ "Error: Operands to <= must be of type Integer or String.\n";
 
  top.typeErrors = er ++ e1.typeErrors ++ e2.typeErrors;
}


aspect production eqeq
top::Expr ::= e1::Expr eq::EQEQ_t e2::Expr
{
  local attribute er :: String;
  er = if (e1.typerep.isPrimative && 
	   e2.typerep.isPrimative && 
	   e1.typerep.typeEquals(e1.typerep, e2.typerep).bValue)
       then ""
       else top.location ++ "Error: Operands to == must be primative and of the same type.\n";

  top.typeErrors = er ++ e1.typeErrors ++ e2.typeErrors;
}


aspect production neq
top::Expr ::= e1::Expr n::NEQ_t e2::Expr
{
  local attribute er :: String;
  er = if (e1.typerep.isPrimative && 
	   e2.typerep.isPrimative && 
	   e1.typerep.typeEquals(e1.typerep, e2.typerep).bValue)
       then ""
       else top.location ++ "Error: Operands to == must be primative and of the same type.\n";

  top.typeErrors = er ++ e1.typeErrors ++ e2.typeErrors;
}


aspect production ifThenElse
top::Expr ::= i::If_kwd e1::Expr t::Then_kwd e2::Expr e::Else_kwd e3::Expr
{
  top.typeErrors = boolError ++ branchError ++ e1.typeErrors ++ e2.typeErrors ++ e3.typeErrors;

  local attribute boolError ::String;
  boolError = if (e1.typerep.typeName == "Boolean") then ""
	       else top.location ++ "Error: " ++ e1.pp ++ " must have type boolean.\n";


  local attribute branchError ::String;
  branchError = if e2.typerep.typeEquals(e2.typerep, e3.typerep).bValue then ""
 	         else top.location ++ "Error: The 'then' and 'else' branches must evaluate to the same type.\n" ++
                                      "\tInstead they are '" ++ e2.typerep.pp ++ "' and '" ++ e3.typerep.pp ++ "'\n";
}

aspect production intConst
top::Expr ::= i::Int_t
{
  top.typeErrors = "";
}

aspect production floatConst
top::Expr ::= f::Float_t
{
  top.typeErrors = "";
} 

aspect production plus
top::Expr ::= e1::Expr p::Plus_t e2::Expr
{
  local attribute er :: String;
  er = if ((e1.typerep.typeName == "Integer") && (e2.typerep.typeName == "Integer"))
	  ||
	  ((e1.typerep.typeName == "Float") && (e2.typerep.typeName == "Float"))
       then ""
       else top.location ++ "Error: Operands to '+' must be of type Integer or Float and of the same type.\n";

  top.typeErrors = er ++ e1.typeErrors ++ e2.typeErrors;
}
aspect production minus
top::Expr ::= e1::Expr m::Minus_t e2::Expr
{
  local attribute er :: String;
  er = if ((e1.typerep.typeName == "Integer") && (e2.typerep.typeName == "Integer"))
	  ||
	  ((e1.typerep.typeName == "Float") && (e2.typerep.typeName == "Float"))
       then ""
       else top.location ++ "Error: Operands to '-' must be of type Integer or Float and of the same type.\n";

  top.typeErrors = er ++ e1.typeErrors ++ e2.typeErrors;
}
aspect production multiply
top::Expr ::= e1::Expr m::Multiply_t e2::Expr
{
  local attribute er :: String;
  er = if ((e1.typerep.typeName == "Integer") && (e2.typerep.typeName == "Integer"))
	  ||
	  ((e1.typerep.typeName == "Float") && (e2.typerep.typeName == "Float"))
       then ""
       else top.location ++ "Error: Operands to '*' must be of type Integer or Float and of the same type.\n";

  top.typeErrors = er ++ e1.typeErrors ++ e2.typeErrors;
}
aspect production divide
top::Expr ::= e1::Expr d::Divide_t e2::Expr
{
  local attribute er :: String;
  er = if ((e1.typerep.typeName == "Integer") && (e2.typerep.typeName == "Integer"))
	  ||
	  ((e1.typerep.typeName == "Float") && (e2.typerep.typeName == "Float"))
       then ""
       else top.location ++ "Error: Operands to '/' must be of type Integer or Float and of the same type.\n";

  top.typeErrors = er ++ e1.typeErrors ++ e2.typeErrors;
}
aspect production neg
top::Expr ::= m::Minus_t e::Expr
{
  local attribute er :: String;
  er = if ((e.typerep.typeName == "Integer") || (e.typerep.typeName == "Float"))
       then ""
       else top.location ++ "Error: Opperand to unary '-' must be of type Integer or Float.\n";

  top.typeErrors = er ++ e.typeErrors;
}

aspect production stringConst
top::Expr ::= s::String_t
{
  top.typeErrors = "";
}

aspect production defaultPlusPlus
top::Expr ::= e1::Expr e2::Expr
{
  local attribute er :: String;
  er = if ((e1.typerep.typeName == "String") && (e2.typerep.typeName == "String"))
       then ""
       else top.location ++ "Error: Operands to overloaded operator '++' are not compatible.\n\t" ++
		e1.typerep.typeName ++ " ++ " ++ e2.typerep.typeName ++ "\n";

  top.typeErrors = er ++ e1.typeErrors ++ e2.typeErrors;
}

aspect production stringPlusPlus
top::Expr ::= e1::Expr e2::Expr
{
  top.typeErrors = e1.typeErrors ++ e2.typeErrors;
}


attribute typeErrors occurs on Exprs;

aspect production exprsEmpty
top::Exprs ::=
{
  top.typeErrors = "";
}

aspect production exprsSingle
top::Exprs ::= e::Expr
{
  top.typeErrors = e.typeErrors;
}

aspect production exprsCons
top::Exprs ::= e1::Expr c::Comma_t e2::Exprs
{
  top.typeErrors = e1.typeErrors ++ e2.typeErrors;
}

aspect production exprWith
top::Expr ::= ev::Decorate_kwd e::Expr w::With_kwd l::LCurly_t inh::ExprInh r::RCurly_t
{
  local attribute er :: String;
  er = if (e.typerep.isNonTerminal)
       then ""
       else top.location ++ "Error: Operand to 'decorate/with' must have a non-terminal type.\n" ++
                " Instead it is " ++ e.typerep.typeName ++ ".\n" ;

  top.typeErrors = e.typeErrors ++ inh.typeErrors ++ er;
}


attribute typeErrors occurs on ExprInh;
aspect production exprStmtsEmpty
top::ExprInh ::= 
{
  top.typeErrors = "";
}

aspect production exprStmtsCons
top::ExprInh ::= lhs::ExprLHSExpr ekwd::Equal_t e::Expr s::Semi_t inh::ExprInh
{ 
  local attribute valid :: Boolean;
  valid = lhs.typerep.typeEquals(lhs.typerep, e.typerep).bValue;

  local attribute er :: String;
  er = if valid then "" 
       else top.location ++ "Error: The LHS and RHS of the assignment must be the same type.\n" ++
                                  "\tInstead they are '" ++ lhs.typerep.typeName ++ "' and '" ++ e.typerep.typeName ++ "'\n" ; 

  top.typeErrors = e.typeErrors ++ inh.typeErrors ++ er;
}

