grammar silver:base:core;

terminal Aspect_kwd /aspect/ lexer precedence = 5;
concrete production aspectAGDcl
 top::AGDcl ::=
   a::Aspect_kwd p::Production_kwd id::QName ns::AspectSignature
   body::ProductionBody
{
  top.pp =  "aspect production " ++ id.pp ++ "\n" ++ 
	     ns.pp  ++ "  " ++ "\n{\n" ++ body.pp ++ "\n}";

  top.location = "[line: " ++ toString(a.line) ++ ", column: " ++ toString(a.column) ++ "] ";  

  top.defs = emptyDefs();
  top.globalDefs = addAspectDcl(fName, body.stmtList, emptyDefs());

  production attribute fNames :: [Decorated EnvItem];
  fNames = getFullNameDclOne(id.name, top.env);

  production attribute fName :: String;
  fName = if !null(fNames) then head(fNames).fullName else "FULL_NAME_NOT_FOUND";

  production attribute prods :: [Decorated EnvItem];
  prods = getProductionDclOne(fName, top.env);

  local attribute er :: String;
  er = if null(prods) || !head(prods).typerep.isProduction
        then top.location ++ "Error: Production for aspect '" ++ id.name ++ "' is not declared.\n"
        else "";	

  top.errors = er ++ ns.errors ++ body.errors;

  local attribute prodAtts :: Decorated Defs;
  prodAtts = getProdAtts(fName, top.env);



  body.env = appendDefsEnv(
		appendDefs(prodAtts, 
	     	appendDefs(body.defs, 
	     	appendDefs(ns.defs, 
	     	addThisDcl(fName, emptyDefs())))), top.env
		);


  body.signatureEnv = toEnv(ns.defs);


  body.localsEnv = toEnv(appendDefs(prodAtts, body.defs));
  body.signatureNames = ns.sigNames;
  body.signatureTypes = ns.sigTypes;

  ns.env = appendDefsEnv(ns.defs, top.env);
  
  ns.realNames = if null(prods) then [] else cons(head(prods).lhsName, head(prods).rhsNames);
  ns.realTypes = if null(prods) then [] else cons(head(prods).typerep.outputType, head(prods).typerep.inputTypes);
}


inherited attribute realNames :: [String];
inherited attribute realTypes :: [Decorated TypeRep];

attribute realNames occurs on AspectSignature, AspectLHS, AspectLHSElem, AspectRHS, AspectRHSElem;
attribute realTypes occurs on AspectSignature, AspectLHS, AspectLHSElem, AspectRHS, AspectRHSElem;


nonterminal AspectSignature with {pp, location, grammarName, defs, env, errors};
concrete production aspectSignatureEmptyRHS
top::AspectSignature ::= lhs::AspectLHS kwd::CCEQ_t
{
  top.pp = lhs.pp ++ " ::=";
  top.location = lhs.location;

  top.defs = lhs.defs;
  top.errors = lhs.errors;

  top.sigNames = lhs.sigNames;
  top.sigTypes = lhs.sigTypes;

  lhs.realNames = if null(top.realNames) then [] else [head(top.realNames)];
  lhs.realTypes = if null(top.realTypes) then [] else [head(top.realTypes)];
}

concrete production aspectSignature
top::AspectSignature ::= lhs::AspectLHS kwd::CCEQ_t rhs::AspectRHS 
{
  top.pp = lhs.pp ++ " ::= " ++ rhs.pp;
  top.location = lhs.location;

  top.defs = appendDefs(lhs.defs, rhs.defs);
  top.errors = lhs.errors ++ rhs.errors;

  top.sigNames = lhs.sigNames +++ rhs.sigNames;
  top.sigTypes = lhs.sigTypes +++ rhs.sigTypes;

  lhs.realNames = if null(top.realNames) then [] else [head(top.realNames)];
  lhs.realTypes = if null(top.realTypes) then [] else [head(top.realTypes)];

  rhs.realNames = if null(top.realNames) then [] else tail(top.realNames);
  rhs.realTypes = if null(top.realTypes) then [] else tail(top.realTypes);
}

nonterminal AspectLHS with {pp, location, grammarName, defs, env, errors};
concrete production aspectLHSSingle
top::AspectLHS ::= lhs::AspectLHSElem
{
  top.pp = lhs.pp;
  top.location = lhs.location;
  top.defs = lhs.defs;
  top.errors = lhs.errors;

  top.sigNames = lhs.sigNames;
  top.sigTypes = lhs.sigTypes;

  lhs.realNames = top.realNames;
  lhs.realTypes = top.realTypes;
}

concrete production aspectLHSElemNone
top::AspectLHSElem ::= id::UnderScore_t
{
  top.pp = "_";

  forwards to aspectLHSElem(terminal(Id_t, "_" ++ toString(genint())));
}

nonterminal AspectLHSElem with {pp, location, grammarName, defs, env, errors};

--The reason this does not forward to aspectLHSElemFull is becuase
--we do not have the Type nonterminal
concrete production aspectLHSElem
top::AspectLHSElem ::= id::Id_t
{
  top.pp = id.lexeme;
  top.location = "[line: " ++ toString(id.line) ++ ", column: " ++ toString(id.column) ++ "] ";  

  production attribute rName :: String;
  rName = if null(top.realNames) then id.lexeme else head(top.realNames);

  production attribute rType :: Decorated TypeRep;
  rType = if null(top.realTypes) then topTypeRep() else head(top.realTypes);

  production attribute fName :: String;
--  fName = id.lexeme;
  fName = rName;

  top.defs = addValueDcl(rName, rType, 
	     addFullNameDcl(id.lexeme, rName,  emptyDefs()));

  top.sigNames = [id.lexeme];
  top.sigTypes = [rType];

  local attribute er1 :: String;
  er1 = if length(getFullNameDcl(id.lexeme, top.env)) > 1
       then top.location ++ "Error: Name '" ++ id.lexeme ++ "' is already bound.\n"
       else "";	

  local attribute er2 :: String;
  er2 = if length(getValueDcl(fName, top.env)) > 1
       then top.location ++ "Error: Value '" ++ fName ++ "' is already bound.\n"
       else "";	

  top.errors = er1 ++ er2;
}

concrete production aspectLHSElemFull
top::AspectLHSElem ::= id::Id_t kwd::HasType_t t::Type
{
  top.pp = id.lexeme ++ "::" ++ t.pp;
  top.location = "[line: " ++ toString(id.line) ++ ", column: " ++ toString(id.column) ++ "] ";  

  production attribute rName :: String;
  rName = if null(top.realNames) then id.lexeme else head(top.realNames);

  production attribute fName :: String;
--  fName = id.lexeme;
  fName = rName;

  top.defs = addValueDcl(rName, t.typerep, 
	     addFullNameDcl(id.lexeme, rName,  emptyDefs()));

  top.sigNames = [id.lexeme];
  top.sigTypes = [t.typerep];

  local attribute er1 :: String;
  er1 = if length(getFullNameDcl(id.lexeme, top.env)) > 1
       then top.location ++ "Error: Name '" ++ id.lexeme ++ "' is already bound.\n"
       else "";	

  local attribute er2 :: String;
  er2 = if length(getValueDcl(fName, top.env)) > 1
       then top.location ++ "Error: Value '" ++ fName ++ "' is already bound.\n"
       else "";	

  top.errors = er1 ++ er2 ++ t.errors;
}

nonterminal AspectRHS with {pp, location, grammarName, defs, env, errors};
abstract production aspectRHSEmpty
top::AspectRHS ::=
{
  top.pp = "";
  top.location = "[line: -1, column: -1, aspectRHSEmpty()] ";
  top.defs = emptyDefs();
  top.errors = "";

  top.sigNames = [];
  top.sigTypes = [];
}

concrete production aspectRHSSingle
top::AspectRHS ::= rhs::AspectRHSElem
{
  top.pp = rhs.pp;
  top.location = rhs.location;
  top.defs = rhs.defs;
  top.errors = rhs.errors;

  --We don't want to give more than one.  This happens when the aspect has less names than the production.
  rhs.realNames = if null(top.realNames) then [] else [head(top.realNames)];
  rhs.realTypes = if null(top.realTypes) then [] else [head(top.realTypes)];

  top.sigNames = rhs.sigNames;
  top.sigTypes = rhs.sigTypes;
}

concrete production aspectRHSCons
top::AspectRHS ::= h::AspectRHSElem t::AspectRHS
{
  top.pp = h.pp ++ " " ++ t.pp;
  top.location = h.location;
  top.defs = appendDefs(h.defs, t.defs);
  top.errors = h.errors ++ t.errors;

  top.sigNames = h.sigNames +++ t.sigNames;
  top.sigTypes = h.sigTypes +++ t.sigTypes;

  h.realNames = if null(top.realNames) then [] else [head(top.realNames)];
  h.realTypes = if null(top.realTypes) then [] else [head(top.realTypes)];

  t.realNames = if null(top.realNames) then [] else tail(top.realNames);
  t.realTypes = if null(top.realTypes) then [] else tail(top.realTypes);
}

nonterminal AspectRHSElem with {pp, location, grammarName, defs, env, errors, name, typerep};
terminal UnderScore_t /[\_]/ lexer precedence = 5;

concrete production aspectRHSElemNone
top::AspectRHSElem ::= id::UnderScore_t
{
  top.pp = "_";

  forwards to aspectRHSElem(terminal(Id_t, "_" ++ toString(genint())));
}

concrete production aspectRHSElem
top::AspectRHSElem ::= id::Id_t
{
  top.pp = id.lexeme;

  production attribute rName :: String;
  rName = if null(top.realNames) then id.lexeme else head(top.realNames);

  production attribute rType :: Decorated TypeRep;
  rType = if null(top.realTypes) then topTypeRep() else head(top.realTypes);

  production attribute fName :: String;
--  fName = id.lexeme;
  fName = rName;

  top.defs = addValueDcl(rName, rType, 
	     addFullNameDcl(id.lexeme, rName,  emptyDefs()));

  top.location = "[line: " ++ toString(id.line) ++ ", column: " ++ toString(id.column) ++ "] ";  

  top.sigNames = [id.lexeme];
  top.sigTypes = [rType];

  local attribute er1 :: String;
  er1 = if length(getFullNameDcl(id.lexeme, top.env)) > 1
       then top.location ++ "Error: Name '" ++ id.lexeme ++ "' is already bound.\n"
       else "";	

  local attribute er2 :: String;
  er2 = if length(getValueDcl(fName, top.env)) > 1
       then top.location ++ "Error: Value '" ++ fName ++ "' is already bound.\n"
       else "";	

  top.errors = er1 ++ er2;
}

concrete production aspectRHSElemFull
top::AspectRHSElem ::= id::Id_t kwd::HasType_t t::Type
{
  top.pp = id.lexeme ++ "::" ++ t.pp;
  top.location = "[line: " ++ toString(id.line) ++ ", column: " ++ toString(id.column) ++ "] ";  

  production attribute rName :: String;
  rName = if null(top.realNames) then id.lexeme else head(top.realNames);

  production attribute fName :: String;
--  fName = id.lexeme;
  fName = rName;

  top.defs = addValueDcl(rName, t.typerep, 
	     addFullNameDcl(id.lexeme, rName,  emptyDefs()));

  top.sigNames = [id.lexeme];
  top.sigTypes = [t.typerep];

  local attribute er1 :: String;
  er1 = if length(getFullNameDcl(id.lexeme, top.env)) > 1
       then top.location ++ "Error: Name '" ++ id.lexeme ++ "' is already bound.\n"
       else "";	

  local attribute er2 :: String;
  er2 = if length(getValueDcl(fName, top.env)) > 1
       then top.location ++ "Error: Value '" ++ fName ++ "' is already bound.\n"
       else "";	

  top.errors = er1 ++ er2 ++ t.errors;
}