grammar silver:base:core;

terminal Length_kwd /length/ lexer precedence = 5;
concrete production lengthFunction
top::Expr ::= l::Length_kwd lp::LParen_t e::Expr rp::RParen_t
{
  production attribute handlers :: [Expr] with epp;
  handlers := [];

  forwards to if null(handlers) then unknownLength(e'') else head(handlers);
}

function epp
[Expr] ::= e1::[Expr] e2::[Expr]
{
  return e1 +++ e2;
}

abstract production unknownLength
top::Expr ::= e::Expr
{
  top.pp = "length ( " ++ e.pp ++ " )";
  top.location = e.location;
 
  top.errors = e.errors;
  top.typerep = topTypeRep();
}


aspect production lengthFunction
top::Expr ::= l::Length_kwd lp::LParen_t e::Expr rp::RParen_t
{
  handlers <- if e.typerep.typeName == "String"
	      then [stringLength(e'')]
	      else [];
}

abstract production stringLength
top::Expr ::= e::Expr
{
  top.pp = "length ( " ++ e.pp ++ " )";
  top.location = e.location;
 
  top.errors = e.errors;
  top.typerep = integerTypeRep();
}


terminal IndexOf_kwd /indexof/ lexer precedence = 5;
concrete production indexOfFunction
top::Expr ::= i::IndexOf_kwd l::LParen_t e1::Expr c::Comma_t e2::Expr r::RParen_t
{
  top.pp = "indexof ( " ++ e1.pp ++ ", " ++ e2.pp ++ " )";
  top.location = "[line: " ++ toString(i.line) ++ ", column: " ++ toString(i.column) ++ "] ";

  top.errors = e1.errors ++ e2.errors;
  top.typerep = integerTypeRep();
}


terminal SubString_kwd /substring/ lexer precedence = 5;
concrete production subStringFunction
top::Expr ::= s::SubString_kwd l::LParen_t e1::Expr c1::Comma_t e2::Expr c2::Comma_t e3::Expr r::RParen_t
{
  top.pp = "substring ( " ++ e1.pp ++ ", " ++ e2.pp ++ ", " ++ e3.pp ++ " )";
  top.location = "[line: " ++ toString(s.line) ++ ", column: " ++ toString(s.column) ++ "] ";

  top.errors = e1.errors ++ e2.errors ++ e3.errors;

  top.typerep = stringTypeRep();
}


terminal Error_kwd /error/ lexer precedence = 5;
concrete production errorFunction
top::Expr ::= er::Error_kwd l::LParen_t e::Expr r::RParen_t
{
  top.pp = "error ( " ++ e.pp ++ " )";
  top.location = "[line: " ++ toString(er.line) ++ ", column: " ++ toString(er.column) ++ "] ";

  top.errors = e.errors;
  top.typerep = topTypeRep();
}

terminal ToInt_kwd /toInt/ lexer precedence = 5;
concrete production stringToIntFunction
top::Expr ::= t::ToInt_kwd l::LParen_t e::Expr r::RParen_t
{
  top.pp = "toInt ( " ++ e.pp ++ " )";
  top.location = "[line: " ++ toString(t.line) ++ ", column: " ++ toString(t.column) ++ "] ";

  top.errors = e.errors;
  top.typerep = integerTypeRep();
}

terminal ToFloat_kwd /toFloat/ lexer precedence = 5;
concrete production stringToFloatFunction
top::Expr ::= t::ToFloat_kwd l::LParen_t e::Expr r::RParen_t
{
  top.pp = "toFloat ( " ++ e.pp ++ " )";
  top.location = "[line: " ++ toString(t.line) ++ ", column: " ++ toString(t.column) ++ "] ";

  top.errors = e.errors;
  top.typerep = floatTypeRep();
}

terminal ToString_kwd /toString/ lexer precedence = 5;
concrete production numToStringFunction
top::Expr ::= t::ToString_kwd l::LParen_t e::Expr r::RParen_t
{
  top.pp = "toString ( " ++ e.pp ++ " )";
  top.location = "[line: " ++ toString(t.line) ++ ", column: " ++ toString(t.column) ++ "] ";

  top.errors = e.errors;
  top.typerep = stringTypeRep();
}

terminal FTOI_kwd /ftoi/ lexer precedence = 5;
concrete production ftoiFunction
top::Expr ::= t::FTOI_kwd l::LParen_t e::Expr r::RParen_t
{
  top.pp = "ftoi ( " ++ e.pp ++ " )";
  top.location = "[line: " ++ toString(t.line) ++ ", column: " ++ toString(t.column) ++ "] ";

  top.errors = e.errors;
  top.typerep = integerTypeRep();
}

terminal ITOF_kwd /itof/ lexer precedence = 5;
concrete production itofFunction
top::Expr ::= t::ITOF_kwd l::LParen_t e::Expr r::RParen_t
{
  top.pp = "itof ( " ++ e.pp ++ " )";
  top.location = "[line: " ++ toString(t.line) ++ ", column: " ++ toString(t.column) ++ "] ";

  top.errors = e.errors;
  top.typerep = floatTypeRep();
}


terminal IsDigit_kwd /isDigit/ lexer precedence = 5;
concrete production isDigitFunction
top::Expr ::= t::IsDigit_kwd l::LParen_t e::Expr r::RParen_t
{
  top.pp = "isdigit ( " ++ e.pp ++ " )";
  top.location = "[line: " ++ toString(t.line) ++ ", column: " ++ toString(t.column) ++ "] ";

  top.errors = e.errors;
  top.typerep = booleanTypeRep();
}

terminal IsAlpha_kwd /isAlpha/ lexer precedence = 5;
concrete production isAlphaFunction
top::Expr ::= t::IsAlpha_kwd l::LParen_t e::Expr r::RParen_t
{
  top.pp = "isAlpha ( " ++ e.pp ++ " )";
  top.location = "[line: " ++ toString(t.line) ++ ", column: " ++ toString(t.column) ++ "] ";

  top.errors = e.errors;
  top.typerep = booleanTypeRep();
}

terminal IsSpace_kwd /isSpace/ lexer precedence = 5;
concrete production isSpaceFunction
top::Expr ::= t::IsSpace_kwd l::LParen_t e::Expr r::RParen_t
{
  top.pp = "isSpace ( " ++ e.pp ++ " )";
  top.location = "[line: " ++ toString(t.line) ++ ", column: " ++ toString(t.column) ++ "] ";

  top.errors = e.errors;
  top.typerep = booleanTypeRep();
}

terminal IsLower_kwd /isLower/ lexer precedence = 5;
concrete production isLowerFunction
top::Expr ::= t::IsLower_kwd l::LParen_t e::Expr r::RParen_t
{
  top.pp = "isLower ( " ++ e.pp ++ " )";
  top.location = "[line: " ++ toString(t.line) ++ ", column: " ++ toString(t.column) ++ "] ";

  top.errors = e.errors;
  top.typerep = booleanTypeRep();
}

terminal IsUpper_kwd /isUpper/ lexer precedence = 5;
concrete production isUpperFunction
top::Expr ::= t::IsUpper_kwd l::LParen_t e::Expr r::RParen_t
{
  top.pp = "isUpper ( " ++ e.pp ++ " )";
  top.location = "[line: " ++ toString(t.line) ++ ", column: " ++ toString(t.column) ++ "] ";

  top.errors = e.errors;
  top.typerep = booleanTypeRep();
}

terminal New_kwd /new/ lexer precedence = 5;
concrete production newFunction
top::Expr ::= n::New_kwd l::LParen_t e::Expr r::RParen_t
{
  top.pp = "new ( " ++ e.pp ++ " )";
  top.location = "[line: " ++ toString(n.line) ++ ", column: " ++ toString(n.column) ++ "] ";

  top.errors = e.errors;
--  top.typerep = ntTypeRep(e.typerep.typeName);
  top.typerep = e.typerep.decoratedType ;

  e.userFriendly = 1;
}

concrete production terminalFunction
top::Expr ::= term::Terminal_kwd l::LParen_t t::Type c::Comma_t e::Expr r::RParen_t
{
  top.pp = "terminal ( " ++ t.pp ++ ", " ++ e.pp ++ " )";
  top.location = "[line: " ++ toString(term.line) ++ ", column: " ++ toString(term.column) ++ "] ";

  top.errors = t.errors ++ e.errors;
  top.typerep = t.typerep;
}

concrete production terminalFunction2
top::Expr ::= term::Terminal_kwd l::LParen_t t::Type c1::Comma_t e1::Expr c2::Comma_t e2::Expr c3::Comma_t e3::Expr r::RParen_t
{
  top.pp = "terminal ( " ++ t.pp ++ ", " ++ e1.pp ++ ", " ++ e2.pp ++ ", " ++ e3.pp ++ " )";
  top.location = "[line: " ++ toString(term.line) ++ ", column: " ++ toString(term.column) ++ "] ";

  top.errors = t.errors ++ e1.errors ++ e2.errors ++ e3.errors;
  top.typerep = t.typerep;
}
