grammar silver:base:core;

closed nonterminal EnvItem 
  with {itemName, fullName, decoratesName, typerep, stmtList,
        isValueDeclaration, isTypeDeclaration, isOccursDeclaration, isInheritedDeclaration,
        isSynthesizedDeclaration, isThisDeclaration, isAttributeDeclaration,
        isFullNameDeclaration, isCloseDeclaration, isAspectDeclaration, pp, isProdAttsDeclaration, defs,
	lhsName, rhsNames, isProductionDeclaration};

synthesized attribute itemName :: String;

synthesized attribute typerep :: Decorated TypeRep;
synthesized attribute fullName :: String;
synthesized attribute decoratesName :: String;

synthesized attribute lhsName :: String;
synthesized attribute rhsNames :: [String];

synthesized attribute isValueDeclaration :: Boolean;
synthesized attribute isFullNameDeclaration :: Boolean;
synthesized attribute isTypeDeclaration :: Boolean;
synthesized attribute isAttributeDeclaration :: Boolean;
synthesized attribute isProductionDeclaration :: Boolean;
synthesized attribute isOccursDeclaration :: Boolean;
synthesized attribute isInheritedDeclaration :: Boolean;
synthesized attribute isSynthesizedDeclaration :: Boolean;
synthesized attribute isThisDeclaration :: Boolean;
synthesized attribute isCloseDeclaration :: Boolean;
synthesized attribute isAspectDeclaration :: Boolean;
synthesized attribute isProdAttsDeclaration ::Boolean;

----------------------------------------------------------------------------------------------------
--item creation functions---------------------------------------------------------------------------
----------------------------------------------------------------------------------------------------
function valueEnvItem
Decorated EnvItem ::= n::String value::Decorated TypeRep
{
  return decorate internal_valueEnvItem(n, value) with {};
}
abstract production internal_valueEnvItem
top::EnvItem ::= n::String value::Decorated TypeRep
{
  -- required to be defined.
  top.itemName = n;
  top.pp = n ++ " :: " ++ value.pp ; 

  -- will be defined based on what type of item it is.  
  top.typerep = value;

  -- required to be defined.
  top.isValueDeclaration = true;

  forwards to internal_defaultEnvItem();
}

function typeEnvItem
Decorated EnvItem ::= n::String type::Decorated TypeRep
{
  return decorate internal_typeEnvItem(n, type) with {};
}

abstract production internal_typeEnvItem
top::EnvItem ::= n::String value::Decorated TypeRep
{
  -- required to be defined.
  top.itemName = n;
  top.pp = n ++ " |-> " ++ value.pp ;

  -- will be defined based on what type of item it is.  
  top.typerep = value ;

  top.isTypeDeclaration = true;

  forwards to internal_defaultEnvItem();
}

function productionEnvItem
Decorated EnvItem ::= n::String lhs::String rhs::[String] t::Decorated TypeRep
{
  return decorate internal_productionEnvItem(n, lhs, rhs, t) with {};
}

abstract production internal_productionEnvItem
top::EnvItem ::= n::String lhs::String rhs::[String] t::Decorated TypeRep
{
  -- required to be defined.
  top.itemName = n;
  top.pp = n ++ " :: " ++ t.pp;

  -- will be defined based on what type of item it is.  
  top.typerep = t;

  top.lhsName = lhs;
  top.rhsNames = rhs;

  -- required to be defined.
  top.isProductionDeclaration = true;

  forwards to internal_defaultEnvItem();
}

function attributeEnvItem
Decorated EnvItem ::= n::String value::Decorated TypeRep
{
  return decorate internal_attributeEnvItem(n, value) with {};
}

abstract production internal_attributeEnvItem
top::EnvItem ::= n::String value::Decorated TypeRep
{
  -- required to be defined.
  top.itemName = n;
  top.pp = "attr " ++ n ++ " :: " ++ value.pp ; 

  -- will be defined based on what type of item it is.  
  top.typerep = value;

  top.isAttributeDeclaration = true;

  forwards to internal_defaultEnvItem();
}


function occursEnvItem
Decorated EnvItem ::= n::String dn::String
{
  return decorate internal_occursEnvItem(n, dn) with {};
}

abstract production internal_occursEnvItem
top::EnvItem ::= n::String dn::String
{
  -- required to be defined.
  top.itemName = n;
  top.pp = n ++ " @ " ++ dn ; 
  
  top.decoratesName = dn;

  top.isOccursDeclaration = true;

  forwards to internal_defaultEnvItem();
}

function inheritedEnvItem
Decorated EnvItem ::= n::String
{
  return decorate internal_inheritedEnvItem(n) with {};
}
abstract production internal_inheritedEnvItem
top::EnvItem ::= n::String
{
  -- required to be defined.
  top.itemName = n;
  top.pp = "inh " ++ n ;

  top.isInheritedDeclaration = true;

  forwards to internal_defaultEnvItem();
}

function synthesizedEnvItem
Decorated EnvItem ::= n::String
{
  return decorate internal_synthesizedEnvItem(n) with {};
}
abstract production internal_synthesizedEnvItem
top::EnvItem ::= n::String
{
  -- required to be defined.
  top.itemName = n;
  top.pp = "syn " ++ n ;

  top.isSynthesizedDeclaration = true;

  forwards to internal_defaultEnvItem();
}
function thisEnvItem
Decorated EnvItem ::= n::String
{
  return decorate internal_thisEnvItem(n) with {};
}
abstract production internal_thisEnvItem
top::EnvItem ::= n::String
{
  -- required to be defined.
  top.itemName = n;
  top.pp = "this " ++ n;

  top.isThisDeclaration = true;

  forwards to internal_defaultEnvItem();
}

function fullNameEnvItem
Decorated EnvItem ::= n::String fname::String
{
  return decorate internal_fullNameEnvItem(n, fname) with {};
}
abstract production internal_fullNameEnvItem
top::EnvItem ::= n::String fname::String
{
  -- required to be defined.
  top.itemName = n;
  top.pp = n  ++ " <-> " ++ fname;

  top.fullName = fname;
  top.isFullNameDeclaration = true;
  forwards to internal_defaultEnvItem();
}


function closeEnvItem
Decorated EnvItem ::= n::String
{
  return decorate internal_closeEnvItem(n) with {};
}
abstract production internal_closeEnvItem
top::EnvItem ::= n::String
{
  -- required to be defined.
  top.itemName = n;
  top.pp = "!" ++ n;

  top.isCloseDeclaration = true;

  forwards to internal_defaultEnvItem();
}

function aspectEnvItem
Decorated EnvItem ::= n::String stmts::[StmtRep]
{
  return decorate internal_aspectEnvItem(n, stmts'') with {};
}

abstract production internal_aspectEnvItem
top::EnvItem ::= n::String stmts::[StmtRep]
{
  top.itemName = n;
  top.pp = "aspect " ++ n;

  top.stmtList = stmts'';

  top.isAspectDeclaration = true;

  forwards to internal_defaultEnvItem();
}

function prodAttsEnvItem
Decorated EnvItem ::= n::String d::Decorated Defs
{
  return decorate internal_prodAttsEnvItem(n, d) with {};
}

abstract production internal_prodAttsEnvItem
top::EnvItem ::= n::String d::Decorated Defs
{
  top.itemName = n;
  top.pp = "prodAtts " ++ n;

  top.defs = d;

  top.isProdAttsDeclaration = true;

  forwards to internal_defaultEnvItem();
}

--function startEnvItem
--Decorated EnvItem ::= n::String d::Decorated TypeRep
--{
--  return decorate internal_startEnvItem(n, d) with {};
--}
--
--abstract production internal_startEnvItem
--top::EnvItem ::= n::String d::Decorated TypeRep
--{
--  top.itemName = n;
--  top.pp = "start " ++ n;
--
--  top.typerep = d;
--
--  top.isStartDeclaration = true;
--
--  forwards to internal_defaultEnvItem();
--}

function defaultEnvItem
Decorated EnvItem ::=
{
  return decorate internal_defaultEnvItem() with {};
}


abstract production internal_defaultEnvItem
top::EnvItem ::= 
{
  top.itemName = "_DEFAULT_ENV_ITEM";
  top.pp = "DEFAULT ENV ITEM";

  top.typerep = decorate internal_defaultTypeRep() with {};
  top.lhsName = "_DEFAULT_ENV_ITEM";
  top.rhsNames = [];
  top.fullName = "_DEFAULT_ENV_ITEM";
  top.decoratesName = "_DEFAULT_ENV_ITEM";
  top.stmtList = [];
  top.defs = emptyDefs();

  -- required to be defined.
  top.isValueDeclaration = false;
  top.isTypeDeclaration = false;
  top.isProductionDeclaration = false;
  top.isAttributeDeclaration = false;
  top.isOccursDeclaration = false;
  top.isInheritedDeclaration = false;
  top.isSynthesizedDeclaration = false;
  top.isThisDeclaration = false;
  top.isFullNameDeclaration = false;
  top.isCloseDeclaration = false;
  top.isAspectDeclaration = false;
  top.isProdAttsDeclaration = false;
--  top.isStartDeclaration = false;
}

function containsItem
Boolean ::= s::String e::[Decorated EnvItem]
{
  return if null(e)
	 then false
	 else head(e).itemName == s || containsItem(s, tail(e));
}

function mapEnvItems
[Decorated EnvItem] ::= mapper::EnvMap e::[Decorated EnvItem]
{
  mapper.inEnvItem = head(e);

  return if null(e)
	 then []
	 else cons(mapper.newEnvItem, mapEnvItems(mapper'', tail(e)));
}

function filterEnvItems
[Decorated EnvItem] ::= fil::EnvFilter e::[Decorated EnvItem]
{
  fil.inEnvItem = head(e);

  local attribute recurse :: [Decorated EnvItem];
  recurse = filterEnvItems(fil'', tail(e));

  return if null(e)
	 then []  
	 else if fil.keep
	      then cons(head(e), recurse)
         else recurse;
}

function printEnvItems
String ::= es::[Decorated EnvItem]
{
  return if null(es) then "\n" else head(es).pp ++ "\n" ++ printEnvItems(tail(es));
}
