grammar silver:base:core;
import silver:utils;
import core;

terminal Import_kwd  /import/ lexer precedence = 5 ;
terminal Export_kwd  /export/ lexer precedence = 5 ;
terminal Only_kwd    /only/   lexer precedence = 5 ;
terminal Hiding_kwd  /hiding/ lexer precedence = 5 ;
terminal With_kwd    /with/   lexer precedence = 5 ;
terminal As_kwd      /as/     lexer precedence = 5 ;


synthesized attribute includedGrammars :: [String];

synthesized attribute localDefs :: Decorated Defs;
attribute localDefs occurs on ImportExportStmts;

nonterminal ImportExportStmts with {pp, grammarName, location, env, errors, includedGrammars, compiledGrammars, exportedDefs, importedEnv} ;

abstract production importStmtsNone
top::ImportExportStmts ::=
{
  top.pp = "" ;
  top.location = "[line: -1, column: -1, importStmtsNone()] ";
  top.importedEnv = emptyEnv();
  top.errors = "";
  top.includedGrammars = [::String];
  top.exportedDefs = emptyDefs();
  top.localDefs = emptyDefs();
}

concrete production importStmtsOne 
top::ImportExportStmts ::= im::ImportStmt
{
  top.pp = im.pp ;
  top.location = im.location;
  top.importedEnv = im.importedEnv;
  top.errors = im.errors;
  top.includedGrammars = im.includedGrammars;
  top.exportedDefs = emptyDefs();
  top.localDefs = emptyDefs();
  
  im.compiledGrammars = top.compiledGrammars;
}

concrete production exportStmtsOne 
top::ImportExportStmts ::= ex::ExportStmt
{
  top.pp = ex.pp ;
  top.location = ex.location;
  top.importedEnv = emptyEnv();
  top.errors = ex.errors;
  top.includedGrammars = ex.includedGrammars;
  top.exportedDefs = ex.exportedDefs;
  top.localDefs = emptyDefs();

  ex.compiledGrammars = top.compiledGrammars;
}

concrete production importStmtsCons
top::ImportExportStmts ::= h::ImportStmt t::ImportExportStmts
{
  top.pp = h.pp ++ "\n" ++ t.pp ;
  top.location = h.location;
  top.importedEnv = appendEnv(h.importedEnv, t.importedEnv);
  top.errors = h.errors ++ t.errors;

  top.includedGrammars = h.includedGrammars +++ t.includedGrammars;
  top.exportedDefs = t.exportedDefs;
  top.localDefs = t.localDefs;
 
  h.compiledGrammars = top.compiledGrammars;
  t.compiledGrammars = top.compiledGrammars;
}

concrete production exportStmtsCons
top::ImportExportStmts ::= h::ExportStmt t::ImportExportStmts
{
  top.pp = h.pp ++ "\n" ++ t.pp ;
  top.location = h.location;
  top.importedEnv = t.importedEnv;

  top.errors = h.errors ++ t.errors;
  top.includedGrammars = h.includedGrammars +++ t.includedGrammars;
  top.exportedDefs = appendDefs(h.exportedDefs, t.exportedDefs);
  top.localDefs = t.localDefs;

  h.compiledGrammars = top.compiledGrammars;
  t.compiledGrammars = top.compiledGrammars;
}

abstract production appendImportsExports
top::ImportExportStmts ::= fst::ImportExportStmts snd::ImportExportStmts
{
  top.pp = fst.pp ++ "\n" ++ snd.pp ;
  top.location = fst.location;

  top.importedEnv = appendEnv(fst.importedEnv, snd.importedEnv);
  top.errors = fst.errors ++ snd.errors;

  top.includedGrammars = fst.includedGrammars +++ snd.includedGrammars;
  top.exportedDefs = appendDefs(fst.exportedDefs, snd.exportedDefs);
  top.localDefs = appendDefs(fst.localDefs, snd.localDefs);
  fst.compiledGrammars = top.compiledGrammars;
  snd.compiledGrammars = top.compiledGrammars;
}

--Import definitons in Imports.sv, Export definitions in Exports.sv.

nonterminal OptWithClause with {pp, grammarName, location, env, errors, envMapList} ;
nonterminal WithElems with {pp, grammarName, location, env, errors, envMapList} ;
nonterminal WithElem with {pp, grammarName, location, env, errors, envMapList} ;

synthesized attribute envMapList :: [EnvMap];

concrete production withClauseEmpty
top::OptWithClause ::= 
{
  top.pp = "" ;
  top.location = "[line: -1, column: -1, withClauseEmpty()] ";
  top.errors = "";
  top.envMapList = [];
}
concrete production withClause
top::OptWithClause ::= w::With_kwd  wcs:: WithElems
{ 
  top.pp = " with " ++ wcs.pp ;
  top.location = "[line: " ++ toString(w.line) ++ ", column: " ++ toString(w.column) ++ "] ";
  top.errors = wcs.errors;
  top.envMapList = wcs.envMapList;
}
concrete production withElemsOne
top::WithElems ::= we::WithElem
{
  top.pp = we.pp ;
  top.location = we.location;
  top.errors = we.errors;
  top.envMapList = we.envMapList;
}
concrete production withElemsCons
top::WithElems  ::= h::WithElem  c::Comma_t  t::WithElems
{
  top.pp = h.pp ++ ", " ++ t.pp ;
  top.location = h.location;
  top.errors = h.errors ++ t.errors;
  top.envMapList = h.envMapList +++ t.envMapList;
}
concrete production withElement
top::WithElem ::= n::Id_t a::As_kwd  newname::Id_t 
{
  top.pp = n.lexeme ++ " as " ++ newname.lexeme ;
  top.location = "[line: " ++ toString(n.line) ++ ", column: " ++ toString(n.column) ++ "] ";
  top.errors = "";

  top.envMapList = [renameMap(n.lexeme, newname.lexeme)];
}

synthesized attribute names :: [String];

nonterminal NameList with {pp, grammarName, location, env, errors, names} ;
concrete production nameListOne
top::NameList ::= n::QName
{ 
  top.pp = n.pp ;
  top.location = n.location;
  top.errors = "";
  top.names = [n.name];
}
concrete production nameListCons
top::NameList ::= h::QName  c::Comma_t  t::NameList
{
  top.pp = h.pp ++ ", " ++ t.pp ;
  top.location = h.location;
  top.errors = t.errors;
  top.names = cons(h.name, t.names);
}

nonterminal MaybeRootSpec with {found, rootSpec};
abstract production getGrammar
top::MaybeRootSpec ::= search::String grammars::[GrammarSpec]
{
  local attribute recurse :: MaybeRootSpec;
  recurse = getGrammar(search, tail(grammars));

  top.found = if null(grammars)
	      then false
	      else if head(grammars).name == search
	           then true
                   else recurse.found;

  top.rootSpec = if null(grammars)
	         then error("Value " ++ search ++ " was not found in the PackageEnv")
	         else if head(grammars).name == search
	              then head(grammars).rootSpec
                      else recurse.rootSpec;
}

abstract production renameMap
top::EnvMap ::= old::String n::String
{
  local attribute item :: Decorated EnvItem;
  item = top.inEnvItem;

  top.newEnvItem = if item.isFullNameDeclaration && item.itemName == old
		   then fullNameEnvItem(n, item.fullName)
		   else item;
}

abstract production prependMap
top::EnvMap ::= prefix::String
{
  local attribute item :: Decorated EnvItem;
  item = top.inEnvItem;

  top.newEnvItem = if item.isFullNameDeclaration
		   then fullNameEnvItem(prefix ++ ":" ++ item.itemName, item.fullName)
		   else item;
}

function applyMappings
Decorated Defs ::= maps::[EnvMap] old::Decorated Defs
{
  return if null(maps)
	 then old
	 else mapDefs(head(maps), applyMappings(tail(maps), old));
}

abstract production removeFilter
top::EnvFilter ::= n::String
{
  local attribute item :: Decorated EnvItem;
  item = top.inEnvItem;

  top.keep = (item.itemName != n) || (!item.isFullNameDeclaration);     
}

abstract production keepFilter
top::EnvFilter ::= n::[String]
{
  local attribute item :: Decorated EnvItem;
  item = top.inEnvItem;

  top.keep = (!item.isFullNameDeclaration) || (containsString(item.itemName, n));
}

function buildRemoveFilters
[EnvFilter] ::= ns::[String]
{
  return if null(ns)
	 then []
	 else cons(removeFilter(head(ns)), buildRemoveFilters(tail(ns)));
}

function applyFilters
Decorated Defs ::= fils::[EnvFilter] old::Decorated Defs
{
  return if null(fils) 
	 then old
	 else filterDefs(head(fils), applyFilters(tail(fils), old));
}