grammar silver:base:core;

terminal NonTerminal_kwd / nonterminal /
 lexer precedence = 5 ;

terminal Closed_kwd / closed /
 lexer precedence = 5 ;

terminal Close_kwd / close /
 lexer precedence = 5 ;

abstract production defaultNonterminalDcl
top::AGDcl ::= id::Id_t
{
  top.location = "[line: " ++ toString(id.line) ++ ", column: " ++ toString(id.column) ++ "] ";
  top.pp = "nonterminal " ++ id.lexeme ++ " ;" ;

  production attribute fName :: String;
  fName = top.grammarName.name ++ ":" ++ id.lexeme;

  top.defs = addTypeDcl(fName, ntTypeRep(fName), 
	     addFullNameDcl(id.lexeme, fName,  emptyDefs()));

  top.globalDefs = addTypeDcl(fName, ntTypeRep(fName), emptyDefs()); 

  local attribute er1 :: String;
  er1 = if length(getFullNameDcl(id.lexeme, top.env)) > 1 
       then top.location ++ "Error: Name '" ++ id.lexeme ++ "' is already bound.\n"
       else "";	

  local attribute er2 :: String;
  er2 = if length(getTypeDcl(fName, top.env)) > 1 
       then top.location ++ "Error: Type '" ++ fName ++ "' is already bound.\n"
       else "";	

  top.errors = er1 ++ er2;
}

concrete production nonterminalDcl
top::AGDcl ::= n::NonTerminal_kwd  id::Id_t  s::Semi_t
{
  production attribute extraDcls :: AGDcl with agDclSeq;
  extraDcls := defaultNonterminalDcl(id);

  forwards to extraDcls'';
}

concrete production closedNonterminalDcl
top::AGDcl ::= c::Closed_kwd n::NonTerminal_kwd  id::Id_t  s::Semi_t
{  
  production attribute extraDcls :: AGDcl with agDclSeq;
  extraDcls := defaultClosedNonterminalDcl(id) ;
 
  forwards to extraDcls'';
}

abstract production defaultClosedNonterminalDcl
top::AGDcl ::= id::Id_t
{
  top.location = "[line: " ++ toString(id.line) ++ ", column: " ++ toString(id.column) ++ "] ";
  top.pp = "closed nonterminal " ++ id.lexeme ++ " ;" ;

  production attribute fName :: String;
  fName = top.grammarName.name ++ ":" ++ id.lexeme;

  top.defs = addTypeDcl(fName, ntTypeRep(fName),
	     addCloseDcl(fName,		
	     addFullNameDcl(id.lexeme, fName, emptyDefs())));

  top.globalDefs = addTypeDcl(fName, ntTypeRep(fName), addCloseDcl(fName, emptyDefs())); 

  local attribute er1 :: String;
  er1 = if length(getFullNameDcl(id.lexeme, top.env)) > 1 
       then top.location ++ "Error: Name '" ++ id.lexeme ++ "' is already bound.\n"
       else "";	

  local attribute er2 :: String;
  er2 = if length(getTypeDcl(fName, top.env)) > 1 
       then top.location ++ "Error: Type '" ++ fName ++ "' is already bound.\n"
       else "";	

  top.errors = er1 ++ er2;
}


concrete production closeNonterminalDcl
top::AGDcl ::= c::Close_kwd q::QName  s::Semi_t
{
  top.location = "[line: " ++ toString(c.line) ++ ", column: " ++ toString(c.column) ++ "] ";
  top.pp = "close " ++ q.name ++ " ;" ;

  production attribute fNames :: [Decorated EnvItem];
  fNames = getFullNameDclOne(q.name, top.env);

  production attribute fName ::String;
  fName = if !null(fNames) then head(fNames).fullName else "FULL_NAME_NOT_DEFINED";

  --whether it is bound to a value
  production attribute types :: [Decorated EnvItem];
  types = getTypeDcl(fName, top.env);

  local attribute er1 :: String;
  er1 = if null(types)
	then top.location ++ "Error: Type '" ++ q.name ++ "' is not declared.\n" 
	else "";

  top.errors = er1;

  top.defs = addCloseDcl(q.name,  emptyDefs());
  top.globalDefs = addCloseDcl(q.name,  emptyDefs());
}
