grammar silver:base:core;

autocopy attribute signatureEnv :: Decorated Env;
autocopy attribute localsEnv :: Decorated Env;

synthesized attribute prodDefs :: Decorated Defs;

synthesized attribute sigNames :: [String];
synthesized  attribute sigTypes :: [Decorated TypeRep];

attribute sigNames occurs on AspectSignature, AspectLHS, AspectLHSElem, AspectRHS, AspectRHSElem;
attribute sigNames occurs on ProductionSignature, ProductionLHS, ProductionLHSElem, ProductionRHS, ProductionRHSElem;

attribute sigTypes occurs on AspectSignature, AspectLHS, AspectLHSElem, AspectRHS, AspectRHSElem;
attribute sigTypes occurs on ProductionSignature, ProductionLHS, ProductionLHSElem, ProductionRHS, ProductionRHSElem;

autocopy attribute signatureNames :: [String];
attribute signatureNames occurs on ProductionBody, ProductionStmts, ProductionStmt, AttributeDef, LHSExpr, Expr, Exprs;

autocopy attribute signatureTypes :: [Decorated TypeRep];
attribute signatureTypes occurs on ProductionBody, ProductionStmts, ProductionStmt, AttributeDef, LHSExpr, Expr, Exprs;

terminal Abstract_kwd /abstract/ lexer precedence = 5;
terminal Production_kwd /production/ lexer precedence = 5;
concrete production productionDcl
 top::AGDcl ::=
   a::Abstract_kwd p::Production_kwd id::Id_t ns::ProductionSignature
   body::ProductionBody
{
  top.pp = "abstract production " ++ id.lexeme ++  "\n" ++ 
	     ns.pp  ++ "  \n{\n" ++ body.pp ++ "\n}";

  top.location = "[line: " ++ toString(a.line) ++ ", column: " ++ toString(a.column) ++ "] ";  

  production attribute fName :: String;
  fName = top.grammarName.name ++ ":" ++ id.lexeme;

  top.defs = addValueDcl(fName, prodTypeRep(tail(ns.sigTypes),  head(ns.sigTypes)), 
	     addProductionDcl(fName, head(ns.sigNames), tail(ns.sigNames), prodTypeRep(tail(ns.sigTypes), head(ns.sigTypes)), 
	     addFullNameDcl(id.lexeme, fName,
	     addProdAttsDcl(fName, body.prodDefs, emptyDefs()))));

  top.globalDefs = addValueDcl(fName, prodTypeRep(tail(ns.sigTypes),  head(ns.sigTypes)),
	           addProductionDcl(fName, head(ns.sigNames), tail(ns.sigNames), prodTypeRep(tail(ns.sigTypes), head(ns.sigTypes)), 
	           emptyDefs())); 
	     
  local attribute er1 :: String;
  er1 = if length(getFullNameDcl(id.lexeme, top.env)) > 1
       then top.location ++ "Error: Name '" ++ id.lexeme ++ "' is already bound.\n"
       else "";	

  local attribute er2 :: String;
  er2 = if length(getValueDcl(fName, top.env)) > 1
       then top.location ++ "Error: Value '" ++ fName ++ "' is already bound.\n"
       else "";	

  top.errors = er1 ++ er2 ++ ns.errors ++ body.errors;


  ns.env = appendDefsEnv(ns.defs, top.env);

  body.env = appendDefsEnv(appendDefs(body.defs, appendDefs(ns.defs, addThisDcl(fName, emptyDefs()))), top.env);

  body.signatureEnv = toEnv(ns.defs);
  body.localsEnv = toEnv(body.defs);

  body.signatureNames = ns.sigNames;
  body.signatureTypes = ns.sigTypes;
}
nonterminal ProductionSignature with {pp, location, grammarName, defs, env, errors};

concrete production productionSignatureEmptyRHS
top::ProductionSignature ::= lhs::ProductionLHS kwd::CCEQ_t
{
  top.pp = lhs.pp ++ " ::=";
  top.location = lhs.location;

  top.defs = lhs.defs;
  top.errors = lhs.errors;

  top.sigNames = lhs.sigNames;
  top.sigTypes = lhs.sigTypes;
}

concrete production productionSignature
top::ProductionSignature ::= lhs::ProductionLHS kwd::CCEQ_t rhs::ProductionRHS 
{
  top.pp = lhs.pp ++ " ::= " ++ rhs.pp;
  top.location = lhs.location;

  top.sigNames = lhs.sigNames +++ rhs.sigNames;
  top.sigTypes = lhs.sigTypes +++ rhs.sigTypes;

  top.defs = appendDefs(lhs.defs, rhs.defs);
  top.errors = lhs.errors ++ rhs.errors;

  rhs.cnt = 0;
}


nonterminal ProductionLHS with {pp, location, grammarName, defs, env, errors};
concrete production productionLHS
top::ProductionLHS ::= lhs::ProductionLHSElem
{
  top.pp = lhs.pp;
  top.location = lhs.location;
  top.defs = lhs.defs;
  top.errors = lhs.errors;

  top.sigNames = lhs.sigNames;
  top.sigTypes = lhs.sigTypes;
}

nonterminal ProductionLHSElem with {pp, location, grammarName, defs, env, errors};
concrete production productionLHSElem
top::ProductionLHSElem ::= id::Id_t kwd::HasType_t t::Type
{
  top.pp = id.lexeme ++ "::" ++ t.pp;
  top.location = "[line: " ++ toString(id.line) ++ ", column: " ++ toString(id.column) ++ "] ";  

  production attribute fName :: String;
  fName = id.lexeme;

  top.defs = addValueDcl(fName, t.typerep, 
	     addFullNameDcl(id.lexeme, fName,  emptyDefs()));

  local attribute er1 :: String;
  er1 = if length(getFullNameDcl(id.lexeme, top.env)) > 1
       then top.location ++ "Error: Name '" ++ id.lexeme ++ "' is already bound.\n"
       else "";	

  local attribute er2 :: String;
  er2 = if length(getValueDcl(fName, top.env)) > 1
       then top.location ++ "Error: Value '" ++ fName ++ "' is already bound.\n"
       else "";	

  top.errors = er1 ++ er2 ++ t.errors;

  top.sigNames = [id.lexeme];
  top.sigTypes = [t.typerep];
}

concrete production productionLHSElemType
top::ProductionLHSElem ::= t::Type
{
  forwards to productionLHSElem(terminal(Id_t, "top"), terminal(HasType_t, "::="), t'');
}

inherited attribute cnt :: Integer;
nonterminal ProductionRHS with {pp, location, grammarName, defs, env, errors, cnt};
abstract production productionRHSEmpty
top::ProductionRHS ::=
{
  top.pp = "";
  top.location = "[line: -1, column: -1, productionRHSEmpty()] ";
  top.defs = emptyDefs();
  top.errors = "";

  top.sigNames = [];
  top.sigTypes = [];
}

concrete production productionRHSSingle
top::ProductionRHS ::= rhs::ProductionRHSElem
{
  top.pp = rhs.pp;
  top.location = rhs.location;
  top.defs = rhs.defs;
  top.errors = rhs.errors;

  top.sigNames = rhs.sigNames;
  top.sigTypes = rhs.sigTypes;

  rhs.cnt = top.cnt;
}

concrete production productionRHSCons
top::ProductionRHS ::= h::ProductionRHSElem t::ProductionRHS
{
  top.pp = h.pp ++ " " ++ t.pp;
  top.location = h.location;
  top.defs = appendDefs(h.defs, t.defs);
  top.errors = h.errors ++ t.errors;

  top.sigNames = h.sigNames +++ t.sigNames;
  top.sigTypes = h.sigTypes +++ t.sigTypes;

  h.cnt = top.cnt;
  t.cnt = top.cnt + 1;
}

nonterminal ProductionRHSElem with {pp, location, grammarName, defs, env, errors, cnt};
concrete production productionRHSElem
top::ProductionRHSElem ::= id::Id_t kwd::HasType_t t::Type
{
  top.pp = id.lexeme ++ "::" ++ t.pp;

  production attribute fName :: String;
  fName = id.lexeme;

  top.defs = addValueDcl(fName, t.typerep, 
	     addFullNameDcl(id.lexeme, fName,  emptyDefs()));

  top.location = "[line: " ++ toString(id.line) ++ ", column: " ++ toString(id.column) ++ "] ";  

  top.sigNames = [id.lexeme];
  top.sigTypes = [t.typerep];

  local attribute er1 :: String;
  er1 = if length(getFullNameDcl(id.lexeme, top.env)) > 1
       then top.location ++ "Error: Name '" ++ id.lexeme ++ "' is already bound.\n"
       else "";	

  local attribute er2 :: String;
  er2 = if length(getValueDcl(fName, top.env)) > 1
       then top.location ++ "Error: Value '" ++ fName ++ "' is already bound.\n"
       else "";	

  top.errors = er1 ++ er2 ++ t.errors;
}


concrete production productionRHSElemType
top::ProductionRHSElem ::= t::Type
{
  forwards to productionRHSElem(terminal(Id_t, "rhs_" ++ toString(top.cnt)), terminal(HasType_t, "::="), t'');
}