grammar silver:base:core;
import core;

closed nonterminal TypeRep with {typeName, inputTypes, outputType, doDecorate,
                          isPrimative, isTerminal, isNonTerminal, isDecorated, isProduction, pp, typeEquals, applicationDispatcher, decoratedType, userFriendlyLHS};

synthesized attribute typeEquals :: Production (TypeEquals ::= Decorated TypeRep Decorated TypeRep);

synthesized attribute userFriendlyLHS :: Integer;

synthesized attribute typeName :: String;
synthesized attribute decoratedType :: Decorated TypeRep;
synthesized attribute inputTypes :: [Decorated TypeRep];
synthesized attribute outputType :: Decorated TypeRep;

synthesized attribute doDecorate :: Boolean;
synthesized attribute isPrimative :: Boolean;
synthesized attribute isTerminal :: Boolean;
synthesized attribute isNonTerminal :: Boolean;
synthesized attribute isDecorated :: Boolean;
synthesized attribute isProduction :: Boolean;


synthesized attribute applicationDispatcher :: Production (Expr ::= Expr Exprs);

function integerTypeRep
Decorated TypeRep ::= 
{
  return decorate internal_integerTypeRep() with {};
}
abstract production internal_integerTypeRep
top::TypeRep ::= 
{
  top.typeName = "Integer";
  top.pp = "Integer" ;
  top.isPrimative = true;
  forwards to internal_defaultTypeRep();
}

function floatTypeRep
Decorated TypeRep ::= 
{
  return decorate internal_floatTypeRep() with {};
}

abstract production internal_floatTypeRep
top::TypeRep ::= 
{
  top.typeName = "Float";
  top.pp = "Float" ;
  top.isPrimative = true;
  forwards to internal_defaultTypeRep();
}

function stringTypeRep
Decorated TypeRep ::= 
{
  return decorate internal_stringTypeRep() with {};
}
abstract production internal_stringTypeRep
top::TypeRep ::= 
{
  top.typeName = "String";
  top.pp = "String" ;
  top.isPrimative = true;
  forwards to internal_defaultTypeRep();
}

function booleanTypeRep
Decorated TypeRep ::= 
{
  return decorate internal_booleanTypeRep() with {};
}
abstract production internal_booleanTypeRep
top::TypeRep ::= 
{
  top.typeName = "Boolean";
  top.pp = "Boolean" ;
  top.isPrimative = true;
  forwards to internal_defaultTypeRep();
}

function termTypeRep
Decorated TypeRep ::= n::String
{
  return decorate internal_termTypeRep(n) with {};
}

abstract production internal_termTypeRep
top::TypeRep ::= n::String
{
  top.typeName = n;
  top.pp = "Terminal " ++ n ;
  top.isTerminal = true;
  forwards to internal_defaultTypeRep();
}

function ntTypeRep
Decorated TypeRep ::= n::String
{
  return decorate internal_ntTypeRep(n) with {};
}

abstract production internal_ntTypeRep
top::TypeRep ::= n::String
{
  top.typeName = n;
  top.pp = "NT " ++ n ;
  top.isNonTerminal = true;
  top.doDecorate = true;

  top.userFriendlyLHS = -1;

  forwards to internal_defaultTypeRep();
}

function refTypeRep
Decorated TypeRep ::= t::Decorated TypeRep
{
  return decorate internal_refTypeRep(t) with {};
}

abstract production internal_refTypeRep
top::TypeRep ::= t::Decorated TypeRep
{
  top.typeName = t.typeName;
  top.pp = "Decorated " ++ t.typeName ;
  top.decoratedType = t;
  top.isDecorated = true;

  top.userFriendlyLHS = 1;

  forwards to internal_defaultTypeRep();
}

function prodTypeRep
Decorated TypeRep ::= it::[Decorated TypeRep] ot::Decorated TypeRep
{
  return decorate internal_prodTypeRep(it,ot) with {};
}


abstract production internal_prodTypeRep
top::TypeRep ::= it::[Decorated TypeRep] ot::Decorated TypeRep
{
  top.typeName = "Production";
  top.pp = "Production( " ++ ot.pp ++ " ::= " ++ foldTypes(it) ++ ")"  ;
  top.isProduction = true;

  top.inputTypes = it;
  top.outputType = ot;
  forwards to internal_defaultTypeRep();
}

function foldTypes
String ::= l::[Decorated TypeRep]{
  return if null(l) then "" else head(l).pp ++ " " ++ foldTypes(tail(l));
}

--abstract production internal_prodTypeRep
--top::TypeRep ::= it::[Decorated TypeRep] ot::Decorated TypeRep
--{
--  top.typeName = "Production";
--  top.pp = "Prod "  ;
--  top.isProduction = true;
--
--  top.inputTypes = it;
--  top.outputType = ot;
--  forwards to internal_defaultTypeRep();
--}
--

function topTypeRep
Decorated TypeRep ::=
{
  return decorate internal_topTypeRep() with {};
}

abstract production internal_topTypeRep
top::TypeRep ::= 
{
  top.typeName = "TOP";
  top.pp = "TOP";
  top.isPrimative = true;
  top.isTerminal = true;
  top.isNonTerminal = true;
  top.isDecorated = true;
  top.isProduction = true;
  forwards to internal_defaultTypeRep();
}

function defaultTypeRep
Decorated TypeRep ::=
{
  return decorate internal_defaultTypeRep() with {};
}

abstract production internal_defaultTypeRep
top::TypeRep ::= 
{
  top.typeName = "DEFAULT";
  top.pp = "DEFAULT";
  top.isPrimative = false;
  top.isTerminal = false;
  top.isNonTerminal = false;
  top.isDecorated = false;
  top.isProduction = false;
  top.doDecorate = false;
  top.decoratedType = topTypeRep();
  top.typeEquals = genEquals;

  top.userFriendlyLHS = 0;

  top.inputTypes = [];
  top.outputType = topTypeRep();

  top.applicationDispatcher = genericApplicationDispatcher;
}

nonterminal TypeEquals with {bValue};
abstract production genEquals
top::TypeEquals ::= t1::Decorated TypeRep t2::Decorated TypeRep
{
    top.bValue = t1.typeName == "TOP" || t2.typeName == "TOP" ||
   	   ((t1.typeName == t2.typeName) &&  
	    ((t1.isPrimative && t2.isPrimative) ||
	     (t1.isTerminal && t2.isTerminal) ||
	     (t1.isNonTerminal && t2.isNonTerminal) ||
	     (t1.isDecorated && t2.isDecorated) ||
	     (t1.isProduction && t2.isProduction &&
	      t1.outputType.typeEquals(t1.outputType, t2.outputType).bValue &&
	      genListEquals(t1.inputTypes, t2.inputTypes))
	));
}

function genListEquals
Boolean ::= t1::[Decorated TypeRep] t2::[Decorated TypeRep]
{
  return if null(t1) && null(t2)
	 then true
	 else if null(t1) || null(t2)
	      then false
	      else head(t1).typeEquals(head(t1), head(t2)).bValue
		   && genListEquals(tail(t1), tail(t2));
}
