grammar silver:base:cs:default;
import silver:base:core;


attribute parserUses occurs on Expr, Exprs;

--Base Expressions
aspect production nestedExpr
top::Expr ::= l::LParen_t e::Expr r::RParen_t
{
  top.parserUses = e.parserUses;
}

aspect production productionReference
top::Expr ::= q::QName
{
  top.parserUses = if substring(length(fName) - 5, length(fName), fName) == "parse" then [fName] else [];
}

aspect production dontDecorateExpr
top::Expr ::= q::QName
{
  top.parserUses = [];
}

aspect production decorateExpr
top::Expr ::= q::QName
{
  top.parserUses = [];
}

aspect production genericApplicationDispatcher
top::Expr ::= e::Expr es::Exprs
{
  top.parserUses = e.parserUses ++ es.parserUses;
}

aspect production atAccess
top::Expr ::= e::Expr d::At_t q::QName
{
  top.parserUses = e.parserUses;
}

aspect production trueConst
top::Expr ::= t::True_kwd
{
  top.parserUses = [];
}

aspect production falseConst
top::Expr ::= f::False_kwd
{
  top.parserUses = [];
}

aspect production and
top::Expr ::= e1::Expr a::And_t e2::Expr
{
  top.parserUses = e1.parserUses ++ e2.parserUses;
}

aspect production or
top::Expr ::= e1::Expr o::Or_t e2::Expr
{
  top.parserUses = e1.parserUses ++ e2.parserUses;
}

aspect production not
top::Expr ::= n::Not_t e::Expr
{
  top.parserUses = e.parserUses;
}

aspect production gt
top::Expr ::= e1::Expr g::GT_t e2::Expr
{
  top.parserUses = e1.parserUses ++ e2.parserUses;
}

aspect production lt
top::Expr ::= e1::Expr l::LT_t e2::Expr
{
  top.parserUses = e1.parserUses ++ e2.parserUses;
}


aspect production gteq
top::Expr ::= e1::Expr g::GTEQ_t e2::Expr
{
  top.parserUses = e1.parserUses ++ e2.parserUses;
}


aspect production lteq
top::Expr ::= e1::Expr l::LTEQ_t e2::Expr
{
  top.parserUses = e1.parserUses ++ e2.parserUses;
}


aspect production eqeq
top::Expr ::= e1::Expr eq::EQEQ_t e2::Expr
{
  top.parserUses = e1.parserUses ++ e2.parserUses;
}


aspect production neq
top::Expr ::= e1::Expr n::NEQ_t e2::Expr
{
  top.parserUses = e1.parserUses ++ e2.parserUses;
}


aspect production ifThenElse
top::Expr ::= i::If_kwd e1::Expr t::Then_kwd e2::Expr e::Else_kwd e3::Expr
{
  top.parserUses = e1.parserUses ++ e2.parserUses ++ e3.parserUses;
}

aspect production intConst
top::Expr ::= i::Int_t
{
  top.parserUses = [];
}

aspect production floatConst
top::Expr ::= f::Float_t
{
  top.parserUses = [];
} 

aspect production plus
top::Expr ::= e1::Expr p::Plus_t e2::Expr
{
  top.parserUses = e1.parserUses ++ e2.parserUses;
}
aspect production minus
top::Expr ::= e1::Expr m::Minus_t e2::Expr
{
  top.parserUses = e1.parserUses ++ e2.parserUses;
}
aspect production multiply
top::Expr ::= e1::Expr m::Multiply_t e2::Expr
{
  top.parserUses = e1.parserUses ++ e2.parserUses;
}
aspect production divide
top::Expr ::= e1::Expr d::Divide_t e2::Expr
{
  top.parserUses = e1.parserUses ++ e2.parserUses;
}
aspect production neg
top::Expr ::= m::Minus_t e::Expr
{
  top.parserUses = e.parserUses;
}

aspect production stringConst
top::Expr ::= s::String_t
{
  top.parserUses = [];
}

aspect production defaultPlusPlus
top::Expr ::= e1::Expr e2::Expr
{
  top.parserUses = e1.parserUses ++ e2.parserUses;
}

aspect production stringPlusPlus
top::Expr ::= e1::Expr e2::Expr
{
  top.parserUses = e1.parserUses ++ e2.parserUses;
}

aspect production exprsEmpty
top::Exprs ::=
{
  top.parserUses = [];
}

aspect production exprsSingle
top::Exprs ::= e::Expr
{
  top.parserUses = e.parserUses;
}

aspect production exprsCons
top::Exprs ::= e1::Expr c::Comma_t e2::Exprs
{
  top.parserUses = e1.parserUses ++ e2.parserUses;
}

aspect production exprWith
top::Expr ::= ev::Decorate_kwd e::Expr w::With_kwd l::LCurly_t inh::ExprInh r::RCurly_t
{
  top.parserUses = e.parserUses ++ inh.parserUses;
}


attribute parserUses occurs on ExprInh;

aspect production exprStmtsEmpty
top::ExprInh ::= 
{
  top.parserUses = [];
}

aspect production exprStmtsCons
top::ExprInh ::= lhs::ExprLHSExpr ekwd::Equal_t e::Expr s::Semi_t inh::ExprInh
{ 
  top.parserUses = e.parserUses ++ inh.parserUses;
}