grammar silver:base:cs:default;
import silver:base:core;

attribute hasStartSymbol occurs on ImportExportStmts;
attribute startName occurs on ImportExportStmts;

synthesized attribute hiddenSyntax :: [String];
attribute hiddenSyntax occurs on SyntaxStmt;

attribute terminalDcls occurs on SyntaxStmt;
attribute ruleDcls occurs on SyntaxStmt;

attribute localDefs occurs on SyntaxStmt;

concrete production syntaxStmtsOne 
top::ImportExportStmts ::= syn::SyntaxStmt
{
  top.pp = syn.pp ;
  top.location = syn.location;
  top.importedEnv = emptyEnv();
  top.errors = syn.errors;
  top.includedGrammars = syn.includedGrammars;
  top.exportedDefs = emptyDefs();
  top.localDefs = syn.localDefs;

  top.hasStartSymbol = syn.hasStartSymbol;
  top.startName = syn.startName;

  top.terminalDcls = syn.terminalDcls;
  top.ruleDcls = syn.ruleDcls;
  top.hiddenSyntax = syn.hiddenSyntax;

  syn.compiledGrammars = top.compiledGrammars;
}

concrete production syntaxStmtsCons
top::ImportExportStmts ::= h::SyntaxStmt t::ImportExportStmts
{
  top.pp = h.pp ++ "\n" ++ t.pp ;
  top.location = h.location;
  top.importedEnv = t.importedEnv;
  top.errors = h.errors ++ t.errors;

  top.hasStartSymbol = h.hasStartSymbol || t.hasStartSymbol;
  top.startName = if h.hasStartSymbol then h.startName else t.startName;

  top.includedGrammars = h.includedGrammars +++ t.includedGrammars;
  top.exportedDefs = t.exportedDefs;
  top.localDefs = appendDefs(h.localDefs, t.localDefs);

  top.terminalDcls = h.terminalDcls +++ t.terminalDcls;
  top.ruleDcls = h.ruleDcls +++ t.ruleDcls;
  top.hiddenSyntax = h.hiddenSyntax +++ t.hiddenSyntax;

  h.compiledGrammars = top.compiledGrammars;
  t.compiledGrammars = top.compiledGrammars;
}

terminal Syntax_kwd  /syntax/ lexer precedence = 5 ;
nonterminal SyntaxStmt with {pp, grammarName, location, env, errors, includedGrammars, compiledGrammars, hasStartSymbol, startName};

concrete production syntaxAll
top::SyntaxStmt ::= s::Syntax_kwd pkg::QName se::Semi_t
{
  top.pp = "syntax " ++ pkg.pp ++ " ;" ;
  top.location = "[line: " ++ toString(s.line) ++ ", column: " ++ toString(s.column) ++ "] ";

  local attribute e :: String;
  e = if mitem.found then "" else top.location ++ "Error: Package '" ++ pkg.name ++ "' cannot be found.\n";

  top.errors = e;

  top.includedGrammars = [pkg.name];

  local attribute mitem :: MaybeRootSpec;
  mitem = getGrammar(pkg.name, top.compiledGrammars);

  production attribute spec :: Decorated RootSpec;
  spec = if mitem.found then mitem.rootSpec else decorate emptyRootSpec() with {};

  top.hasStartSymbol = spec.hasStartSymbol;
  top.startName = spec.startName;

  top.terminalDcls = spec.terminalDcls;
  top.ruleDcls = spec.ruleDcls;
  top.hiddenSyntax = spec.hiddenSyntax;

  top.localDefs = if   !spec.hasStartSymbol
 	     then emptyDefs()
	     else addValueDcl(top.grammarName.name ++ ":" ++ "parse", 
			      prodTypeRep([stringTypeRep()], ntTypeRep(spec.startName)),
                  addFullNameDcl("parse", top.grammarName.name ++ ":" ++ "parse", emptyDefs()));
}

concrete production syntaxHiding
top::SyntaxStmt ::= s::Syntax_kwd  pkg::QName h::Hiding_kwd ns::NameList se::Semi_t
{
  top.pp = "syntax " ++ pkg.pp ++ " ;" ;
  top.location = "[line: " ++ toString(s.line) ++ ", column: " ++ toString(s.column) ++ "] ";

  local attribute e :: String;
  e = if mitem.found then "" else top.location ++ "Error: Package '" ++ pkg.name ++ "' cannot be found.\n";

  top.errors = e;

  top.includedGrammars = [pkg.name];

  local attribute mitem :: MaybeRootSpec;
  mitem = getGrammar(pkg.name, top.compiledGrammars);

  production attribute spec :: Decorated RootSpec;
  spec = if mitem.found then mitem.rootSpec else decorate emptyRootSpec() with {};

  top.hasStartSymbol = spec.hasStartSymbol;
  top.startName = spec.startName;

  top.terminalDcls = spec.terminalDcls;
  top.ruleDcls = spec.ruleDcls;
  top.hiddenSyntax = ns.names +++ spec.hiddenSyntax;
--  top.hiddenSyntax = makeFullNames(pkg.name, ns.names) +++ spec.hiddenSyntax;

  top.localDefs = if   !spec.hasStartSymbol
 	     then emptyDefs()
	     else addValueDcl(top.grammarName.name ++ ":" ++ "parse", 
			      prodTypeRep([stringTypeRep()], ntTypeRep(spec.startName)),
                  addFullNameDcl("parse", top.grammarName.name ++ ":" ++ "parse", emptyDefs()));

}

function makeFullNames
[String] ::= pkg::String ns::[String]
{
  return if null(ns)
	 then []
	 else cons(pkg ++ ":" ++ head(ns), makeFullNames(pkg, tail(ns)));
}