grammar silver:base:regex:abstractsyntax;
export silver:base:regex:abstractsyntax;

synthesized attribute regex_pp :: String;
synthesized attribute copper_regexpr :: String;

nonterminal Regex_abst with {regex_pp, copper_regexpr};
nonterminal Range_abst with {regex_pp, copper_regexpr};

abstract production regex_concat
combined::Regex_abst ::= first::Regex_abst rest::Regex_abst
{
     combined.regex_pp = first.regex_pp ++ rest.regex_pp;
     combined.copper_regexpr = first.copper_regexpr ++ rest.copper_regexpr;
}

abstract production regex_kleene
combined::Regex_abst ::= first::Regex_abst rest::Regex_abst
{
     combined.regex_pp = "(" ++ first.regex_pp ++ ")*" ++ rest.regex_pp;
     combined.copper_regexpr = first.copper_regexpr ++ "*" ++ rest.copper_regexpr;
}

abstract production regex_plus
combined::Regex_abst ::= first::Regex_abst rest::Regex_abst
{
     combined.regex_pp = "(" ++ first.regex_pp ++ ")+" ++ rest.regex_pp;
     combined.copper_regexpr = first.copper_regexpr ++ "+" ++ rest.copper_regexpr;
}

abstract production regex_optional
combined::Regex_abst ::= first::Regex_abst rest::Regex_abst
{
     combined.regex_pp = "(" ++ first.regex_pp ++ ")?" ++ rest.regex_pp;
     combined.copper_regexpr = first.copper_regexpr ++ "?" ++ rest.copper_regexpr;
}

abstract production regex_choice
combined::Regex_abst ::= first::Regex_abst rest::Regex_abst
{
     combined.regex_pp = first.regex_pp ++ " | " ++ rest.regex_pp;
     combined.copper_regexpr = first.copper_regexpr ++ "|" ++ rest.copper_regexpr;
}

abstract production regex_empty
emp::Regex_abst ::=
{
     emp.regex_pp = "";
     emp.copper_regexpr = "";
}

abstract production regex_wildcard
wildcard::Regex_abst ::=
{
     wildcard.regex_pp = ".";
     wildcard.copper_regexpr = ".";
}

abstract production regex_parenthesized
parenth::Regex_abst ::= nonparenth::Regex_abst
{
     parenth.regex_pp = "(" ++ nonparenth.regex_pp ++ ")";
     parenth.copper_regexpr = "(" ++ nonparenth.copper_regexpr ++ ")";
}

abstract production regex_posCharrange
rangeRegex::Regex_abst ::= range::Range_abst
{
     rangeRegex.regex_pp = "[" ++ range.regex_pp ++ "]";
     rangeRegex.copper_regexpr = "[" ++ range.copper_regexpr ++ "]";
}

abstract production regex_negCharrange
rangeRegex::Regex_abst ::= range::Range_abst
{
     rangeRegex.regex_pp = "[^" ++ range.regex_pp ++ "]";
     rangeRegex.copper_regexpr = "[^" ++ range.copper_regexpr ++ "]";
}

abstract production range_rangeunion
range::Range_abst ::= first::Range_abst rest::Range_abst
{
     range.regex_pp = first.regex_pp ++ rest.regex_pp;
     range.copper_regexpr = first.copper_regexpr ++ rest.copper_regexpr;
}

abstract production range_loosechar
range::Range_abst ::= char::Regex_abst
{
     range.regex_pp = char.regex_pp;
     range.copper_regexpr = char.copper_regexpr;
}

abstract production range_rangechar
range::Range_abst ::= leastchar::Regex_abst greatestchar::Regex_abst
{
     range.regex_pp = leastchar.regex_pp ++ "-" ++ greatestchar.regex_pp;
     range.copper_regexpr = leastchar.copper_regexpr ++ "-" ++ greatestchar.copper_regexpr;
}

abstract production range_empty
emp::Range_abst ::=
{
     emp.regex_pp = "";
     emp.copper_regexpr = "";
}

abstract production regex_char
regex::Regex_abst ::= char::String
{
     local attribute re :: String;
     re = if char == "+" ||
             char == "*" ||
             char == "?" ||
             char == "|" ||
             char == "-" ||
             char == "^" ||
             char == "[" ||
             char == "]" ||
             char == "(" ||
             char == ")" ||
             char == "{" ||
             char == "}" ||
             char == "." ||
             char == ":" ||
             char == "\\"
          then "\\" ++ char
          else char;

     regex.regex_pp = re;
     regex.copper_regexpr = re;
}

abstract production regex_escaped_char
regex::Regex_abst ::= char::String
{
     local attribute re :: String;
     re = if char == "\\+" ||
             char == "\\*" ||
             char == "\\?" ||
             char == "\\|" ||
             char == "\\-" ||
             char == "\\^" ||
             char == "\\[" ||
             char == "\\]" ||
             char == "\\(" ||
             char == "\\)" ||
             char == "\\{" ||
             char == "\\}" ||
             char == "\\." ||
             char == "\\:" ||
             char == "\\r" ||
             char == "\\n" ||
             char == "\\t" ||
             char == "\\\\"
          then char
          else substring(1,length(char),char);

     regex.regex_pp = re;
     regex.copper_regexpr = re;
}
