grammar silver:base:regex:abstractsyntax;

-- import silver:base:regex:concretesyntax;

synthesized attribute lex_regexpr :: String;

attribute lex_regexpr occurs on Regex_abst;
attribute lex_regexpr occurs on Range_abst;

aspect production regex_concat
combined::Regex_abst ::= first::Regex_abst rest::Regex_abst
{
     combined.lex_regexpr = first.lex_regexpr ++ rest.lex_regexpr;
}

aspect production regex_kleene
combined::Regex_abst ::= first::Regex_abst rest::Regex_abst
{
     combined.lex_regexpr = first.lex_regexpr ++ "*" ++ rest.lex_regexpr;
}

aspect production regex_plus
combined::Regex_abst ::= first::Regex_abst rest::Regex_abst
{
     combined.lex_regexpr = first.lex_regexpr ++ "+" ++ rest.lex_regexpr;
}

aspect production regex_optional
combined::Regex_abst ::= first::Regex_abst rest::Regex_abst
{
     combined.lex_regexpr = first.lex_regexpr ++ "?" ++ rest.lex_regexpr;
}

aspect production regex_choice
combined::Regex_abst ::= first::Regex_abst rest::Regex_abst
{
     combined.lex_regexpr = first.lex_regexpr ++ "|" ++ rest.lex_regexpr;
}

aspect production regex_empty
emp::Regex_abst ::=
{
     emp.lex_regexpr = "";
}

aspect production regex_wildcard
wildcard::Regex_abst ::=
{
     wildcard.lex_regexpr = ".";
}

aspect production regex_parenthesized
parenth::Regex_abst ::= nonparenth::Regex_abst
{
     parenth.lex_regexpr = "(" ++ nonparenth.lex_regexpr ++ ")";
}

aspect production regex_posCharrange
rangeRegex::Regex_abst ::= range::Range_abst
{
     rangeRegex.lex_regexpr = "[" ++ range.lex_regexpr ++ "]";
}

aspect production regex_negCharrange
rangeRegex::Regex_abst ::= range::Range_abst
{
     rangeRegex.lex_regexpr = "[^" ++ range.lex_regexpr ++ "]";
}

aspect production range_rangeunion
range::Range_abst ::= first::Range_abst rest::Range_abst
{
     range.lex_regexpr = first.lex_regexpr ++ rest.lex_regexpr;
}

aspect production range_loosechar
range::Range_abst ::= char::Regex_abst
{
     range.lex_regexpr = char.lex_regexpr;
}

aspect production range_rangechar
range::Range_abst ::= leastchar::Regex_abst greatestchar::Regex_abst
{
     range.lex_regexpr = leastchar.lex_regexpr ++ "-" ++ greatestchar.lex_regexpr;
}

aspect production range_empty
emp::Range_abst ::=
{
     emp.lex_regexpr = "";
}

aspect production regex_char
regex::Regex_abst ::= char::String
{
     local attribute re_lex :: String;
     re_lex = if char == "+" ||
                 char == "*" ||
                 char == "?" ||
                 char == "|" ||
                 char == "-" ||
                 char == "^" ||
                 char == "[" ||
                 char == "]" ||
                 char == "(" ||
                 char == ")" ||
                 char == "{" ||
                 char == "}" ||
                 char == "." ||
                 char == "\\"
              then "\\" ++ char
              else char;

     regex.lex_regexpr = re_lex;
}

aspect production regex_escaped_char
regex::Regex_abst ::= char::String
{
     local attribute re_lex :: String;
     re_lex = if char == "\\+" ||
                 char == "\\*" ||
                 char == "\\?" ||
                 char == "\\|" ||
                 char == "\\-" ||
                 char == "\\^" ||
                 char == "\\[" ||
                 char == "\\]" ||
                 char == "\\(" ||
                 char == "\\)" ||
                 char == "\\{" ||
                 char == "\\}" ||
                 char == "\\." ||
                 char == "\\r" ||
                 char == "\\n" ||
                 char == "\\t" ||
                 char == "\\\\"
              then char
              else substring(1,length(char),char);

     regex.lex_regexpr = re_lex;
}
