grammar silver:extension:easyterminal;
import silver:base;

synthesized attribute easyRegsDefs :: [EasyReg];
attribute easyRegsDefs occurs on AGDcl, AGDcls, Root, ImportExportStmts, ImportStmt, RootSpec, Roots;

synthesized attribute exportedEasyRegsDefs :: [EasyReg];
attribute exportedEasyRegsDefs occurs on Root, RootSpec, ImportExportStmts, ExportStmt;

autocopy attribute easyRegsEnv :: [EasyReg];
attribute easyRegsEnv occurs on AGDcl, Root, AGDcls, Roots, ImportExportStmts, ImportStmt, ExportStmt, ProductionSignature, ProductionRHS, ProductionRHSElem;
attribute easyRegsEnv occurs on AspectSignature, AspectRHS, AspectRHSElem;
attribute easyRegsEnv occurs on ProductionBody, ProductionStmts, ProductionStmt, AttributeDef, Expr, Exprs, ExprInh, ForwardInh;

aspect production compileGrammar
top::Grammar ::= iIn::IO gn::String sPath::[String]
{
  cu.easyRegsEnv = cu.easyRegsDefs;
}
aspect production compileFiles
top::Roots ::= iIn::IO files::[String] gpath::String
{
  top.easyRegsDefs = if null(files) then [] else r.easyRegsDefs +++ recurse.easyRegsDefs;

  r.easyRegsEnv = top.easyRegsEnv;
  recurse.easyRegsEnv = top.easyRegsEnv;
}
aspect production emptyRootSpec
top::RootSpec ::= 
{
  top.easyRegsDefs = [];
  top.exportedEasyRegsDefs = [];
}
aspect production consRootSpec
top::RootSpec ::= c1::Decorated Root c2::Decorated RootSpec
{
  top.easyRegsDefs = c1.easyRegsDefs +++ c2.easyRegsDefs;
  top.exportedEasyRegsDefs = c1.exportedEasyRegsDefs +++ c2.exportedEasyRegsDefs;
}
aspect production appendRootSpec
top::RootSpec ::= c1::Decorated RootSpec c2::Decorated RootSpec
{
  top.easyRegsDefs = c1.easyRegsDefs +++ c2.easyRegsDefs;
  top.exportedEasyRegsDefs = c1.exportedEasyRegsDefs +++ c2.exportedEasyRegsDefs;
}
aspect production root
top::Root ::= gdcl::GrammarDcl ims::ImportExportStmts  ags::AGDcls
{
  top.easyRegsDefs = ags.easyRegsDefs;
  top.exportedEasyRegsDefs = ims.exportedEasyRegsDefs;

  ags.easyRegsEnv = top.easyRegsEnv +++ ims.easyRegsDefs;
}
aspect production agDclsOne
top::AGDcls ::= ag::AGDcl
{
  top.easyRegsDefs = ag.easyRegsDefs;
}

aspect production agDclsCons
top::AGDcls ::= h::AGDcl t::AGDcls
{
  top.easyRegsDefs = h.easyRegsDefs +++ t.easyRegsDefs;
}

aspect production agDclSeq
top::AGDcl ::= ag1::AGDcl ag2::AGDcl
{
  top.easyRegsDefs = ag1.easyRegsDefs +++ ag2.easyRegsDefs;
}

aspect production agDclNone
top::AGDcl ::=
{
  top.easyRegsDefs = [];
}


aspect production attributeDclInh
top::AGDcl ::= i::Inherited_kwd  at::Attribute_kwd a::Id_t ht::HasType_t te::Type s::Semi_t
{
  top.easyRegsDefs = [];
}
aspect production attributeDclSyn
top::AGDcl ::= sn::Synthesized_kwd  at::Attribute_kwd a::Id_t ht::HasType_t te::Type  s::Semi_t
{
  top.easyRegsDefs = [];
}
aspect production defaultNonterminalDcl
top::AGDcl ::= id::Id_t
{
  top.easyRegsDefs = [];
}
aspect production defaultClosedNonterminalDcl
top::AGDcl ::= id::Id_t
{
  top.easyRegsDefs = [];
}
aspect production closeNonterminalDcl
top::AGDcl ::= c::Close_kwd q::QName  s::Semi_t
{
  top.easyRegsDefs = [];
}
aspect production defaultStartNonterminalDcl
top::AGDcl ::= id::Id_t
{
  top.easyRegsDefs = [];
}
aspect production attributionDcl
top::AGDcl ::= at::Attribute_kwd a::QName oc::Occurs_kwd o::On_kwd nt::QName s::Semi_t
{
  top.easyRegsDefs = [];
}
aspect production productionDcl
 top::AGDcl ::=
   a::Abstract_kwd p::Production_kwd id::Id_t ns::ProductionSignature
   body::ProductionBody
{
  top.easyRegsDefs = [];
}
aspect production aspectAGDcl
 top::AGDcl ::=
   a::Aspect_kwd p::Production_kwd id::QName ns::AspectSignature
   body::ProductionBody
{
  top.easyRegsDefs = [];
}

aspect production importStmtsNone
top::ImportExportStmts ::=
{
  top.easyRegsDefs = [];
  top.exportedEasyRegsDefs = []; 
}

aspect production importStmtsOne 
top::ImportExportStmts ::= im::ImportStmt
{
  top.easyRegsDefs = im.easyRegsDefs;
  top.exportedEasyRegsDefs = [];
}

aspect production exportStmtsOne 
top::ImportExportStmts ::= ex::ExportStmt
{
  top.easyRegsDefs = [];
  top.exportedEasyRegsDefs = ex.exportedEasyRegsDefs;
}
aspect production syntaxStmtsOne 
top::ImportExportStmts ::= syn::SyntaxStmt
{
  top.easyRegsDefs = [];
  top.exportedEasyRegsDefs = []; 
}

aspect production importStmtsCons
top::ImportExportStmts ::= h::ImportStmt t::ImportExportStmts
{
  top.easyRegsDefs = h.easyRegsDefs +++ t.easyRegsDefs;
  top.exportedEasyRegsDefs = t.exportedEasyRegsDefs;
}
aspect production exportStmtsCons
top::ImportExportStmts ::= h::ExportStmt t::ImportExportStmts
{
  top.easyRegsDefs = t.easyRegsDefs;
  top.exportedEasyRegsDefs = h.exportedEasyRegsDefs +++ t.exportedEasyRegsDefs;
}
aspect production syntaxStmtsCons
top::ImportExportStmts ::= h::SyntaxStmt t::ImportExportStmts
{
  top.easyRegsDefs = t.easyRegsDefs;
  top.exportedEasyRegsDefs = t.exportedEasyRegsDefs;
}

aspect production appendImportsExports
top::ImportExportStmts ::= fst::ImportExportStmts snd::ImportExportStmts
{
  top.easyRegsDefs = fst.easyRegsDefs +++ snd.easyRegsDefs;
  top.exportedEasyRegsDefs = fst.exportedEasyRegsDefs +++ snd.exportedEasyRegsDefs;
}


--TODO filter this.
aspect production importAll
top::ImportStmt ::= i::Import_kwd  pkg::QName  wc::OptWithClause  s::Semi_t
{
  top.easyRegsDefs = mitem.rootSpec.exportedEasyRegsDefs;
}
aspect production importOnly
top::ImportStmt ::= i::Import_kwd pkg::QName o::Only_kwd ns::NameList wc::OptWithClause  s::Semi_t 
{
  top.easyRegsDefs = mitem.rootSpec.exportedEasyRegsDefs;
}
aspect production importHiding
top::ImportStmt ::= i::Import_kwd pkg::QName h::Hiding_kwd ns::NameList wc::OptWithClause  s::Semi_t 
{
  top.easyRegsDefs = mitem.rootSpec.exportedEasyRegsDefs;
}
aspect production importAs
top::ImportStmt ::= i::Import_kwd pkg1::QName a::As_kwd pkg2::QName s::Semi_t
{
  top.easyRegsDefs = mitem.rootSpec.exportedEasyRegsDefs;
}

aspect production exportAll
top::ExportStmt ::= i::Export_kwd  pkg::QName  wc::OptWithClause  s::Semi_t
{
  top.exportedEasyRegsDefs = 
	  if pkg.name == top.grammarName.name 
	  then top.easyRegsEnv 
	  else if mitem.found 
	       then mitem.rootSpec.exportedEasyRegsDefs
	       else [];
}
aspect production exportOnly
top::ExportStmt ::= i::Export_kwd pkg::QName o::Only_kwd ns::NameList wc::OptWithClause  s::Semi_t 
{
  top.exportedEasyRegsDefs = 
	  if pkg.name == top.grammarName.name 
	  then top.easyRegsEnv 
	  else if mitem.found 
	       then mitem.rootSpec.exportedEasyRegsDefs
	       else [];
}
aspect production exportHiding
top::ExportStmt ::= i::Export_kwd pkg::QName h::Hiding_kwd ns::NameList wc::OptWithClause  s::Semi_t 
{
  top.exportedEasyRegsDefs = 
	  if pkg.name == top.grammarName.name 
	  then top.easyRegsEnv 
	  else if mitem.found 
	       then mitem.rootSpec.exportedEasyRegsDefs
	       else [];
}
aspect production exportAs
top::ExportStmt ::= i::Export_kwd pkg1::QName a::As_kwd pkg2::QName s::Semi_t
{
  top.exportedEasyRegsDefs = 
	  if pkg1.name == top.grammarName.name 
	  then top.easyRegsEnv 
	  else if mitem.found 
	       then mitem.rootSpec.exportedEasyRegsDefs
	       else [];
}

function keepER
[EasyReg] ::= sl::[String] er::[EasyReg]
{
  return if null(sl) then []
	 else keepERHelp(head(sl), er) +++ keepER(tail(sl), er);
}

function keepERHelp
[EasyReg] ::= s::String er::[EasyReg]
{
  return if null(er) then []
	 else if head(er).name == s then [head(er)]
	      else keepERHelp(s, tail(er));
}

function throwER
[EasyReg] ::= sl::[String] er::[EasyReg]
{
  return if null(sl) then er
	 else throwERHelp(head(sl), er) +++ throwER(tail(sl), er);
}

function throwERHelp
[EasyReg] ::= s::String er::[EasyReg]
{
  return if null(er) then []
	 else if head(er).name == s then tail(er)
	      else [head(er)] +++ throwERHelp(s, tail(er));
}

function prependER
[EasyReg] ::= s::String er::[EasyReg]
{
  return if null(er) then []
	 else [mkEasyReg(head(er).easyReg, head(er).name, head(er).typerep)] +++ prependER(s, tail(er));
}

