grammar silver:extension:easyterminal;
import silver:base;

terminal Terminal_t /[\']([^\']|([\\][\']))*[\']/ lexer precedence = 5;

synthesized attribute isEasyReg ::Boolean;
attribute isEasyReg occurs on RegExpr;


aspect production terminalDcl
top::AGDcl ::= t::Terminal_kwd id::Id_t r::RegExpr s::Semi_t
{
  top.easyRegsDefs = if r.isEasyReg then [mkEasyReg(r.pp, id.lexeme, termTypeRep(fName))] else [];
}
aspect production regExpr
top::RegExpr ::= r::RegExpr_t
{
  top.isEasyReg = false;
}
concrete production easyRegExpr
top::RegExpr ::= r::Terminal_t
{
  top.pp = r.lexeme;
  top.location = "[line: " ++ toString(r.line) ++ ", column: " ++ toString(r.column) ++ "] ";
  top.isEasyReg = true;

  local attribute newreg :: String;
  newreg = "/" ++ makereg(substring(1, length(r.lexeme)-1, r.lexeme)) ++ "/";

  forwards to regExpr(terminal(RegExpr_t, newreg));
}

concrete production productionRhsElemEasyReg
top::ProductionRHSElem ::= reg::Terminal_t
{
  local attribute regName :: [Decorated TypeRep];
  regName = getEasyReg(reg.lexeme, top.easyRegsEnv);

  local attribute fNames :: [Decorated EnvItem];
  fNames = getAllFullNameDcls(top.env);

  local attribute aNames :: [Decorated EnvItem];
  aNames = if null(regName) then [] else filterEnvItems(reverseLookUp(head(regName).typeName), fNames);

  local attribute aName :: String;
  aName = if null(aNames) then "NOT FOUND"
	  else head(aNames).itemName;

  top.errors = if null(regName) 
	       then top.location ++ "Error: Could not find terminal declaration for " ++ reg.lexeme ++ "\n"
 	       else "";

  forwards to productionRHSElem(terminal(Id_t, "NOT_NEEDED" ++ toString(genint()), reg.line, reg.column),
			        terminal(HasType_t, "::="), nttType(makeQName(aName)));
}

concrete production aspectRHSElemEasyReg
top::AspectRHSElem ::= reg::Terminal_t
{
  local attribute regName :: [Decorated TypeRep];
  regName = getEasyReg(reg.lexeme, top.easyRegsEnv);

  local attribute fNames :: [Decorated EnvItem];
  fNames = getAllFullNameDcls(top.env);

  local attribute aNames :: [Decorated EnvItem];
  aNames = if null(regName) then [] else filterEnvItems(reverseLookUp(head(regName).typeName), fNames);

  local attribute aName :: String;
  aName = if null(aNames) then "NOT FOUND"
	  else head(aNames).itemName;

  top.errors = if null(regName) 
	       then top.location ++ "Error: Could not find terminal declaration for " ++ reg.lexeme ++ "\n"
 	       else "";

  forwards to aspectRHSElemFull(terminal(Id_t, "NOT_NEEDED" ++ toString(genint()), reg.line, reg.column),
			        terminal(HasType_t, "::="), nttType(makeQName(aName)));
}

concrete production terminalExpr
top::Expr ::= t::Terminal_t
{
  local attribute regName :: [Decorated TypeRep];
  regName = getEasyReg(t.lexeme, top.easyRegsEnv);


  local attribute fNames :: [Decorated EnvItem];
  fNames = getAllFullNameDcls(top.env);

  local attribute aNames :: [Decorated EnvItem];
  aNames = if null(regName) 
	   then [] 
	   else filterEnvItems(reverseLookUp(head(regName).typeName), fNames);

  local attribute aName :: String;
  aName = if null(aNames) 
	  then "NOT_FOUND"
	  else head(aNames).itemName;

  local attribute escapedName :: String;
  escapedName = makeEscapedName(substring(1, length(t.lexeme)-1, t.lexeme));

  top.errors = if null(regName) 
	       then top.location ++ "Error: Could not find terminal declaration for " ++ t.lexeme ++ "\n"
 	       else "";

  forwards to terminalFunction(terminal(Terminal_kwd, "terminal", t.line, t.column),
			       terminal(LParen_t, "("),
			       nttType(makeQName(aName)),
			       terminal(Comma_t, ","),
			       stringConst(terminal(String_t, "\"" ++ escapedName ++ "\"")),
			       terminal(RParen_t, ")"));
}

function makeEscapedName
String ::= s::String
{
  local attribute ch :: String;
  ch = substring(0, 1, s);

  local attribute rest :: String;
  rest = substring(1, length(s), s);

  return if length(s) == 0 
	 then ""
	 else if ch == "\\" || ch == "\""
	      then "\\" ++ ch ++ makeEscapedName(rest)
 	      else ch ++ makeEscapedName(rest);
}

function makereg 
String ::= s::String
{
  local attribute ch :: String;
  ch = substring(0, 1, s);

  local attribute rest :: String;
  rest = substring(1, length(s), s);

  return if length(s) == 0 
	 then ""
	 else if isAlpha(ch) || isDigit(ch)
	      then ch ++ makereg(rest)
	      else "[\\" ++ ch ++ "]" ++ makereg(rest);
}

nonterminal EasyReg with {easyReg, typerep, name};
synthesized attribute easyReg :: String;

abstract production mkEasyReg
top::EasyReg ::= s::String n::String t::Decorated TypeRep
{
  top.easyReg = s;
  top.name = n;
  top.typerep = t;
}


function getEasyReg
[Decorated TypeRep] ::= s::String l::[EasyReg]
{
  return if null(l) then []
	 else if head(l).easyReg == s 
	      then [head(l).typerep]
	      else getEasyReg(s, tail(l));
}