grammar silver:extension:list;
import silver:base;
import core;


synthesized attribute isUnknownListType :: Boolean;
attribute isUnknownListType occurs on Expr;

autocopy attribute resolvedType :: Decorated TypeRep;
attribute resolvedType occurs on Expr, Exprs, ExprInh;

aspect production forwardsDcl
top::ProductionStmt ::= f::Forwards_kwd t::To_kwd e::Expr s::Semi_t
{
  e.resolvedType = emptyListTypeRep();
}

aspect production forwardsDclWith
top::ProductionStmt ::= f::Forwards_kwd t::To_kwd e::Expr w::With_kwd l::LCurly_t inh::ForwardInh r::RCurly_t s::Semi_t
{
  e.resolvedType = emptyListTypeRep();
}

aspect production forwardStmtsCons
top::ForwardInh ::= lhs::ForwardLHSExpr ekwd::Equal_t e::Expr s::Semi_t inh::ForwardInh
{ 
  e.resolvedType = emptyListTypeRep();
}

aspect production concreteAttrDef
top::AttributeDef ::= lhs::LHSExpr eq::Equal_t e::Expr s::Semi_t
{
  e.resolvedType = emptyListTypeRep();
}

aspect production attrDef
top::AttributeDef ::= lhs::LHSExpr e::Expr
{
  e.resolvedType = emptyListTypeRep();
}

--Base Expressions
aspect production nestedExpr
top::Expr ::= l::LParen_t e::Expr r::RParen_t
{
  top.isUnknownListType = false;
}

aspect production productionReference
top::Expr ::= q::QName
{
  top.isUnknownListType = false;
}

aspect production dontDecorateExpr
top::Expr ::= q::QName
{
  top.isUnknownListType = false;
}

aspect production decorateExpr
top::Expr ::= q::QName
{
  top.isUnknownListType = false;
}

aspect production genericApplicationDispatcher
top::Expr ::= e::Expr es::Exprs
{
  top.isUnknownListType = false;
}

aspect production atAccess
top::Expr ::= e::Expr d::At_t q::QName
{
  top.isUnknownListType = false;
}

aspect production trueConst
top::Expr ::= t::True_kwd
{
  top.isUnknownListType = false;
}

aspect production falseConst
top::Expr ::= f::False_kwd
{
  top.isUnknownListType = false;
}

aspect production and
top::Expr ::= e1::Expr a::And_t e2::Expr
{
  top.isUnknownListType = false;
}

aspect production or
top::Expr ::= e1::Expr o::Or_t e2::Expr
{
  top.isUnknownListType = false;
}

aspect production not
top::Expr ::= n::Not_t e::Expr
{
  top.isUnknownListType = false;
}

aspect production gt
top::Expr ::= e1::Expr g::GT_t e2::Expr
{
  top.isUnknownListType = false;
}

aspect production lt
top::Expr ::= e1::Expr l::LT_t e2::Expr
{
  top.isUnknownListType = false;
}

aspect production gteq
top::Expr ::= e1::Expr g::GTEQ_t e2::Expr
{
  top.isUnknownListType = false;
}

aspect production lteq
top::Expr ::= e1::Expr l::LTEQ_t e2::Expr
{
  top.isUnknownListType = false;
}

aspect production eqeq
top::Expr ::= e1::Expr eq::EQEQ_t e2::Expr
{
  top.isUnknownListType = false;
}

aspect production neq
top::Expr ::= e1::Expr n::NEQ_t e2::Expr
{
  top.isUnknownListType = false;
}

aspect production ifThenElse
top::Expr ::= i::If_kwd e1::Expr t::Then_kwd e2::Expr e::Else_kwd e3::Expr
{
  top.isUnknownListType = false;

  e2.resolvedType = if e2.isUnknownListType && !e3.isUnknownListType && e3.typerep.isList
		    then e3.typerep
	            else emptyListTypeRep();

  e3.resolvedType = if e3.isUnknownListType && !e2.isUnknownListType && e2.typerep.isList
		    then e2.typerep
	            else emptyListTypeRep();
}

aspect production intConst
top::Expr ::= i::Int_t
{
  top.isUnknownListType = false;
}

aspect production floatConst
top::Expr ::= f::Float_t
{
  top.isUnknownListType = false;
} 

aspect production plus
top::Expr ::= e1::Expr p::Plus_t e2::Expr
{
  top.isUnknownListType = false;
}
aspect production minus
top::Expr ::= e1::Expr m::Minus_t e2::Expr
{
  top.isUnknownListType = false;
}
aspect production multiply
top::Expr ::= e1::Expr m::Multiply_t e2::Expr
{
  top.isUnknownListType = false;
}
aspect production divide
top::Expr ::= e1::Expr d::Divide_t e2::Expr
{
  top.isUnknownListType = false;
}
aspect production neg
top::Expr ::= m::Minus_t e::Expr
{
  top.isUnknownListType = false;
}

aspect production stringConst
top::Expr ::= s::String_t
{
  top.isUnknownListType = false;
}

aspect production stringPlusPlus
top::Expr ::= e1::Expr e2::Expr
{
  top.isUnknownListType = false;
}

aspect production defaultPlusPlus
top::Expr ::= e1::Expr e2::Expr
{
  top.isUnknownListType = false;
}

aspect production exprsEmpty
top::Exprs ::=
{

}

aspect production exprsSingle
top::Exprs ::= e::Expr
{

}

aspect production exprsCons
top::Exprs ::= e1::Expr c::Comma_t e2::Exprs
{

}

aspect production exprWith
top::Expr ::= ev::Decorate_kwd e::Expr w::With_kwd l::LCurly_t inh::ExprInh r::RCurly_t
{
  top.isUnknownListType = false;
}

aspect production exprStmtsEmpty
top::ExprInh ::= 
{

}

aspect production exprStmtsCons
top::ExprInh ::= lhs::ExprLHSExpr ekwd::Equal_t e::Expr s::Semi_t inh::ExprInh
{ 

}

aspect production exprLhsExpr
top::ExprLHSExpr ::= q::QName
{

}

aspect production lengthFunction
top::Expr ::= l::Length_kwd lp::LParen_t e::Expr rp::RParen_t
{
  top.isUnknownListType = false;
}

aspect production indexOfFunction
top::Expr ::= i::IndexOf_kwd l::LParen_t e1::Expr c::Comma_t e2::Expr r::RParen_t
{
  top.isUnknownListType = false;
}

aspect production subStringFunction
top::Expr ::= s::SubString_kwd l::LParen_t e1::Expr c1::Comma_t e2::Expr c2::Comma_t e3::Expr r::RParen_t
{
  top.isUnknownListType = false;
}
aspect production errorFunction
top::Expr ::= er::Error_kwd l::LParen_t e::Expr r::RParen_t
{
  top.isUnknownListType = false;
}
aspect production stringToIntFunction
top::Expr ::= t::ToInt_kwd l::LParen_t e::Expr r::RParen_t
{
  top.isUnknownListType = false;
}
aspect production stringToFloatFunction
top::Expr ::= t::ToFloat_kwd l::LParen_t e::Expr r::RParen_t
{
  top.isUnknownListType = false;
}
aspect production numToStringFunction
top::Expr ::= t::ToString_kwd l::LParen_t e::Expr r::RParen_t
{
  top.isUnknownListType = false;
}

aspect production ftoiFunction
top::Expr ::= t::FTOI_kwd l::LParen_t e::Expr r::RParen_t
{
  top.isUnknownListType = false;
}

aspect production itofFunction
top::Expr ::= t::ITOF_kwd l::LParen_t e::Expr r::RParen_t
{
  top.isUnknownListType = false;
}

aspect production isDigitFunction
top::Expr ::= t::IsDigit_kwd l::LParen_t e::Expr r::RParen_t
{
  top.isUnknownListType = false;
}

aspect production isAlphaFunction
top::Expr ::= t::IsAlpha_kwd l::LParen_t e::Expr r::RParen_t
{
  top.isUnknownListType = false;
}

aspect production isSpaceFunction
top::Expr ::= t::IsSpace_kwd l::LParen_t e::Expr r::RParen_t
{
  top.isUnknownListType = false;
}

aspect production isLowerFunction
top::Expr ::= t::IsLower_kwd l::LParen_t e::Expr r::RParen_t
{
  top.isUnknownListType = false;
}

aspect production isUpperFunction
top::Expr ::= t::IsUpper_kwd l::LParen_t e::Expr r::RParen_t
{
  top.isUnknownListType = false;
}

aspect production newFunction
top::Expr ::= n::New_kwd l::LParen_t e::Expr r::RParen_t
{
  top.isUnknownListType = false;
}

aspect production terminalFunction
top::Expr ::= term::Terminal_kwd l::LParen_t t::Type c::Comma_t e::Expr r::RParen_t
{
  top.isUnknownListType = false;
}

aspect production terminalFunction2
top::Expr ::= term::Terminal_kwd l::LParen_t t::Type c1::Comma_t e1::Expr c2::Comma_t e2::Expr c3::Comma_t e3::Expr r::RParen_t
{
  top.isUnknownListType = false;
}

aspect production isFileFunction
top::Expr ::= i::IsFile_kwd l::LParen_t e1::Expr c::Comma_t e2::Expr r::RParen_t
{
  top.isUnknownListType = false;
}

aspect production isDirectoryFunction
top::Expr ::= i::IsDirectory_kwd l::LParen_t e1::Expr c::Comma_t e2::Expr r::RParen_t
{
  top.isUnknownListType = false;
}

aspect production readFunction
top::Expr ::= r::ReadFile_kwd l::LParen_t e1::Expr c::Comma_t e2::Expr rp::RParen_t
{
  top.isUnknownListType = false;
}

aspect production cwdFunction
top::Expr ::= c::CWD_kwd l::LParen_t e::Expr r::RParen_t
{
  top.isUnknownListType = false;
}

aspect production envVarFunction
top::Expr ::= e::EnvVar_kwd l::LParen_t e1::Expr c::Comma_t e2::Expr r::RParen_t
{
  top.isUnknownListType = false;
}

aspect production systemFunction
top::Expr ::= s::System_kwd l::LParen_t e1::Expr c::Comma_t e2::Expr r::RParen_t
{
  top.isUnknownListType = false;
}

aspect production writeFunction
top::Expr ::= w::WriteFile_kwd l::LParen_t e1::Expr c1::Comma_t e2::Expr c2::Comma_t e3::Expr r::RParen_t
{  
  top.isUnknownListType = false;
}

aspect production appendFunction
top::Expr ::= a::AppendFile_kwd l::LParen_t e1::Expr c1::Comma_t e2::Expr c2::Comma_t e3::Expr r::RParen_t
{
  top.isUnknownListType = false;
}

aspect production printFunction
top::Expr ::= p::Print_kwd l::LParen_t e1::Expr c::Comma_t e2::Expr r::RParen_t
{
  top.isUnknownListType = false;
}

aspect production listContentsFunction
top::Expr ::= li::ListContents_kwd l::LParen_t e1::Expr c::Comma_t e2::Expr r::RParen_t
{
  top.isUnknownListType = false;
}

aspect production unsafeIOFunction
top::Expr ::= u::UnsafeIO_kwd
{  
  top.isUnknownListType = false;
}

aspect production genIntFunction
top::Expr ::= g::GenInt_kwd l::LParen_t r::RParen_t
{
   top.isUnknownListType = false;
}

aspect production emptyList
top::Expr ::= lb::LSqr_t rb::RSqr_t
{
  top.isUnknownListType = true;
}

aspect production emptyListWType
top::Expr ::= lb::LSqr_t ht::HasType_t t::Type rb::RSqr_t
{
  top.isUnknownListType = false;
}

aspect production fullList
top::Expr ::= lb::LSqr_t es::Exprs rb::RSqr_t
{ 
  top.isUnknownListType = false;
}

aspect production consList
top::Expr ::= ck::Cons_t lp::LParen_t h::Expr c::Comma_t t::Expr rp::RParen_t 
{ 
  top.isUnknownListType = false;
}

aspect production consListOp
top::Expr ::= h::Expr o::ConsOp_t t::Expr
{ 
  top.isUnknownListType = false;
}

aspect production appendList
top::Expr ::= l::Expr o::AppendOp_t r::Expr
{ 
  top.isUnknownListType = l.isUnknownListType && r.isUnknownListType;
}

aspect production nullList
top::Expr ::= n::Null_t lp::LParen_t l::Expr rp::RParen_t 
{ 
  top.isUnknownListType = false;
}

aspect production headList
top::Expr ::= h::Head_t lp::LParen_t l::Expr rp::RParen_t 
{ 
  top.isUnknownListType = false;
}

aspect production tailList
top::Expr ::= t::Tail_t lp::LParen_t l::Expr rp::RParen_t 
{ 
  top.isUnknownListType = false;
}
