grammar silver:translation:ag:haskell:core;
import silver:base;
import silver:utils;

attribute haskell occurs on Expr, Exprs;

--Base Expressions
aspect production nestedExpr
top::Expr ::= l::LParen_t e::Expr r::RParen_t
{
  top.haskell = "(" ++ e.haskell ++ ")";
}

aspect production productionReference
top::Expr ::= q::QName
{
  top.haskell = "production_" ++ makeName(fName);
}

aspect production dontDecorateExpr
top::Expr ::= q::QName
{
  top.haskell = "param_" ++  makeName(fName);
}

aspect production decorateExpr
top::Expr ::= q::QName
{
  top.haskell = "param_" ++  makeName(fName) ++ "_node";
}

aspect production genericApplicationDispatcher
top::Expr ::= e::Expr es::Exprs
{
  top.haskell = "(" ++ e.haskell ++ " " ++ es.haskell ++ ")";
}

aspect production atAccess
top::Expr ::= e::Expr d::At_t q::QName
{
  local attribute htn :: String;
  htn = if e.typerep.isTerminal then "TERMINAL" else "NT_" ++ makeName(e.typerep.typeName);

  local attribute attr_str :: String;
  attr_str = "(attr_" ++  htn ++ "__"  ++ makeName(fName);

  top.haskell = attr_str ++ (if isSynthesized(fName, top.env) 
			     then " (syns_"  
 			     else " (inhs_")
			 ++ htn ++ " " ++ e.haskell ++ "))";
}

aspect production trueConst
top::Expr ::= t::True_kwd
{
  top.haskell = "True";
}

aspect production falseConst
top::Expr ::= f::False_kwd
{
  top.haskell = "False";
}

aspect production and
top::Expr ::= e1::Expr a::And_t e2::Expr
{
  top.haskell = "(" ++ e1.haskell ++ " && " ++ e2.haskell ++ ")";
}

aspect production or
top::Expr ::= e1::Expr o::Or_t e2::Expr
{
  top.haskell = "(" ++ e1.haskell ++ " || " ++ e2.haskell ++ ")";
}

aspect production not
top::Expr ::= n::Not_t e::Expr
{
  top.haskell = "(not " ++ e.haskell ++ ")";
}

aspect production gt
top::Expr ::= e1::Expr g::GT_t e2::Expr
{
  top.haskell = "(" ++ e1.haskell ++ " > " ++ e2.haskell ++ ")";
}

aspect production lt
top::Expr ::= e1::Expr l::LT_t e2::Expr
{
  top.haskell = "(" ++ e1.haskell ++ " < " ++ e2.haskell ++ ")";
}


aspect production gteq
top::Expr ::= e1::Expr g::GTEQ_t e2::Expr
{
  top.haskell = "(" ++ e1.haskell ++ " >= " ++ e2.haskell ++ ")";
}


aspect production lteq
top::Expr ::= e1::Expr l::LTEQ_t e2::Expr
{
  top.haskell = "(" ++ e1.haskell ++ " <= " ++ e2.haskell ++ ")";
}


aspect production eqeq
top::Expr ::= e1::Expr eq::EQEQ_t e2::Expr
{
  top.haskell = "(" ++ e1.haskell ++ " == " ++ e2.haskell ++ ")";
}


aspect production neq
top::Expr ::= e1::Expr n::NEQ_t e2::Expr
{
  top.haskell = "(" ++ e1.haskell ++ " /= " ++ e2.haskell ++ ")";
}


aspect production ifThenElse
top::Expr ::= i::If_kwd e1::Expr t::Then_kwd e2::Expr e::Else_kwd e3::Expr
{
  top.haskell = "(if " ++ e1.haskell ++ " then " ++ e2.haskell ++ " else " ++ e3.haskell ++ ")";
}

aspect production intConst
top::Expr ::= i::Int_t
{
  top.haskell = i.lexeme;
}

aspect production floatConst
top::Expr ::= f::Float_t
{
  top.haskell = f.lexeme;
} 

aspect production plus
top::Expr ::= e1::Expr p::Plus_t e2::Expr
{
  top.haskell = "(" ++ e1.haskell ++ " + " ++ e2.haskell ++ ")";
}
aspect production minus
top::Expr ::= e1::Expr m::Minus_t e2::Expr
{
  top.haskell = "(" ++ e1.haskell ++ " - " ++ e2.haskell ++ ")";
}
aspect production multiply
top::Expr ::= e1::Expr m::Multiply_t e2::Expr
{
  top.haskell = "(" ++ e1.haskell ++ " * " ++ e2.haskell ++ ")";
}
aspect production divide
top::Expr ::= e1::Expr d::Divide_t e2::Expr
{
  top.haskell = "(" ++ e1.haskell ++ " / " ++ e2.haskell ++ ")";
}
aspect production neg
top::Expr ::= m::Minus_t e::Expr
{
  top.haskell = "(" ++ " - " ++ e.haskell ++ ")";
}

aspect production stringConst
top::Expr ::= s::String_t
{
  top.haskell = "(" ++ s.lexeme ++ ")";
}

aspect production defaultPlusPlus
top::Expr ::= e1::Expr e2::Expr
{
  top.haskell = "(" ++ e1.haskell ++ " ++ " ++ e2.haskell ++ ")";
}

aspect production stringPlusPlus
top::Expr ::= e1::Expr e2::Expr
{
  top.haskell = "(" ++ e1.haskell ++ " ++ " ++ e2.haskell ++ ")";
}

aspect production exprsEmpty
top::Exprs ::=
{
  top.haskell = "";
}

aspect production exprsSingle
top::Exprs ::= e::Expr
{
  top.haskell = e.haskell;
}

aspect production exprsCons
top::Exprs ::= e1::Expr c::Comma_t e2::Exprs
{
  top.haskell = e1.haskell ++ " " ++ e2.haskell;
}

autocopy attribute typeHaskell :: String;
aspect production exprWith
top::Expr ::= ev::Decorate_kwd e::Expr w::With_kwd l::LCurly_t inh::ExprInh r::RCurly_t
{
  top.haskell = "( " ++ e.haskell ++ " INHS_" ++ "NT_" ++ makeName(e.typerep.typeName) ++ " { " ++ inh.haskell ++ " } )";

  inh.typeHaskell = e.typerep.typeName;
  inh.definedInhs = [];
  inh.loc = top.location;
}


attribute haskell occurs on ExprInh;
attribute typeHaskell occurs on ExprInh;

inherited attribute definedInhs :: [String];
attribute definedInhs occurs on ExprInh;


inherited attribute loc :: String;
attribute loc occurs on ExprInh;

aspect production exprStmtsEmpty
top::ExprInh ::= 
{
  local attribute all :: [Decorated EnvItem];
  all = getAllOccuringOn(top.typeHaskell, top.globalEnv);

  local attribute ns :: [String];
  ns = toNames(all);

  local attribute needed :: [String];
  needed = removeDefined(ns, top.definedInhs);

  local attribute inhs ::[String];
  inhs = removeSyns(needed, top.globalEnv);

  top.haskell = makeDefs(top.loc, top.typeHaskell, inhs);
}

aspect production exprStmtsCons
top::ExprInh ::= lhs::ExprLHSExpr ekwd::Equal_t e::Expr s::Semi_t inh::ExprInh
{ 
  top.haskell =  lhs.haskell ++ " = " ++ e.haskell ++ 
		 if inh.haskell == "" 
	         then "" 
	         else (", " ++ inh.haskell);

  inh.definedInhs = top.definedInhs +++ [lhs.fullName];
  inh.loc = top.loc;
}

attribute haskell occurs on ExprLHSExpr;
attribute typeHaskell occurs on ExprLHSExpr;
attribute fullName occurs on ExprLHSExpr;

aspect production exprLhsExpr
top::ExprLHSExpr ::= q::QName
{
  top.haskell = "attr_" ++ "NT_" ++ makeName(top.typeHaskell) ++ "__" ++ makeName(fName);
  top.fullName = fName;
}

function toNames
[String] ::= is::[Decorated EnvItem]
{
  return if null(is) then [] else [head(is).itemName] +++ toNames(tail(is));
}

function removeDefined
[String] ::= all::[String] def::[String]
{
  return if null(all) 
	 then [] 
	 else if containsString(head(all), def)
	 then removeDefined(tail(all), def)
	 else [head(all)] +++ removeDefined(tail(all), def);
}

function removeSyns
[String] ::= all::[String] e::Decorated Env
{
  return if null(all) 
	 then [] 
	 else if isInherited(head(all), e) 
	 then [head(all)] +++ removeSyns(tail(all), e) 
	 else removeSyns(tail(all), e);
}
function makeDefs
String ::= l::String th::String def::[String]
{
  local attribute e :: String;
  e = "(error \"" ++ l ++ " Runtime Error: Inherited attribute" ++ head(def) ++ "is not defined.\")";

  return if null(def)
	 then ""
	 else ("attr_" ++ "NT_" ++ makeName(th) ++ "__" ++ makeName(head(def)) ++ " = " ++ e ++ 
	       if null(tail(def))
	       then "" 
	       else (", " ++ makeDefs(l, th, tail(def))));
}
