grammar silver:translation:ag:haskell:core;
import silver:base;
import silver:utils;

aspect production defaultNonterminalDcl
top::AGDcl ::= id::Id_t
{
  top.haskell = nonTerminalHaskell(id.lexeme, top.grammarName, top.globalEnv);
}

aspect production defaultClosedNonterminalDcl
top::AGDcl ::= id::Id_t
{
  top.haskell = nonTerminalHaskell(id.lexeme, top.grammarName, top.globalEnv);
}

aspect production closeNonterminalDcl
top::AGDcl ::= c::Close_kwd q::QName  s::Semi_t
{
  top.haskell = "";
}

aspect production defaultStartNonterminalDcl
top::AGDcl ::= id::Id_t
{
  top.haskell = nonTerminalHaskell(id.lexeme, top.grammarName, top.globalEnv);
}

function nonTerminalHaskell
String ::= n::String pn::Decorated QName e::Decorated Env
{
  local attribute hn :: String;
  hn = "NT_" ++ makeName(pn.name) ++ "_" ++ n;

  local attribute fqn :: String;
  fqn = pn.name ++ ":" ++ n;

  local attribute decorating :: [Decorated EnvItem];
  decorating = getAllOccuringOn(fqn, e);

  local attribute syns :: [Decorated EnvItem];
  syns = filterEnvItems(keepSyns(e), decorating);

  local attribute inhs :: [Decorated EnvItem];
  inhs = filterEnvItems(keepInhs(e), decorating);
 
  local attribute string_syns :: String;
  string_syns = getAttrStr("attr_" ++ hn, ",\n  ", syns);

  local attribute string_inhs :: String;
  string_inhs = getAttrStr("attr_" ++ hn, ",\n  ", inhs);
  
  local attribute node_def :: String;
  node_def = "data " ++ hn ++ " = " ++ hn ++ 
      " {\n new_" ++ hn ++ " :: INHS_" ++ hn ++ " -> " ++ hn ++ ",\n" ++
           " syns_" ++ hn ++ " :: SYNS_" ++ hn ++ ",\n" ++
           " inhs_" ++ hn ++ " :: INHS_" ++ hn ++ "\n}\n";

  local attribute syns_def :: String;
  syns_def = "data SYNS_" ++ hn ++ " = SYNS_" ++ hn ++ " {\n  " ++ string_syns ++ "\n}\n";

  local attribute inhs_def :: String;
  inhs_def = "data INHS_" ++ hn ++ " = INHS_" ++ hn ++ " {\n  " ++ string_inhs ++ "\n}\n";

  return node_def ++ syns_def ++ inhs_def;
}

abstract production keepInhs
top::EnvFilter ::= e::Decorated Env
{
  top.keep = isInherited(top.inEnvItem.itemName, e);
}

abstract production keepSyns
top::EnvFilter ::= e::Decorated Env
{
  top.keep = isSynthesized(top.inEnvItem.itemName, e);
}

function getAttrStr
String ::= pref::String sep::String atts::[Decorated EnvItem]
{  
  local attribute fn :: String;
  fn = makeName(head(atts).itemName);

  return if null(atts)
	 then ""
	 else pref ++ "__" ++ fn ++ " :: " ++ head(atts).typerep.haskellType
 	      ++ (if length(tail(atts)) == 0 then "" else sep )
	      ++ getAttrStr(pref, sep, tail(atts));
}

