grammar silver:translation:haskell:core;
import silver:base;
import silver:utils;
         
aspect production productionDcl
 top::AGDcl ::=
   a::Abstract_kwd p::Production_kwd id::Id_t ns::ProductionSignature
   body::ProductionBody
{

  local attribute all_stmts :: [StmtRep];
  all_stmts = body.stmtList +++ getAspects(top.grammarName.name ++ ":" ++ id.lexeme, top.globalEnv);

  local attribute lhs_hn :: String;
  lhs_hn = "NT_" ++ makeName(head(ns.sigTypes).typeName);

  local attribute sigEnv :: Decorated Env;
  sigEnv = toEnv(ns.defs);

  local attribute onfns :: [Decorated EnvItem];
  onfns = getFullNameDclOne(head(ns.sigNames), sigEnv);

  local attribute onfn :: String;
  onfn = if !null(onfns) then head(onfns).fullName 
	 else "FULL_NAME_NOT_FOUND";

  local attribute lhs_syns :: [StmtRep];
  lhs_syns = getDefsForName(onfn, all_stmts);

  local attribute decorating :: [Decorated EnvItem];
  decorating = getAllOccuringOn(head(ns.sigTypes).typeName, top.globalEnv);

  local attribute all_syns :: [Decorated EnvItem];
  all_syns = filterEnvItems(keepSyns(top.globalEnv), decorating);

  local attribute lhs_syns_str :: String;
  lhs_syns_str = makeLHSDefStr("SYNS_" ++ lhs_hn, id.lexeme, head(ns.sigNames), lhs_hn, lhs_syns, all_syns, !null(f_dcls));

  local attribute l_dcls :: [StmtRep];
  l_dcls = getLocalDeclarations(all_stmts);

  local attribute f_dcls :: [StmtRep];
  f_dcls = getForwardDeclarations(all_stmts);

  local attribute fing_dcls :: [StmtRep];
  fing_dcls = getForwardingDeclarations(all_stmts);

  top.haskell =
                "production_" ++ makeName(top.grammarName.name) ++ "_" ++ id.lexeme ++ " " ++ ns.haskell ++ "\n"
             ++ " where\n"
	     ++ "  " ++ ns.lhsHaskell ++ " = ( production_" ++ makeName(top.grammarName.name) ++ "_" ++ id.lexeme ++ " " ++ ns.rhsHaskell ++ " )\n"
	     ++ "  " ++ ns.lhsHaskell ++ "_node = " ++ lhs_hn ++ " {\n"
	     ++ "   new_"  ++ lhs_hn ++ " = " ++ ns.lhsHaskell ++ ",\n"
	     ++ "   syns_" ++ lhs_hn ++ " = syns_" ++ ns.lhsHaskell ++ "_node,\n"
             ++ "   inhs_" ++ lhs_hn ++ " = inhs\n"
	     ++ "  }\n"
	     ++ "  syns_" ++ ns.lhsHaskell ++ "_node = " ++ lhs_syns_str ++ "\n\n"
	     ++ makeChildrenDefs(id.lexeme, tail(ns.sigNames), tail(ns.sigTypes), all_stmts, sigEnv, top.globalEnv)
	     ++ makeLocalDefs(id.lexeme, l_dcls, all_stmts, sigEnv, top.globalEnv)
	     ++ makeForwardDefs(f_dcls, fing_dcls);
}

aspect production aspectAGDcl
 top::AGDcl ::=
   a::Aspect_kwd p::Production_kwd id::QName ns::AspectSignature
   body::ProductionBody
{
  top.haskell = "";
}

attribute haskell occurs on ProductionSignature;
attribute haskell occurs on ProductionRHS;
attribute haskell occurs on ProductionRHSElem;
attribute haskell occurs on ProductionLHS;
attribute haskell occurs on ProductionLHSElem;

function makeLocalDefs
String ::= pname::String dcls::[StmtRep] stmts::[StmtRep] sig::Decorated Env e::Decorated Env
{
  local attribute t_hn :: String;
  t_hn = "NT_" ++ makeName(head(dcls).typerep.typeName);

  local attribute l_def :: [StmtRep];
  l_def = getLocalDefinition(head(dcls).fullName, stmts);

  local attribute l_def_str :: String;
  l_def_str = if null(l_def)
	      then hname ++ " = error \"In production '" ++ pname ++ "': Local attribute '" 
			 ++ head(dcls).name ++ "' is declared but not not defined.\"\n" 
	      else hname ++ " = " ++ head(l_def).rhsExpr.haskell ++ "\n";

  local attribute hname :: String;
  hname = "param_" ++ makeName(head(dcls).fullName);

  local attribute inhs :: [StmtRep];
  inhs = getDefsForName(head(dcls).fullName, stmts);

  local attribute decorating :: [Decorated EnvItem];
  decorating = getAllOccuringOn(head(dcls).typerep.typeName, e);

  local attribute all_inhs :: [Decorated EnvItem];
  all_inhs = filterEnvItems(keepInhs(e), decorating);
 
  return
    if null(dcls) then ""
    else (if !(head(dcls).typerep.isNonTerminal)
          then "  " ++ l_def_str
          else "  " ++ l_def_str
	    ++ "  inhs_" ++ hname ++ " = " ++ makeDefStr("INHS_" ++ t_hn, pname, head(dcls).name, t_hn, inhs, all_inhs) ++ "\n"
	    ++ "  " ++ hname ++ "_node = " ++ hname ++ " inhs_" ++ hname ++ "\n\n")
	  ++ makeLocalDefs(pname, tail(dcls), stmts, sig, e);
}

function makeForwardDefs
String ::= dcls::[StmtRep] fings::[StmtRep]
{
  local attribute t_hn :: String;
  t_hn = "NT_" ++ makeName(head(dcls).typerep.typeName);

  return
    if null(dcls) then "" 
    else if !(head(dcls).typerep.isNonTerminal)
         then "  param__FORWARD__node = error \"Forward node is not a nonterminal.\"\n" 
         else "  param__FORWARD = " ++ head(dcls).rhsExpr.haskell ++ "\n"
	   ++ "  " ++ makeForwardInhStr(t_hn, mergeForwardingDcls(head(dcls).stmtList, fings))
           ++ "  param__FORWARD__node = param__FORWARD inhs__param__FORWARD\n" ;
}

function mergeForwardingDcls
[StmtRep] ::= stmts::[StmtRep] fings::[StmtRep]
{
  return if null(fings)
         then stmts
         else mergeForwardingDcls(stmts +++ head(fings).stmtList, tail(fings));
}

function makeForwardInhStr
String ::= nt_hn::String stmts::[StmtRep]
{
  return if null(stmts)
	 then "inhs__param__FORWARD = inhs\n" 
         else "inhs__param__FORWARD = inhs {\n   " ++ 
	       makeForwardInhStrHelp(nt_hn, stmts) ++ "\n  }\n";
}

function makeForwardInhStrHelp
String ::= nt_hn::String stmts::[StmtRep]
{
  local attribute attr_name :: String;
  attr_name = "attr_" ++ nt_hn ++ "__" ++ makeName(head(stmts).lhsExpr.secondAccessFull);

  return
   if null(stmts) then ""
   else if head(stmts).isAttributeDef && head(stmts).lhsExpr.accessCount == 1
	then attr_name ++ " = " ++ head(stmts).rhsExpr.haskell
   	     ++ (if null(tail(stmts)) then "" else ",\n   ")
	     ++ makeForwardInhStrHelp(nt_hn, tail(stmts))
	else makeForwardInhStrHelp(nt_hn, tail(stmts));
}

function makeChildrenDefs
String ::= pname::String childrenN::[String] childrenT::[Decorated TypeRep] stmts::[StmtRep] sig::Decorated Env e::Decorated Env
{
  local attribute t_hn :: String;
  t_hn = "NT_" ++ makeName(head(childrenT).typeName);

  local attribute hn :: String;
  hn = "param_" ++ head(childrenN);

  local attribute fns :: [Decorated EnvItem];
  fns = if true
        then getFullNameDclOne(head(childrenN), sig)  -- from haskell/core/ProductionDcl.sv
        else getFullNameDcl(head(childrenN), sig);    -- from haskellc/core/ProductionDcl.sv
                                                      -- Why was there any difference?
  local attribute fn :: String;
  fn = if !null(fns) then head(fns).fullName else "FULL_NAME_NOT_FOUND";

  local attribute inhs :: [StmtRep];
  inhs = getDefsForName(fn, stmts);

  local attribute decorating :: [Decorated EnvItem];
  decorating = getAllOccuringOn(head(childrenT).typeName, e);

  local attribute all_inhs :: [Decorated EnvItem];
  all_inhs = filterEnvItems(keepInhs(e), decorating);
 
  return
    if null(childrenT) then ""
    else (if !(head(childrenT).isNonTerminal)
          then ""
          else "  inhs_" ++ hn ++ " = " ++ makeDefStr("INHS_" ++ t_hn, pname, head(childrenN), t_hn, inhs, all_inhs) ++ "\n"
	    ++ "  " ++ hn ++ "_node = " ++ hn ++ " inhs_" ++ hn ++ "\n\n")
  	    ++ makeChildrenDefs(pname, tail(childrenN), tail(childrenT), stmts, sig, e);
}

--takes in an additional paramater to signify whether we are forwarding or not
function makeLHSDefStr
String ::= prefix::String pname::String nname::String nt_hn::String stmts::[StmtRep] all_decorating::[Decorated EnvItem] fing::Boolean
{
  local attribute my_defs :: String;
  my_defs = makeDefStrHelp(nt_hn, stmts);

  local attribute xdefs :: String;
  xdefs = if fing 
	  then makeExtraForwardingDefs(pname, nname, nt_hn, stmts, all_decorating)
	  else makeExtraDefs(pname, nname, nt_hn, stmts, all_decorating);

  return
   prefix ++ " {\n   " ++ my_defs ++ (if my_defs == "" || xdefs == "" then "" else ",\n   ") ++ xdefs ++ "\n  }";	       
}

function makeExtraForwardingDefs
String ::= pname::String nname::String nt_hn::String defined::[StmtRep] all::[Decorated EnvItem]
{
  local attribute attr_name :: String;
  attr_name = "attr_" ++ nt_hn ++ "__" ++ makeName(head(all).itemName);

  local attribute recurse :: String;
  recurse = makeExtraForwardingDefs(pname, nname, nt_hn, defined, tail(all));

  local attribute f_def :: String;
  f_def = attr_name ++ " = (" ++ attr_name ++ " (syns_" ++ nt_hn ++ " param__FORWARD__node))";

  return
	if null(all) then ""
        else if containsAttrDef(head(all).itemName, defined) 
	     then recurse                 
             else (f_def ++ (if recurse == ""  then "" else ",\n   ") ++ recurse);
}

--for id.id
function makeDefStr
String ::= prefix::String pname::String nname::String nt_hn::String stmts::[StmtRep] all_decorating::[Decorated EnvItem]
{
  local attribute my_defs :: String;
  my_defs = makeDefStrHelp(nt_hn, stmts);

  local attribute xdefs :: String;
  xdefs = makeExtraDefs(pname, nname, nt_hn, stmts, all_decorating);

  return
   prefix ++ " {\n   " ++ my_defs ++ (if my_defs == "" || xdefs == "" then "" else ",\n   ") ++ xdefs ++ "\n  }";	       
}

function makeDefStrHelp
String ::= nt_hn::String stmts::[StmtRep]
{
  local attribute attr_name :: String;
  attr_name = "attr_" ++ nt_hn ++ "__" ++ makeName(head(stmts).lhsExpr.secondAccessFull);

  return
   if null(stmts) then ""
   else attr_name ++ " = " ++ head(stmts).rhsExpr.haskell
	++ (if null(tail(stmts)) then "" else ",\n   ")
	++ makeDefStrHelp(nt_hn, tail(stmts));   
}

function makeExtraDefs
String ::= pname::String nname::String nt_hn::String defined::[StmtRep] all::[Decorated EnvItem]
{
  local attribute attr_name :: String;
  attr_name = "attr_" ++ nt_hn ++ "__" ++ makeName(head(all).itemName);

  local attribute recurse :: String;
  recurse = makeExtraDefs(pname, nname, nt_hn, defined, tail(all));

  local attribute error_msg :: String;
  error_msg = attr_name ++ " = error \"Error: Attribute '" ++ nname ++ "." ++ head(all).itemName ++ 
                              "' is not defined in production '" ++ pname ++ "'\"";

  return
	if null(all) then ""
        else if containsAttrDef(head(all).itemName, defined'') 
	     then recurse                 
             else (error_msg ++ (if recurse == ""  then "" else ",\n   ") ++ recurse);
}

function containsAttrDef
Boolean ::= attr::String stmts::[StmtRep]
{
  local attribute fname :: String;
  fname = head(stmts).lhsExpr.secondAccessFull;

  return
	if null(stmts) then false
	else if head(stmts).isAttributeDef && fname == attr then true
	else containsAttrDef (attr, tail(stmts));
}

synthesized attribute lhsHaskell :: String;
attribute lhsHaskell occurs on ProductionSignature;

synthesized attribute rhsHaskell :: String;
attribute rhsHaskell occurs on ProductionSignature;

aspect production productionSignatureEmptyRHS
top::ProductionSignature ::= lhs::ProductionLHS kwd::CCEQ_t
{
  top.rhsHaskell = "";
  top.lhsHaskell = lhs.haskell;

  top.haskell = "inhs = " ++ lhs.haskell ++ "_node";
}

aspect production productionSignature
top::ProductionSignature ::= lhs::ProductionLHS kwd::CCEQ_t rhs::ProductionRHS 
{
  top.rhsHaskell = rhs.haskell;
  top.lhsHaskell = lhs.haskell;

  top.haskell = rhs.haskell ++ " inhs = " ++ lhs.haskell ++ "_node";
}

aspect production productionLHS
top::ProductionLHS ::= lhs::ProductionLHSElem
{
  top.haskell = lhs.haskell;
}

aspect production productionLHSElem
top::ProductionLHSElem ::= id::Id_t kwd::HasType_t t::Type
{
  top.haskell = "param_" ++ id.lexeme;
}

aspect production productionRHSEmpty
top::ProductionRHS ::= 
{
  top.haskell = "";
}

aspect production productionRHSSingle
top::ProductionRHS ::= rhs::ProductionRHSElem
{
  top.haskell = rhs.haskell;
}

aspect production productionRHSCons
top::ProductionRHS ::= h::ProductionRHSElem t::ProductionRHS
{
  top.haskell = h.haskell ++ " " ++ t.haskell;
}

aspect production productionRHSElem
top::ProductionRHSElem ::= id::Id_t kwd::HasType_t t::Type
{
  top.haskell = "param_" ++ id.lexeme ;
}
