grammar silver:translation:ag:haskell:type:anytype;

import core;
import silver:base;

import silver:utils ;
import silver:utils:command hiding grammarName, parse;

import silver:translation:ag:haskell:core ;


aspect production genHaskellHelp
top::Unit ::= args::Command specs::[GrammarSpec]
{
  globalHaskell <- mkHaskell_datatype_and_unwrap_code(specs).sValue;
}


aspect production cast_t
top::Expr ::= cst::Cast_kwd e::Expr t_typerep::Decorated TypeRep
{
  top.haskell = if   t_typerep.typeEquals(t_typerep, e.typerep).bValue
		then e.haskell
		else if   t_typerep.typeName == "AnyType"
                     then -- casting to an AnyType
                      "( Haskell_AnyType_" ++ e.typerep.type_fullname ++ " (" ++ e.haskell ++") )"
                     else -- casting from an AnyType
                      "( unwrap_Haskell_AnyType_" ++ t_typerep.type_fullname ++ "( " ++ e.haskell ++ ") )" ;

}


--------------------------------------------------------
-- Creating the Haskell datatype and unwrap functions --
--------------------------------------------------------

nonterminal StringNT with sValue; 

abstract production mkHaskell_datatype_and_unwrap_code
top::StringNT ::= cl::[GrammarSpec]
{
 -- 1. get all definitions from all packages in the language
 local attribute alldefs :: Decorated Env ;
 alldefs = toEnv(all_package_defs(cl)); 

 -- 2. filter out all defs that are not "has type rep" items.
 --    The ones retained are the terminal and non-terminals in the language.
 local attribute alltypedcls :: [Decorated EnvItem] ;
 alltypedcls =  getAllTypeDcls (alldefs) ;

 -- 3. get the TypeRep of each terminal and non-terminal from the
 --    Env containing only terminals and non-terminals.
 local attribute allTNTtypereps :: [Decorated TypeRep] ;
 allTNTtypereps = env_typereps (alltypedcls ) ;

 -- 4. create a list of TypeReps such that each is a reference to 
 --    a nonterminal in the language.
 local attribute refsToNTs :: [Decorated TypeRep];
 refsToNTs = refs_to_NTtypereps ( ntsOnly ( allTNTtypereps ) ) ;

 -- 5. combine 3 and 4 to have the list of TypeReps that we may (besides
 --    the primitive types) that need to be in the Haskell_AnyType datatype.
 production attribute  allTerminalsNonTerminalsAndRefsToNTs :: [Decorated TypeRep] with ++;
 allTerminalsNonTerminalsAndRefsToNTs := allTNTtypereps +++ refsToNTs;


 -- 6. create the Haskell datatype string
 ----------------------------------------
 local attribute hany_datatype :: String ;
 hany_datatype
  = "data Haskell_AnyType = Haskell_AnyType_" ++ integerTypeRep().type_fullname  ++ "   " ++ 
                                                 integerTypeRep().haskellType      ++ " \n" ++ 
    "                     | Haskell_AnyType_" ++ floatTypeRep().type_fullname    ++ "   " ++ 
                                                 floatTypeRep().haskellType        ++ " \n" ++ 
    "                     | Haskell_AnyType_" ++ stringTypeRep().type_fullname   ++ "   " ++ 
                                                 stringTypeRep().haskellType       ++ " \n" ++ 
    "                     | Haskell_AnyType_" ++ booleanTypeRep().type_fullname  ++ "   " ++ 
                                                 booleanTypeRep().haskellType      ++ " \n" ++ 
    "                     | Haskell_AnyType_Terminal  TERMINAL \n" ++

    hany_datatype_TNT( allTerminalsNonTerminalsAndRefsToNTs ) ++

    "\n\n" ;


 -- 7. create the Haskell "unwrap" functions string
 --------------------------------------------------
 local attribute hany_unwrap :: String ;
 hany_unwrap 
  = "unwrap_Haskell_AnyType_" ++ integerTypeRep().type_fullname ++ "( " ++
           "Haskell_AnyType_" ++ integerTypeRep().type_fullname  ++ " x ) = x \n" ++ 
    "unwrap_Haskell_AnyType_" ++ integerTypeRep().type_fullname ++ " _ = error ( " ++ 
              "\"Incorrect unwrap of an AnyType - it is not an " ++ integerTypeRep().type_fullname  ++
              " type.\\n\" ) \n\n" ++

    "unwrap_Haskell_AnyType_" ++ floatTypeRep().type_fullname ++ "( " ++
           "Haskell_AnyType_" ++ floatTypeRep().type_fullname  ++ " x ) = x \n" ++ 
    "unwrap_Haskell_AnyType_" ++ floatTypeRep().type_fullname ++ " _ = error ( " ++ 
              "\"Incorrect unwrap of an AnyType - it is not an " ++ floatTypeRep().type_fullname  ++
              " type.\\n\" ) \n\n" ++

    "unwrap_Haskell_AnyType_" ++ stringTypeRep().type_fullname ++ "( " ++
           "Haskell_AnyType_" ++ stringTypeRep().type_fullname  ++ " x ) = x \n" ++ 
    "unwrap_Haskell_AnyType_" ++ stringTypeRep().type_fullname ++ " _ = error ( " ++ 
              "\"Incorrect unwrap of an AnyType - it is not an " ++ stringTypeRep().type_fullname  ++
              " type.\\n\" ) \n\n" ++

    "unwrap_Haskell_AnyType_" ++ booleanTypeRep().type_fullname ++ "( " ++
           "Haskell_AnyType_" ++ booleanTypeRep().type_fullname  ++ " x ) = x \n" ++ 
    "unwrap_Haskell_AnyType_" ++ booleanTypeRep().type_fullname ++ " _ = error ( " ++ 
              "\"Incorrect unwrap of an AnyType - it is not an " ++ booleanTypeRep().type_fullname  ++
              " type.\\n\" ) \n\n" ++

    hany_unwrap_TNT( allTerminalsNonTerminalsAndRefsToNTs ) ++

    "\n\n" ;


 -- 8. combine 6 and 7 to return as the Haskell code to add.
 top.sValue = hany_datatype ++ hany_unwrap ;
}


-- Create the unwrap functions for a list of TypeReps
function hany_unwrap_TNT
String ::= d::[Decorated TypeRep]
{
 return     if null(d) then ""
            else "unwrap_Haskell_AnyType_" ++ head(d).type_fullname ++ "( " ++
                 "Haskell_AnyType_" ++ head(d).type_fullname ++ " x ) = x \n" ++ 
                 "unwrap_Haskell_AnyType_" ++ head(d).type_fullname ++ " _ = error ( " ++ 
                    "\"Incorrect unwrap of an AnyType - it is not an " ++ head(d).type_fullname  ++
                    " type.\\n\" ) \n\n" 
                  ++
                  hany_unwrap_TNT(tail(d)) ;
}

-- Create the datatype variants for a list of TypeReps
function hany_datatype_TNT
String ::= d::[Decorated TypeRep]
{
 return     if null(d) then ""
            else "                     | Haskell_AnyType_" ++ head(d).type_fullname ++ "  " ++
                 "( " ++ head(d).haskellType ++ " ) \n" ++
                 hany_datatype_TNT(tail(d)) ;
}



-- Merge the defs in a list of Packages to make a single Env.

function all_package_defs
Decorated Defs ::= cl::[GrammarSpec]
{
 local attribute fn :: Decorated Defs ;
 fn = if   null(cl)
      then emptyDefs()
      else appendDefs ( head(cl).rootSpec.globalDefs, all_package_defs (tail(cl)) ) ;

 return fn ;
}


-- Pull the TypeReps out of an Env.
function env_typereps
[Decorated TypeRep] ::= e::[Decorated EnvItem]
{
 return if null(e) then []
        else cons(head(e).typerep,  env_typereps (tail(e)) ) ;
}


-- Make reference TypeReps to all elements of a TypeRepList.
-- This assumes all elements are nonterminals.
function refs_to_NTtypereps
[Decorated TypeRep] ::= trl::[Decorated TypeRep]
{
 return if null(trl)
        then []
        else cons(refTypeRep(head(trl)), refs_to_NTtypereps(tail(trl))) ;
}

-- Filter out all TypeReps that are not nonterminals.
function ntsOnly
[Decorated TypeRep] ::= trl::[Decorated TypeRep]
{
 return if null(trl)  then []
        else if head(trl).isNonTerminal
             then cons(head(trl), rest )
             else rest;

 local attribute rest :: [Decorated TypeRep];
 rest = ntsOnly (tail(trl)) ;
}

