grammar silver:translation:c:type:io;
import silver:translation:c:core;
import silver:base;

aspect production isFileFunction
top::Expr ::= i::IsFile_kwd l::LParen_t e1::Expr c::Comma_t e2::Expr r::RParen_t
{
  local attribute task :: String;
  task = "return (production_core_ioBoolean ( return True ) (unsafePerformIO (doesFileExist " ++ e1.translation ++ ")) )";

  top.translation = "(unsafePerformIO ( " ++ e2.translation ++ ">>= \\t -> if t then " ++  task ++ " else " ++ task ++ " ))";
}

aspect production isDirectoryFunction
top::Expr ::= i::IsDirectory_kwd l::LParen_t e1::Expr c::Comma_t e2::Expr r::RParen_t
{
  local attribute task :: String;
  task = "return (production_core_ioBoolean ( return True ) (unsafePerformIO (doesDirectoryExist " ++ e1.translation ++ ")) )";

  top.translation = "(unsafePerformIO ( " ++ e2.translation ++ ">>= \\t -> if t then " ++  task ++ " else " ++ task ++ " ))";
}

aspect production readFunction
top::Expr ::= r::ReadFile_kwd l::LParen_t e1::Expr c::Comma_t e2::Expr rp::RParen_t
{
  local attribute task :: String;
  task = "((readFile " ++ e1.translation ++ ") >>= \\str ->  (return (production_core_ioString ( return True ) str )))";

  top.translation = "(unsafePerformIO ( " ++ e2.translation ++ ">>= \\t -> if t then " ++  task ++ " else " ++ task ++ " ))";
}

aspect production cwdFunction
top::Expr ::= c::CWD_kwd l::LParen_t e::Expr r::RParen_t
{
  local attribute task :: String;
  task = "return (production_core_ioString ( return True ) (unsafePerformIO ( getCurrentDirectory )))";

  top.translation = "(unsafePerformIO ( " ++ e.translation ++ ">>= \\t -> if t then " ++  task ++ " else " ++ task ++ " ))";
}

aspect production envVarFunction
top::Expr ::= e::EnvVar_kwd l::LParen_t e1::Expr c::Comma_t e2::Expr r::RParen_t
{
  local attribute task :: String;
  task = "return (production_core_ioString (return True) (unsafePerformIO ((System.getEnv " ++ e1.translation ++ " ) `catch` (\\e -> return \"\")) ))";

  top.translation = "(unsafePerformIO ( " ++ e2.translation ++ ">>= \\t -> if t then " ++  task ++ " else " ++ task ++ " ))";
}

aspect production systemFunction
top::Expr ::= s::System_kwd l::LParen_t e1::Expr c::Comma_t e2::Expr r::RParen_t
{
  local attribute task :: String;
  task = "((System.system " ++ e1.translation ++ ") " ++
	 ">>= \\out ->  return (case out of { ExitSuccess -> 0; ExitFailure i -> -1}) " ++
	 ">>= \\code -> return (production_core_ioInteger (return True) code ))";

  top.translation = "(unsafePerformIO ( " ++ e2.translation ++ ">>= \\t -> if t then " ++  task ++ " else " ++ task ++ " ))";
}

aspect production writeFunction
top::Expr ::= w::WriteFile_kwd l::LParen_t e1::Expr c1::Comma_t e2::Expr c2::Comma_t e3::Expr r::RParen_t
{  
  local attribute task :: String;
  task = "((writeFile " ++ e1.translation ++ " " ++ e2.translation ++ ") >>= \\_ -> (return True ))";

  top.translation = "( " ++ e3.translation ++ ">>= \\t -> if t then " ++  task ++ " else " ++ task ++ " )";
}

aspect production appendFunction
top::Expr ::= a::AppendFile_kwd l::LParen_t e1::Expr c1::Comma_t e2::Expr c2::Comma_t e3::Expr r::RParen_t
{
  local attribute task :: String;
  task = "((appendFile " ++ e1.translation ++ " " ++ e2.translation ++ ") >>= \\_ -> (return True ))";

  top.translation = "( " ++ e3.translation ++ ">>= \\t -> if t then " ++  task ++ " else " ++ task ++ " )";
}

aspect production printFunction
top::Expr ::= p::Print_kwd l::LParen_t e1::Expr c::Comma_t e2::Expr r::RParen_t
{
  local attribute task :: String;
  task = "((putStr " ++ e1.translation ++ ") >>= \\_ -> (return True ))";

  top.translation = "( " ++ e2.translation ++ ">>= \\t -> if t then " ++  task ++ " else " ++ task ++ " )";
}

aspect production listContentsFunction
top::Expr ::= li::ListContents_kwd l::LParen_t e1::Expr c::Comma_t e2::Expr r::RParen_t
{
  local attribute task::String;
  task = "( " ++
	        "(getDirectoryContents " ++ e1.translation ++ ") " ++
		">>= (\\strlist -> return (production_core_ioStringList " ++
			                  "(return True) " ++ 
				          "(foldl (\\l str ->  production_core_cons_string str l ) (production_core_empty_string) strlist)" ++
				         ")" ++ 
		    ")" ++ 
         ")";

  top.translation = "(unsafePerformIO ( " ++ e2.translation ++ ">>= \\t -> if t then " ++  task ++ " else " ++ task ++ " ))";
}

aspect production unsafeIOFunction
top::Expr ::= u::UnsafeIO_kwd
{  
  top.translation = "(return True)";
}

aspect production genIntFunction
top::Expr ::= g::GenInt_kwd l::LParen_t r::RParen_t
{
  top.translation = "(genint () )";
}