grammar silver:translation:haskell:happyalex;
import silver:translation:haskell:core;
import silver:translation:ag:haskell:core;
import silver:base hiding grammarName;

import core;
import silver:utils;
import silver:utils:command;

aspect production run
top::RunUnit ::= iIn::IO args::String p::Production(Root ::= String)
{
  postOps <- if a.noCSGen then [] else [generateCS(a'', grammars)]; 
}

abstract production generateCS
top::Unit ::= a::Command specs::[GrammarSpec]
{
  forwards to generateCSHelp(a'', specs)
	with {
		ioIn = print("Generating Parsers and Scanners.\n", top.ioIn);
	};
}

abstract production generateCSHelp
top::Unit ::= a::Command specs::[GrammarSpec]
{
  top.io = writeAllCS(top.ioIn, a, specs, mergeParsers(specs));
  top.code = 0;
  top.order = 2;
}

function writeAllCS
IO ::= i::IO a::Decorated Command l::[GrammarSpec] s::[String]
{
  local attribute now :: IO;
  now = writeCSSpec(i, head(l).name, head(l).rootSpec, a, s);

  local attribute recurse :: IO;
  recurse = writeAllCS(now, a, tail(l), s);

  return if null(l) then i else recurse;
}

function writeCSSpec
IO ::= i::IO gName::String r::Decorated RootSpec a::Decorated Command s::[String]
{
  --the top level grammar name
  local attribute thn :: String;
  thn = makeName(a.grammarName);

  -- the haskell name of the grammar.
  local attribute hn :: String;
  hn = makeName(gName);
  
  local attribute has_start :: Boolean;
  has_start = r.hasStartSymbol && containsString(gName ++ ":parse", s);

  local attribute shn :: String;
  shn = "NT_" ++ makeName(r.startName);

  local attribute parser_boot :: String;
  parser_boot = 
      "module Module_" ++ hn ++ "_Parser where\n" ++
      "import {-# SOURCE #-} Module_" ++ thn ++ "\n" ++
      "doparse :: String ->  (INHS_" ++ shn ++ " -> " ++ shn ++ ")\n";


  local attribute alexFile :: String;
  alexFile = a.generatedPath ++ "Module_" ++ hn ++ "_Scanner.x";

  local attribute alexFileTemp :: String;
  alexFileTemp = a.generatedPath ++ "Module_" ++ hn ++ "_Scanner.x.temp";

  local attribute existsAlex :: IOBoolean;
  existsAlex = isFile(alexFile, i);

  local attribute cpAlex :: IOInteger;
  cpAlex = if existsAlex.bValue then system("cp " ++ alexFile ++ " " ++ alexFileTemp, existsAlex.io) else ioInteger(existsAlex.io, 0);

  local attribute prevAlex :: IOString;
  prevAlex = if existsAlex.bValue then readFile(alexFileTemp, cpAlex.io) else ioString(cpAlex.io, "");

  local attribute newAlex :: String;
  newAlex = makeAlexSpec(hn, r);

  local attribute needsAlex ::Boolean;
  needsAlex = prevAlex.sValue != newAlex;

  local attribute io2 :: IO;
  io2 = if !has_start || !needsAlex then prevAlex.io
	else writeFile(alexFile, newAlex, prevAlex.io);

  local attribute io3 :: IO;
  io3 = if !has_start  || !needsAlex then io2
	else writeFile(a.generatedPath ++ "build-scanner", "alex --ghc -o " ++ "Module_" ++ hn ++ "_Scanner.hs Module_" ++ hn ++ "_Scanner.x\n" , io2);

  local attribute io4 :: IO;
  io4 = if !has_start  || !needsAlex then io3
	else if a.noScanner
	then system("pushd " ++ a.generatedPath ++ ">> /dev/null; chmod +x ./build-scanner; popd >> /dev/null", io3).io	
	else system("pushd " ++ a.generatedPath ++ ">> /dev/null; chmod +x ./build-scanner; ./build-scanner; popd >> /dev/null", io3).io;

--	else system("alex --ghc -o " ++ a.generatedPath ++ "Module_" ++ hn ++ "_Scanner.hs " ++ alexFile, io2).io;


  local attribute happyFile :: String;
  happyFile = a.generatedPath ++ "Module_" ++ hn ++ "_Parser.y";

  local attribute happyFileTemp :: String;
  happyFileTemp = a.generatedPath ++ "Module_" ++ hn ++ "_Parser.y.temp";

  local attribute existsHappy :: IOBoolean;
  existsHappy = isFile(happyFile, io4);

  local attribute cpHappy :: IOInteger;
  cpHappy = if existsHappy.bValue then system("cp " ++ happyFile ++ " " ++ happyFileTemp, existsHappy.io) else ioInteger(existsHappy.io, 0);

  local attribute prevHappy :: IOString;
  prevHappy = if existsHappy.bValue then readFile(happyFileTemp, cpHappy.io) else ioString(cpHappy.io, "");

  local attribute newHappy :: String;
  newHappy = makeHappySpec(hn, r, a.grammarName);

  local attribute needsHappy ::Boolean;
  needsHappy = prevHappy.sValue != newHappy;

  local attribute io5 :: IO;
  io5 = if !has_start || !needsHappy then prevHappy.io
	else writeFile(a.generatedPath ++ "Module_" ++ hn ++ "_Parser.hs-boot", parser_boot, prevHappy.io);

  local attribute io6 :: IO;
  io6 = if !has_start || !needsHappy then io5
	else writeFile(happyFile, newHappy, io5);

  local attribute io7 :: IO;
  io7 = if !has_start || !needsHappy then io6
	else writeFile(a.generatedPath ++ "build-parser", "happy --ghc -a -c -i" ++ "Info_" ++ hn ++ " -o " ++ "Module_" ++ hn ++ "_Parser.hs Module_" ++ hn ++ "_Parser.y\n", io6);

--                system("pushd " ++ a.generatedPath ++ ">> /dev/null; chmod +x ./ghc-compile;                popd >> /dev/null", i5).io
--           else system("pushd " ++ a.generatedPath ++ ">> /dev/null; chmod +x ./ghc-compile; ./ghc-compile; popd >> /dev/null", i5).io;

  local attribute io8 :: IO;
  io8 = if !has_start || !needsHappy then io7
	else if a.noParser 
	then system("pushd " ++ a.generatedPath ++ ">> /dev/null; chmod +x ./build-parser; popd >> /dev/null", io7).io	
	else system("pushd " ++ a.generatedPath ++ ">> /dev/null; chmod +x ./build-parser; ./build-parser; popd >> /dev/null", io7).io;

  return io8;
}
