grammar silver:translation:haskell:happyalex;
import silver:translation:haskell:core;
import silver:base;
import silver:utils;

synthesized attribute happySpec :: String ;
synthesized attribute happyTokenSpec :: String ;

function makeHappySpec
String ::= gname::String spec::Decorated RootSpec topName::String
{  
  local attribute filtered :: FilterResult;
  filtered = filterSyntax(spec.hiddenSyntax, spec.terminalDcls, spec.ruleDcls);

  local attribute merged :: [RuleSpec];
  merged = mergeRules(filtered.ruleDcls);

  local attribute precs :: [ParserPrecSpec];
  precs = getParserPrecSpecRules(filtered.ruleDcls) +++ getParserPrecSpecTerminals(filtered.terminalDcls);

  return
	"{\n" ++
	"module Module_" ++ gname ++ "_Parser where\n" ++
	"import Module_" ++ gname ++ "_Scanner\n" ++
        "import Module_" ++ makeName(topName) ++ "\n" ++
	"}\n" ++ 
	"%name parser\n" ++
	"%tokentype { Token }\n" ++
	"%token\n" ++
        makeTokenSpec(filtered.terminalDcls) ++ "\n\n" ++
	makeParserPrecSpec(precSpecMergeSort(precs)) ++ "\n" ++
	"%%\n" ++ 
	makeRuleSpec(filterInStart(spec.startName, merged)) ++ "\n--\n" ++
	makeRuleSpec(filterOutStart(spec.startName, merged)) ++ "\n" ++
	"{\n" ++
	"happyError :: [Token] -> a\n" ++
	"happyError [] = error (\"Parse error at token: []\")\n" ++
	"happyError (tk:tks) = error (\"Parse error at token: \" ++ (show tk))\n\n" ++
	"doparse text = parser (alexScanTokens text)\n" ++
	"}\n";
}

function makeTokenSpec
String ::= specs::[TerminalSpec]
{
  return if null(specs)
	 then "    "
	 else "   " ++ makeName(head(specs).terminalName) ++ 
	      " { Token_" ++ makeName(head(specs).terminalName) ++  " $$ }\n" ++ makeTokenSpec(tail(specs)) ;
}

function makeParserPrecSpec
String ::= specs::[ParserPrecSpec]
{
  return makeParserPrecSpecHelp(head(specs).parserPrecedence - 1, specs);
}

function makeParserPrecSpecHelp
String ::= last::Integer specs::[ParserPrecSpec]
{
  return if null(specs) 
	 then "" 
	 else if last == head(specs).parserPrecedence
	      then " " ++ makeName(head(specs).precedenceName) ++ makeParserPrecSpecHelp(last, tail(specs))
	      else "\n%" ++ head(specs).parserAssociation ++ " " ++ makeName(head(specs).precedenceName) ++ 
	                    makeParserPrecSpecHelp(head(specs).parserPrecedence, tail(specs));
} 


function makeRuleSpec
String ::= specs::[RuleSpec]
{
  return if null(specs)
	 then ""
	 else makeName(head(specs).ruleLHS) ++ "\n\t: " ++ buildRHS(head(specs).ruleRHSSpec) ++ "\n\n" ++ makeRuleSpec(tail(specs)); 
}

function buildRHS
String ::= l::[RHSSpec]
{
  return if null(l)
	 then ""
	 else buildRHSSpec(head(l).ruleRHS) ++ (if head(l).parserPrecedence != 0 then " %prec " ++ makeName(head(l).ruleName) else "") ++
	      " { production_" ++ makeName(head(l).ruleName) ++ " " ++ buildHappyParams(1, head(l).ruleRHS) ++ "}\n" ++
	      if null(tail(l)) then "" else ("\t| " ++ buildRHS(tail(l)));
}

function buildHappyParams
String ::= i::Integer rhs::[String]
{
  return if null(rhs) then "" 
	 else "$" ++ toString(i) ++ " " ++ buildHappyParams(i+1, tail(rhs));
}

function buildRHSSpec
String ::= rhs::[String]
{
  return if null(rhs) then "" 
	 else makeName(head(rhs)) ++ " " ++ buildRHSSpec(tail(rhs));
}


function filterInStart
[RuleSpec] ::= s::String c2::[RuleSpec]
{
  return if   null(c2)
         then []
         else if head(c2).ruleLHS == s
	      then [head(c2)]
	      else filterInStart(s, tail(c2));
}

function filterOutStart
[RuleSpec] ::= s::String c2::[RuleSpec]
{
  return if   null(c2)
         then []
         else if head(c2).ruleLHS == s
	      then filterOutStart(s, tail(c2))
	      else cons(head(c2), filterOutStart(s, tail(c2)));
}
