grammar silver:translation:haskellc:lexyacc;
import silver:translation:haskellc:core;
import silver:base;
import silver:utils;

function makeYaccSpec
String ::= gname::String spec::Decorated RootSpec topName::String
{  
  local attribute s :: String;
  s = makeName(spec.startName);

  local attribute filtered :: FilterResult;
  filtered = filterSyntax(spec.hiddenSyntax, spec.terminalDcls, spec.ruleDcls);

  local attribute merged :: [RuleSpec];
  merged = mergeRules(filtered.ruleDcls);

  local attribute precs :: [ParserPrecSpec];
  precs = getParserPrecSpecRules(filtered.ruleDcls) +++ getParserPrecSpecTerminals(filtered.terminalDcls);

  return
	"%{\n" ++
	"#include <string.h>\n" ++
	"#include <stdlib.h>\n" ++
	"#include <stdio.h>\n" ++
	"char * pp (char *to, char *from);\n" ++
	"char * cpy (char *to);\n" ++
	"char * fix (char *str);\n" ++
	"char * parse_result;\n" ++
	"%}\n" ++ 
	"%union{\n" ++
	"  char * str;\n" ++
	"  struct terminal {\n" ++
	"    char * lexeme;\n" ++
	"    char * line;\n" ++
	"    char * column;\n" ++
	"  }term ;\n" ++
	"}\n" ++
        makeYaccTokenSpec(filtered.terminalDcls) ++ "\n\n" ++
	makeParserPrecSpec(precSpecMergeSort(precs)) ++ "\n" ++
	"%type <str> " ++ makeYaccType(merged) ++ "\n" ++
	"%%\n" ++ 
	makeRuleSpec(filterInStart(spec.startName, merged), filtered.terminalDcls) ++ "\n\n" ++

	makeRuleSpec(filterOutStart(spec.startName, merged), filtered.terminalDcls) ++ "\n" ++
	"%%\n" ++
	"main()\n" ++
	"{\n" ++
	"  parse_result = (char*)calloc(1, sizeof(char));\n" ++
	"  yyparse();\n" ++
	"  printf(parse_result);\n" ++
	"  printf(\"\\n\");\n" ++
	"}\n" ++
	"yyerror(s)\n" ++
	"char *s;\n" ++
 	"{\n" ++
	"  free(parse_result);\n" ++
	"  parse_result = s;\n" ++
 	"}\n" ++
 	"yywrap()\n" ++
	"{\n" ++
	"  return(1);\n" ++
 	"}\n" ++
	"char * pp (char *to, char *from)\n" ++
	"{\n" ++
	"  char * result;\n" ++
	"  result = (char *)calloc(strlen(to) + strlen(from) + 1, sizeof(char));\n" ++
	"  strcat(result, to);\n" ++
	"  strcat(result, from);\n" ++
	"  //free(to);\n" ++
	"  //free(from);\n" ++
	"  return result;\n" ++
	"}\n" ++
	"char * cpy (char *to)\n" ++
	"{\n" ++
	"  char * result;\n" ++
	"  result = (char *)calloc(strlen(to) + 1, sizeof(char));\n" ++
	"  strcpy(result, to);\n" ++
	"  return result;\n" ++
	"}\n" ++
	"char* fix(char* str){\n" ++
	"  int size;\n" ++
	"  char * result;\n" ++
	"  int i;\n" ++
	"  int j;\n\n" ++

	"  size = 0;\n" ++
	"  for(i = 0; i < strlen(str); i++) {\n" ++
	"      if(str[i] == '\\\"'){\n" ++
	"	size = size + 1;\n" ++
	"      }else if(str[i] == '\\\\'){\n" ++
	"	size = size + 1;\n" ++
	"      }else if(str[i] == '%'){\n" ++
	"	size = size + 1;\n" ++
	"      }\n" ++
	"      size = size + 1;    \n" ++
	"    }\n\n" ++

	"  result = (char *)calloc(size + 1, sizeof(char));\n\n" ++

	"  for(i = 0, j = 0; i < strlen(str); i++, j++){\n" ++
	"    if(str[i] == '\\\"'){\n" ++
	"      result[j] = '\\\\';\n" ++
	"       j++;\n" ++
	"    }else if(str[i] == '\\\\'){\n" ++
	"      result[j] = '\\\\';\n" ++
	"      j++;\n" ++
	"    }else if(str[i] == '%'){\n" ++
	"      result[j] = '%';\n" ++
	"      j++;\n" ++
	"    }\n" ++
	"    result[j] = str[i];\n" ++
	"  }\n" ++
	"  return result;\n" ++
	"}\n";
}

function makeYaccTokenSpec
String ::= specs::[TerminalSpec]
{
  return if null(specs)
	 then "    "
	 else "%token <term> " ++ makeName(head(specs).terminalName) ++ "\n" ++ makeYaccTokenSpec(tail(specs)) ;
}

function makeParserPrecSpec
String ::= specs::[ParserPrecSpec]
{
  return makeParserPrecSpecHelp(head(specs).parserPrecedence - 1, specs);
}

function makeParserPrecSpecHelp
String ::= last::Integer specs::[ParserPrecSpec]
{
  return if null(specs) 
	 then "" 
	 else if last == head(specs).parserPrecedence
	      then " " ++ makeName(head(specs).precedenceName) ++ makeParserPrecSpecHelp(last, tail(specs))
	      else "\n%" ++ head(specs).parserAssociation ++ " " ++ makeName(head(specs).precedenceName) ++ 
	                    makeParserPrecSpecHelp(head(specs).parserPrecedence, tail(specs));
}

function makeYaccType
String ::= l::[RuleSpec]
{
  return if null(l)
	 then ""
	 else makeName(head(l).ruleLHS) ++ " " ++ makeYaccType(tail(l));
}

function makeRuleSpec
String ::= specs::[RuleSpec] t::[TerminalSpec]
{
  return if null(specs)
	 then ""
	 else makeName(head(specs).ruleLHS) ++ "\n\t: " ++ buildRHS(head(specs).ruleRHSSpec, t) ++ "\n\n" ++ makeRuleSpec(tail(specs), t); 
}

function buildRHS
String ::= l::[RHSSpec] t::[TerminalSpec]
{
  local attribute r :: RHSSpec;
  r = head(l);

  return if null(l)
	 then ""
	 else buildRHSSpec(r.ruleRHS) ++ (if head(l).parserPrecedence != 0 then " %prec " ++ makeName(r.ruleName) else "") ++
	      " {\n\t\t$$ = " ++ rep(length(r.ruleRHS) + 1 + 6*cntTerms(r.ruleRHS, t), "pp(") ++ "\"(Production_" ++ makeName(r.ruleName) ++ " \", " ++ 
	      buildYaccParams(1, r.ruleRHS, t) ++ "\")\");\n\t\t" ++
	      "free(parse_result);\n\t\t" ++
	      "parse_result = cpy($$);\n\t" ++
	      "}\n" ++ if null(tail(l)) then "" else ("\t| " ++ buildRHS(tail(l), t));
}

function buildRHSSpec
String ::= rhs::[String]
{
  return if null(rhs) then "" 
	 else makeName(head(rhs)) ++ " " ++ buildRHSSpec(tail(rhs));
}


function buildYaccParams
String ::= i::Integer rhs::[String] t::[TerminalSpec]
{
  return if null(rhs) 
	 then " " 
	 else if terminalCheck(head(rhs), t)
	 then ("\"(Bridge_TERMINAL1 \\\"\"), fix($" ++ toString(i) ++ ".lexeme)), \"\\\" \"), $" ++ toString(i)  ++ ".line), \" \"), $" ++ toString(i) ++ ".column), \")\"), " ++ buildYaccParams(i+1, tail(rhs), t))
	 else "$" ++ toString(i) ++ ("), " ++ buildYaccParams(i+1, tail(rhs), t));
}

function cntTerms
Integer ::= l::[String] t::[TerminalSpec]
{
  return if null(l) then 0 else (if terminalCheck(head(l), t) then 1 else 0) + cntTerms(tail(l), t);
}

function rep
String ::= i::Integer s::String
{
  return if i == 0 then "" else s ++ rep(i-1, s);
}

function filterInStart
[RuleSpec] ::= s::String c2::[RuleSpec]
{
  return if   null(c2)
         then []
         else if head(c2).ruleLHS == s
	      then [head(c2)]
	      else filterInStart(s, tail(c2));
}

function filterOutStart
[RuleSpec] ::= s::String c2::[RuleSpec]
{
  return if   null(c2)
         then []
         else if head(c2).ruleLHS == s
	      then filterOutStart(s, tail(c2))
	      else cons(head(c2), filterOutStart(s, tail(c2)));
}
