grammar silver:translation:haskelljava:command;
import silver:utils:command;

synthesized attribute noCSGen :: Boolean;
attribute noCSGen occurs on PieceList, Command;
synthesized attribute onlyCSGen :: Boolean;
attribute onlyCSGen occurs on PieceList, Command;
synthesized attribute javaClassPath :: String;
attribute javaClassPath occurs on PieceList, Command;
synthesized attribute jvmParameters :: String;
attribute jvmParameters occurs on PieceList, Command;

-- Aspect for  --jclasspath
aspect production cRootAll
top::Command ::= c1::PieceList
{
  flagLookups <- [flagLookup("--jclasspath", true)];
  uses <- ["\t--jclasspath <path> Java classpath for external calls to Java\n"];

  local attribute javaClassPathFlags :: [Flag];
  javaClassPathFlags = findFlag("--jclasspath", top.flags);

  top.javaClassPath = if null(javaClassPathFlags) then "." else head(javaClassPathFlags).chunk ;
}

-- Aspect for  --jvmparameters
aspect production cRootAll
top::Command ::= c1::PieceList
{
  flagLookups <- [flagLookup("--jvmparameters", true)];
  uses <- ["\n\t--jvmparameters <string> Parameters for the Java virtual machine used to run Copper\n"];

  local attribute jvmParametersFlags :: [Flag];
  jvmParametersFlags = findFlag("--jvmparameters", top.flags);

  top.jvmParameters = if null(jvmParametersFlags) then "" else head(jvmParametersFlags).chunk ;
}

-- Aspect for  --nocs
aspect production cRootAll
top::Command ::= c1::PieceList
{
  flagLookups <- [flagLookup("--nocs", false)];
  uses <- ["\t--nocs (don't generate concrete syntax)\n"];

  local attribute nocsFlags :: [Flag];
  nocsFlags = findFlag("--nocs", top.flags);

  top.noCSGen = ! null(nocsFlags) ;
}

-- Aspect for  --onlycs
aspect production cRootAll
top::Command ::= c1::PieceList
{
  flagLookups <- [flagLookup("--onlycs", false)];
  uses <- ["\t--onlycs (only generate concrete syntax)\n"];

  local attribute onlycsFlags :: [Flag];
  onlycsFlags = findFlag("--onlycs", top.flags);

  top.onlyCSGen = ! null(onlycsFlags) ;
}

--------------------------------------------------------------------------------------------------------
-- Previous Version
--------------------------------------------------------------------------------------------------------
--
--
--aspect production rootAll
--top::Command ::= c1::PieceList
--{
--  uses <- ["\t--nocs (don't generate concrete syntax)\n"];
--  top.noCSGen = c1.noCSGen;
--  uses <- ["\t--onlycs (only generate concrete syntax)\n"];
--  top.onlyCSGen = c1.onlyCSGen;
--  uses <- ["\t--jclasspath Java classpath for external calls to Java\n"];
--  top.javaClassPath = c1.javaClassPath;
--}
--
--aspect production pieceOne
--top::PieceList ::= c1::Chunk_t
--{
--  top.noCSGen = false;
--  top.onlyCSGen = false;
--  top.javaClassPath = ".";
--}
--
--aspect production longFlag
--top::PieceList ::= c1::LongFlag_t c2::PieceList
--{
--  claimed <- (c1.lexeme == "--nocs" || c1.lexeme == "--onlycs");
--
--  top.noCSGen = if c1.lexeme == "--nocs"
--                then if c2.onlyCSGen
--                     then error("Cannot generate no concrete syntax and only concrete syntax concurrently.")
--                     else true
--	        else c2.noCSGen;
--  top.onlyCSGen = if c1.lexeme == "--onlycs"
--                then if c2.noCSGen
--                     then error("Cannot generate no concrete syntax and only concrete syntax concurrently.")
--                     else true
--	        else c2.onlyCSGen;
--  top.javaClassPath = c2.javaClassPath;
--}
--
--aspect production longFlagChunk
--top::PieceList ::= c1::LongFlag_t c2::Chunk_t c3::PieceList
--{
--  claimed <- (c1.lexeme == "--jclasspath");
--  top.javaClassPath = if c1.lexeme == "--jclasspath"
--                      then c2.lexeme
--                      else c3.javaClassPath;
--
--  top.noCSGen = c3.noCSGen;
--  top.onlyCSGen = c3.onlyCSGen;
--}
--
--aspect production shortFlag
--top::PieceList ::= c1::ShortFlag_t c2::PieceList
--{
--  top.noCSGen = c2.noCSGen;
--  top.onlyCSGen = c2.onlyCSGen;
--  top.javaClassPath = c2.javaClassPath;
--}
--
--aspect production shortFlagChunk
--top::PieceList ::= c1::ShortFlag_t c2::Chunk_t c3::PieceList
--{
--  top.noCSGen = c3.noCSGen;
--  top.onlyCSGen = c3.onlyCSGen;
--  top.javaClassPath = c3.javaClassPath;
--}
--
