grammar silver:translation:haskelljava:concretesyntax;

import silver:base:cs:default;
import silver:base:core;

import silver:analysis:typechecking:core;
import silver:base:type:anytype;
import silver:extension:functions;
import silver:extension:list ;
import silver:extension:patternmatching with isAspect as pmIsAspect;
import silver:extension:autocopy with isAspect as acIsAspect;
import silver:utils;

synthesized attribute hasValidActionCode :: Boolean;
synthesized attribute hasActionCode :: Boolean;
synthesized attribute actionCode :: ActionCode;
synthesized attribute aCode :: String;

nonterminal ActionCode with {aCode,pp};
nonterminal ActionCode_c with {hasValidActionCode,actionCode,env,defs,globalEnv,grammarName,localsEnv,signatureNames,signatureTypes,signatureEnv};

--terminal ac /[\{][\@]([^\@]|[\n\t\ ])*[\@][\}]/ lexer precedence = 1;

--concrete production actionCode_c
--top::ActionCode ::= action::ac
--{
--   top.pp = action.lexeme;
--   top.aCode = substring(2,length(action.lexeme) - 2,action.lexeme);
--}

abstract production actionCodeEmpty
top::ActionCode ::=
{
  top.pp = ""; --"{@@}";
  top.aCode = "";
  top.errors = "";
  top.typeErrors = "";
}

abstract production actionCodeError
top::ActionCode ::= err::String
{
  top.errors = err;
  forwards to actionCodeEmpty();
}

abstract production actionCodeConcat
top::ActionCode ::= c1::ActionCode c2::ActionCode
{
  top.pp = c1.pp ++ c2.pp;
  top.aCode = c1.aCode ++ c2.aCode;
  top.errors = c1.errors ++ c2.errors;
  top.typeErrors = c1.typeErrors ++ c2.typeErrors;
}

abstract production actionCodeCode
top::ActionCode ::= acode::String
{
  top.pp = acode;
  top.aCode = acode;
  top.errors = "";
  top.typeErrors = "";
}

attribute actionCode occurs on ProductionStmts,ProductionStmt,AttributeDef,LHSExpr,Expr,Exprs;
attribute hasValidActionCode occurs on ProductionStmts,ProductionStmt,AttributeDef,LHSExpr,Expr,Exprs;

attribute errors occurs on ActionCode,ActionCode_c;
attribute typeErrors occurs on ActionCode,ActionCode_c;

terminal Action_kwd /action/ lexer precedence = 5;

concrete production concreteProductionDclWithAction
top::AGDcl ::= c::Concrete_kwd p::Production_kwd id::Id_t
   ns::ProductionSignature pm::OptProductionModifiers
   b::ProductionBody act::Action_kwd acode::ActionCode_c
{
  production attribute fName :: String;
  fName := top.grammarName.name ++ ":" ++ id.lexeme;

  top.ruleDcls = [ruleSpec(head(ns.sigTypes).typeName, [rhsSpecAction(fName, getNames(tail(ns.sigTypes)), top.grammarName, pm, acode.actionCode)])];

  top.pp = "concrete production " ++ id.lexeme ++  "\n" ++ 
	     ns.pp  ++ "  \n{\n" ++ b.pp ++ "\n}\naction\n  {@" ++
             acode.actionCode.aCode ++ "\n@}";

  acode.actionCodeType = productionActionType();

--  local attribute er1 :: String;
--  er1 = if length(getFullNameDcl(id.lexeme, top.env)) > 1
--       then top.location ++ "Error: Name '" ++ id.lexeme ++ "' is already bound.\n"
--       else "";	

--  local attribute er2 :: String;
--  er2 = if length(getValueDcl(fName, top.env)) > 1
--       then top.location ++ "Error: Value '" ++ fName ++ "' is already bound.\n"
--       else "";	
--
--  top.errors = er1 ++ er2 ++ ns.errors ++ b.errors ++ acode.errors;  
--  top.typeErrors = b.typeErrors ++ acode.typeErrors;

  forwards to concreteProductionDcl(c, p, id, ns'', pm'', b'');
}

aspect production concreteProductionDclWithAction
top::AGDcl ::= c::Concrete_kwd p::Production_kwd id::Id_t
   ns::ProductionSignature pm::OptProductionModifiers
   b::ProductionBody act::Action_kwd acode::ActionCode_c
{
  acode.env = appendDefsEnv(appendDefs(acode.defs, appendDefs(ns.defs, addThisDcl(fName, emptyDefs()))), top.env);

  acode.signatureEnv = toEnv(ns.defs);
  acode.localsEnv = toEnv(acode.defs);

  acode.signatureNames = ns.sigNames;
  acode.signatureTypes = ns.sigTypes;
}

function generateRHSDecls
ActionCode ::= sNames::[String] sTypes::[Decorated TypeRep] count::Integer
{
  local attribute elemType :: ActionCode;
  elemType = if head(sTypes).typeName == "String"
             then actionCodeCode("String ")
             else actionCodeCode("edu.umn.cs.melt.copper.runtime.parsetree.plain." ++
               (if head(sTypes).isNonTerminal
                then "ParseTreeProdNode "
                else "ParseTreeTermNode "));

  local attribute elemName :: ActionCode;
  elemName = actionCodeCode(makeName(head(sNames)));

  local attribute elemRetrievalCode :: ActionCode;
  elemRetrievalCode = if head(sTypes).typeName == "String"
                      then actionCodeEmpty()
                      else if count == -1
                           then actionCodeCode(" = _parseTree")
                           else actionCodeCode(" = getChild" ++
                           (if head(sTypes).isNonTerminal then "NT" else "T") ++
                            "(" ++ toString(count) ++ ")");


  return if null(sNames)
         then actionCodeEmpty()
         else actionCodeConcat(elemType,
               actionCodeConcat(elemName,
                actionCodeConcat(elemRetrievalCode,
                 actionCodeConcat(actionCodeCode(";\n"),
                  generateRHSDecls(tail(sNames),tail(sTypes),count + 1)))));
}

concrete production actionCode_c
top::ActionCode_c ::= l::LCurly_t stmts::ProductionStmts r::RCurly_t
{
  top.defs = stmts.defs;

  local attribute preambleCode :: ActionCode;
  preambleCode = if top.actionCodeType.isProductionAction
                 then generateRHSDecls(top.signatureNames,top.signatureTypes,-1)
                 else actionCodeEmpty();

  top.actionCode = if stmts.hasValidActionCode && stmts.errors == "" && stmts.typeErrors == ""
                   then actionCodeConcat(preambleCode,stmts.actionCode)
                   else error(stmts.errors ++ stmts.actionCode.errors ++ stmts.typeErrors ++ stmts.actionCode.typeErrors);

  top.errors = stmts.errors ++ stmts.actionCode.errors;
  top.typeErrors = stmts.typeErrors ++ stmts.actionCode.typeErrors;
}

concrete production actionCodeEmpty_c
top::ActionCode_c ::= l::LCurly_t r::RCurly_t
{
  top.actionCode = actionCodeEmpty();
  forwards to actionCode_c(l,productionStmtsNone(),r);
}

aspect production productionStmtsSingle
top::ProductionStmts ::= stmt::ProductionStmt
{
  top.hasValidActionCode = stmt.hasValidActionCode;
  top.actionCode = stmt.actionCode;
}

aspect production productionStmtsCons
top::ProductionStmts ::= h::ProductionStmt t::ProductionStmts
{
  top.hasValidActionCode = h.hasValidActionCode && t.hasValidActionCode;
  top.actionCode = actionCodeConcat(h.actionCode,t.actionCode);
}

aspect production productionStmtsAppend
top::ProductionStmts ::= h::ProductionStmts t::ProductionStmts
{
  top.hasValidActionCode = h.hasValidActionCode && t.hasValidActionCode;
  top.actionCode = actionCodeConcat(h.actionCode,t.actionCode);
}
aspect production forwardsDcl
top::ProductionStmt ::= f::Forwards_kwd t::To_kwd e::Expr s::Semi_t
{
  -- Possibly construe this to mean that one action block gets another's code?
  top.hasValidActionCode = false;
  top.actionCode = actionCodeConcat(e.actionCode,
                    actionCodeError(top.location ++ " Error: Construct 'forwardsDcl' not allowed in parser actions\n"));
}

aspect production forwardsDclWith
top::ProductionStmt ::= f::Forwards_kwd t::To_kwd e::Expr w::With_kwd l::LCurly_t inh::ForwardInh r::RCurly_t s::Semi_t
{
  -- Possibly construe this to mean that one action block gets another's code?
  top.hasValidActionCode = false;
  top.actionCode = actionCodeConcat(e.actionCode,
                    actionCodeError(top.location ++ " Error: Construct 'forwardsDclWith' not allowed in parser actions\n"));
}

aspect production forwardingDclWith
top::ProductionStmt ::= f::Forwarding_kwd w::With_kwd l::LCurly_t inh::ForwardInh r::RCurly_t s::Semi_t
{
  top.hasValidActionCode = false;
  top.actionCode = actionCodeError(top.location ++ " Error: Construct 'forwardingDclWith' not allowed in parser actions\n");
}

aspect production attributeDclLocal
top::ProductionStmt ::= l::Local_kwd  at::Attribute_kwd a::Id_t 
                        ht::HasType_t te::Type  s::Semi_t
{
   top.hasValidActionCode = te.typerep.isParserAttrType;
   top.actionCode = if te.typerep.isParserAttrType
                    then actionCodeCode(te.typerep.parserAttrType ++ " " ++ makeName(fName) ++ ";\n")
                    else actionCodeError(top.location ++ "Error: Type " ++ te.typerep.typeName ++ " is not valid for a local attribute declaration in a parser action\n");
}

aspect production stmtToAttrDef
top::ProductionStmt ::= a::AttributeDef
{
  top.hasValidActionCode = a.hasValidActionCode;
  top.actionCode = a.actionCode;
}

aspect production attrDef
top::AttributeDef ::= lhs::LHSExpr e::Expr
{
  top.hasValidActionCode = lhs.hasValidActionCode && e.hasValidActionCode;
  top.actionCode = actionCodeConcat(lhs.actionCode,
                    actionCodeConcat(actionCodeCode(" = "),
                     actionCodeConcat(e.actionCode,actionCodeCode(" ;\n"))));
}

aspect production nestedExpr
top::Expr ::= l::LParen_t e::Expr r::RParen_t
{
  top.hasValidActionCode = e.hasValidActionCode;
  top.actionCode = actionCodeConcat(actionCodeCode("( "),
                     actionCodeConcat(e.actionCode,actionCodeCode(" )")));
}

aspect production productionReference
top::Expr ::= q::QName
{
 -- FIXME
 -- Add protections against non-parser function calls.
  top.hasValidActionCode = true;
  top.actionCode = actionCodeCode(makeName(q.name));
}


aspect production baseExpr
top::Expr ::= q::QName
{
  -- FIXME
  -- Add protections against non-parser variable references.
  top.hasValidActionCode = true;
  top.actionCode = actionCodeCode(makeName(translateTopName(fName)));
}

aspect production dontDecorateExpr
top::Expr ::= q::QName
{
  top.hasValidActionCode = false;
  top.actionCode = actionCodeError(top.location ++ " Error: Construct 'dontDecorateExpr' not allowed in parser actions\n");
}

aspect production decorateExpr
top::Expr ::= q::QName
{
  top.hasValidActionCode = false;
  top.actionCode = actionCodeError(top.location ++ " Error: Construct 'decorateExpr' not allowed in parser actions\n");
}


aspect production productionApp
top::Expr ::= e::Expr l::LParen_t es::Exprs r::RParen_t
{
  top.hasValidActionCode = e.hasValidActionCode && es.hasValidActionCode;
  top.actionCode = actionCodeConcat(e.actionCode,
                    actionCodeConcat(actionCodeCode(" ( "),
                     actionCodeConcat(es.actionCode,actionCodeCode(" ) "))));
}

aspect production emptyList
top::Expr ::= lb::LSqr_t rb::RSqr_t
{
  top.hasValidActionCode = false;
  top.actionCode = actionCodeError(top.location ++ " Error: Untyped lists not allowed in parser action blocks; use [::type] notation instead.\n");
  --top.actionCode = actionCodeCode("new java.util.LinkedList()");
}

aspect production emptyListWType
top::Expr ::= lb::LSqr_t ht::HasType_t t::Type rb::RSqr_t
{
  top.hasValidActionCode = t.typerep.isParserAttrType;
  top.actionCode = if t.typerep.isParserAttrType
                   then actionCodeCode("new java.util.LinkedList< " ++ t.typerep.parserAttrType ++ " >()")
                   else actionCodeError(top.location ++ " Error: Type " ++ t.typerep.typeName ++ " is not valid as a list element type in a parser action\n");
}

aspect production fullList
top::Expr ::= lb::LSqr_t es::Exprs rb::RSqr_t
{
  top.hasValidActionCode = es.hasValidActionCode;
  top.actionCode = actionCodeConcat(actionCodeCode("newList ( "),
                    actionCodeConcat(es.actionCode,actionCodeCode(" ) ")));
}

aspect production consList
top::Expr ::= ck::Cons_t lp::LParen_t h::Expr c::Comma_t t::Expr rp::RParen_t 
{
  top.hasValidActionCode = h.hasValidActionCode && t.hasValidActionCode;
  top.actionCode = actionCodeConcat(actionCodeCode("listCons ( "),
                    actionCodeConcat(h.actionCode,
                     actionCodeConcat(actionCodeCode(","),
                      actionCodeConcat(t.actionCode,actionCodeCode(" ) ")))));
}


aspect production emptyProductionApp
top::Expr ::= e::Expr l::LParen_t r::RParen_t
{
  top.hasValidActionCode = e.hasValidActionCode;
  top.actionCode = actionCodeConcat(e.actionCode,actionCodeCode(" () "));
}

aspect production cast_t
top::Expr ::= cst::Cast_kwd e::Expr t_typerep::Decorated TypeRep
{
  top.hasValidActionCode = e.hasValidActionCode && t_typerep.isParserAttrType;

  -- FIXME
  top.actionCode = if top.hasValidActionCode
                   then e.actionCode
                   else if !t_typerep.isParserAttrType
                        then actionCodeConcat(e.actionCode,
                                              actionCodeError(top.location ++ " Error: Type '" ++ t_typerep.typeName ++ "' not allowed in parser actions\n"))
                        else e.actionCode;
}


aspect production functionApplicationDispatcher
top::Expr ::= e::Expr es::Exprs
{
  top.hasValidActionCode = e.hasValidActionCode && es.hasValidActionCode;
  top.actionCode = actionCodeConcat(e.actionCode,
                    actionCodeConcat(actionCodeCode(" ( "),
                     actionCodeConcat(es.actionCode,actionCodeCode(" ) "))));
}


aspect production genericApplicationDispatcher
top::Expr ::= e::Expr es::Exprs
{
  top.hasValidActionCode = e.hasValidActionCode && es.hasValidActionCode;
  top.actionCode = actionCodeConcat(e.actionCode,
                    actionCodeConcat(actionCodeCode(" "),es.actionCode));
}

function translateTopName
String ::= orig::String
{
   return if orig == "lexeme"
          then "lexeme"
          else if orig == "line"
               then "_pos.getLine()"
               else if orig == "column"
                    then "_pos.getColumn()"
                    else orig;
}


function translateFieldName
String ::= orig::String te::[Decorated TypeRep]
{
   return if !null(te) && head(te).isTerminal
          then if orig == "lexeme"
               then "getToken().getLexeme()"
               else if orig == "line"
                    then "getPosition().getLine()"
                    else if orig == "column"
                         then "getPosition().getColumn()"
                         else ""
          else if orig == "listContains"
               then "listContains"
               else "";
}

aspect production atAccess
top::Expr ::= e::Expr d::At_t q::QName
{
  -- FIXME
  -- Make certain that parser actions are restrained to accessing the
  -- correct type of attributes (using 'translateFieldName' ought to secure
  -- this)

  -- Add type protections.

  local attribute fieldTrans :: String;
  fieldTrans = translateFieldName(makeName(fName),[e.typerep]);

  top.hasValidActionCode = e.hasValidActionCode && fieldTrans != "";
  top.actionCode = if fieldTrans != ""
                   then actionCodeConcat(e.actionCode,
                         actionCodeConcat(actionCodeCode("."),
                          actionCodeCode(fieldTrans)))
                   else actionCodeConcat(e.actionCode,
                         actionCodeError(top.location ++ " Error: '" ++ makeName(fName) ++ "' is not a valid field name on '" ++ e.pp ++ "'"));
}

aspect production trueConst
top::Expr ::= t::True_kwd
{
  top.hasValidActionCode = true;
  top.actionCode = actionCodeCode("true");
}

aspect production falseConst
top::Expr ::= f::False_kwd
{
  top.hasValidActionCode = true;
  top.actionCode = actionCodeCode("false");
}

aspect production and
top::Expr ::= e1::Expr a::And_t e2::Expr
{
  top.hasValidActionCode = e1.hasValidActionCode && e2.hasValidActionCode;
  top.actionCode = actionCodeConcat(e1.actionCode,
                    actionCodeConcat(actionCodeCode(" && "),e2.actionCode));
}

aspect production or
top::Expr ::= e1::Expr o::Or_t e2::Expr
{
  top.hasValidActionCode = e1.hasValidActionCode && e2.hasValidActionCode;
  top.actionCode = actionCodeConcat(e1.actionCode,
                    actionCodeConcat(actionCodeCode(" || "),e2.actionCode));
}

aspect production not
top::Expr ::= n::Not_t e::Expr
{
  top.hasValidActionCode = e.hasValidActionCode;
  top.actionCode = actionCodeConcat(actionCodeCode("!"),e.actionCode);
}

aspect production gt
top::Expr ::= e1::Expr g::GT_t e2::Expr
{
  top.hasValidActionCode = e1.hasValidActionCode && e2.hasValidActionCode;
  top.actionCode = actionCodeConcat(e1.actionCode,
                    actionCodeConcat(actionCodeCode(" > "),e2.actionCode));
}

aspect production lt
top::Expr ::= e1::Expr l::LT_t e2::Expr
{
  top.hasValidActionCode = e1.hasValidActionCode && e2.hasValidActionCode;
  top.actionCode = actionCodeConcat(e1.actionCode,
                    actionCodeConcat(actionCodeCode(" < "),e2.actionCode));
}


aspect production gteq
top::Expr ::= e1::Expr g::GTEQ_t e2::Expr
{
  top.hasValidActionCode = e1.hasValidActionCode && e2.hasValidActionCode;
  top.actionCode = actionCodeConcat(e1.actionCode,
                    actionCodeConcat(actionCodeCode(" >= "),e2.actionCode));
}


aspect production lteq
top::Expr ::= e1::Expr l::LTEQ_t e2::Expr
{
  top.hasValidActionCode = e1.hasValidActionCode && e2.hasValidActionCode;
  top.actionCode = actionCodeConcat(e1.actionCode,
                    actionCodeConcat(actionCodeCode(" <= "),e2.actionCode));
}


aspect production eqeq
top::Expr ::= e1::Expr eq::EQEQ_t e2::Expr
{
  top.hasValidActionCode = e1.hasValidActionCode && e2.hasValidActionCode;
  top.actionCode = actionCodeConcat(e1.actionCode,
                    actionCodeConcat(actionCodeCode(" == "),e2.actionCode));
}


aspect production neq
top::Expr ::= e1::Expr n::NEQ_t e2::Expr
{
  top.hasValidActionCode = e1.hasValidActionCode && e2.hasValidActionCode;
  top.actionCode = actionCodeConcat(e1.actionCode,
                    actionCodeConcat(actionCodeCode(" != "),e2.actionCode));
}


aspect production ifThenElse
top::Expr ::= i::If_kwd e1::Expr t::Then_kwd e2::Expr e::Else_kwd e3::Expr
{
  top.hasValidActionCode = e1.hasValidActionCode && e2.hasValidActionCode && e3.hasValidActionCode;
  top.actionCode = actionCodeConcat(actionCodeCode(" ( ( "),
                    actionCodeConcat(e1.actionCode,
                     actionCodeConcat(actionCodeCode(" ) ? ( "),
                      actionCodeConcat(e2.actionCode,
                       actionCodeConcat(actionCodeCode(" ) : ( "),
                        actionCodeConcat(e3.actionCode,
                         actionCodeCode(" ) ) ")))))));
}

aspect production appendList
top::Expr ::= l::Expr o::AppendOp_t r::Expr
{
  top.hasValidActionCode = l.hasValidActionCode && r.hasValidActionCode;
  top.actionCode = actionCodeConcat(actionCodeCode("listcat ( "),
                    actionCodeConcat(l.actionCode,
                     actionCodeConcat(actionCodeCode(", "),
                      actionCodeConcat(r.actionCode,actionCodeCode(" ) ")))));
}

aspect production listLength
top::Expr ::= e::Expr
{
  top.hasValidActionCode = e.hasValidActionCode;
  top.actionCode = actionCodeConcat(e.actionCode,actionCodeCode(".length()"));
}

aspect production nullList
top::Expr ::= n::Null_t lp::LParen_t l::Expr rp::RParen_t 
{
  top.hasValidActionCode = l.hasValidActionCode;
  top.actionCode = actionCodeConcat(actionCodeCode(" ( "),
                    actionCodeConcat(l.actionCode,actionCodeCode(".isEmpty() )")));
}

aspect production headListNoCast
top::Expr ::= l::Expr 
{
  top.hasValidActionCode = l.hasValidActionCode;
  top.actionCode = actionCodeConcat(actionCodeCode(" ( "),
                    actionCodeConcat(l.actionCode,actionCodeCode(".getFirst() )")));
}

aspect production tailList
top::Expr ::= t::Tail_t lp::LParen_t l::Expr rp::RParen_t 
{
  top.hasValidActionCode = l.hasValidActionCode;
  top.actionCode = actionCodeConcat(actionCodeCode("listCdr ( "),
                    actionCodeConcat(l.actionCode,actionCodeCode(" )")));
}

aspect production intConst
top::Expr ::= i::Int_t
{
  top.hasValidActionCode = true;
  top.actionCode = actionCodeCode(i.lexeme);
}

aspect production floatConst
top::Expr ::= f::Float_t
{
  top.hasValidActionCode = true;
  top.actionCode = actionCodeCode(f.lexeme);
} 

aspect production plus
top::Expr ::= e1::Expr p::Plus_t e2::Expr
{
  top.hasValidActionCode = e1.hasValidActionCode && e2.hasValidActionCode;
  top.actionCode = actionCodeConcat(e1.actionCode,
                    actionCodeConcat(actionCodeCode(" + "),e2.actionCode));
}
aspect production minus
top::Expr ::= e1::Expr m::Minus_t e2::Expr
{
  top.hasValidActionCode = e1.hasValidActionCode && e2.hasValidActionCode;
  top.actionCode = actionCodeConcat(e1.actionCode,
                    actionCodeConcat(actionCodeCode(" - "),e2.actionCode));
}
aspect production multiply
top::Expr ::= e1::Expr m::Multiply_t e2::Expr
{
  top.hasValidActionCode = e1.hasValidActionCode && e2.hasValidActionCode;
  top.actionCode = actionCodeConcat(e1.actionCode,
                    actionCodeConcat(actionCodeCode(" * "),e2.actionCode));
}
aspect production divide
top::Expr ::= e1::Expr d::Divide_t e2::Expr
{
  top.hasValidActionCode = e1.hasValidActionCode && e2.hasValidActionCode;
  top.actionCode = actionCodeConcat(e1.actionCode,
                    actionCodeConcat(actionCodeCode(" / "),e2.actionCode));
}
aspect production neg
top::Expr ::= m::Minus_t e::Expr
{
  top.hasValidActionCode = e.hasValidActionCode;
  top.actionCode = actionCodeConcat(actionCodeCode("-( "),
                    actionCodeConcat(e.actionCode,
                     actionCodeCode(" ) ")));
}

aspect production stringConst
top::Expr ::= s::String_t
{
  top.hasValidActionCode = true;
  top.actionCode = actionCodeCode(s.lexeme);
}

aspect production defaultPlusPlus
top::Expr ::= e1::Expr e2::Expr
{
  top.hasValidActionCode = e1.hasValidActionCode && e2.hasValidActionCode;
  top.actionCode = actionCodeConcat(e1.actionCode,
                    actionCodeConcat(actionCodeCode(" + "),e2.actionCode));
}

aspect production stringPlusPlus
top::Expr ::= e1::Expr e2::Expr
{
  top.hasValidActionCode = e1.hasValidActionCode && e2.hasValidActionCode;
  top.actionCode = actionCodeConcat(e1.actionCode,
                    actionCodeConcat(actionCodeCode(" + "),e2.actionCode));
}

aspect production exprsEmpty
top::Exprs ::=
{
  top.hasValidActionCode = true;
  top.actionCode = actionCodeEmpty();
}

aspect production exprsSingle
top::Exprs ::= e::Expr
{
  top.hasValidActionCode = e.hasValidActionCode;
  top.actionCode = e.actionCode;
}

aspect production exprsCons
top::Exprs ::= e1::Expr c::Comma_t e2::Exprs
{
  top.hasValidActionCode = e1.hasValidActionCode && e2.hasValidActionCode;
  top.actionCode = actionCodeConcat(e1.actionCode,
                    actionCodeConcat(actionCodeCode(","),e2.actionCode));
}

aspect production exprWith
top::Expr ::= ev::Decorate_kwd e::Expr w::With_kwd l::LCurly_t inh::ExprInh r::RCurly_t
{
  top.hasValidActionCode = false;
  top.actionCode = actionCodeConcat(e.actionCode,
                    actionCodeError(top.location ++ " Error: Construct 'exprWith' not allowed in parser actions\n"));
}

aspect production fakeLHSExpr
top::LHSExpr ::= c1::QName c2::Decorated TypeRep
{
  -- FIXME
  -- Add any typechecks.
  top.hasValidActionCode = true;
  top.actionCode = actionCodeCode(makeName(fName));
}

aspect production lhsExprOne
top::LHSExpr ::= id::Id_t
{
  -- FIXME
  -- Add any typechecks.
  top.hasValidActionCode = id.lexeme != head(top.signatureNames);
  top.actionCode = if top.hasValidActionCode
                   then actionCodeCode(makeName(fName))
                   else actionCodeError(top.location ++ " Error: Cannot reference LHS in a parser action\n");
}

aspect production lhsExprTwo
top::LHSExpr ::= id::Id_t d::Dot_t q::QName
{
  -- FIXME
  -- Add any typechecks.
  top.hasValidActionCode = id.lexeme != head(top.signatureNames);--true;
  top.actionCode = if top.hasValidActionCode
                   then actionCodeCode(makeName(fName1 ++ "." ++ fName2))
                   else actionCodeError(top.location ++ " Error: Cannot reference LHS in a parser action\n");
}

function generateSigTypeList
[Decorated TypeRep] ::= signatNames::[String] e::Decorated Env
{
   local attribute typeItem :: [Decorated EnvItem];
   typeItem = if null(signatNames)
              then []
              else getTypeDclOne(head(signatNames),e);

   return if null(signatNames)
          then []
          else if null(typeItem)
               then [topTypeRep()] +++ generateSigTypeList(tail(signatNames),e)
               else [head(typeItem).typerep] +++ generateSigTypeList(tail(signatNames),e);
}
