grammar silver:translation:haskelljava:concretesyntax;
import silver:base:cs:default;
import silver:base:core;
import silver:analysis:typechecking:core;

terminal Operator_kwd /operator/ lexer precedence = 5;
terminal Static_kwd /static/ lexer precedence = 5;
terminal Dynamic_kwd /dynamic/ lexer precedence = 5;
terminal Class_kwd /class/ lexer precedence = 5;
terminal Prefix_kwd /prefix/ lexer precedence = 5;

terminal Dominates_t 'dominates' lexer precedence = 5;
terminal SubmitsTo_t 'submitsTo' lexer precedence = 5;

terminal Classes_kwd /classes/ lexer precedence = 5;

attribute env occurs on TerminalSpec;
attribute env occurs on TermPrecList, ClassList;
attribute env occurs on TerminalModifiers, TerminalModifier;

attribute errors occurs on TerminalSpec;
attribute errors occurs on TermPrecList, ClassList;
attribute errors occurs on TerminalModifiers, TerminalModifier;

attribute defs occurs on TermPrecList;

attribute typeErrors occurs on TerminalSpec;
attribute typeErrors occurs on TermPrecList, ClassList;
attribute typeErrors occurs on TerminalModifiers, TerminalModifier;

synthesized attribute hasLexerClass :: Boolean;
synthesized attribute hasStaticPrec :: Boolean;
synthesized attribute hasDynamicPrec :: Boolean;

synthesized attribute lexerClass :: String;
synthesized attribute staticPrec :: Integer;
synthesized attribute dynamicPrec :: Integer;

synthesized attribute usePrefix :: Boolean;

attribute hasLexerClass occurs on TerminalModifier;
attribute hasLexerClass occurs on TerminalModifiers;

attribute hasStaticPrec occurs on TerminalModifier;
attribute hasStaticPrec occurs on TerminalModifiers;

attribute hasDynamicPrec occurs on TerminalModifier;
attribute hasDynamicPrec occurs on TerminalModifiers;

attribute lexerClass occurs on TerminalModifier;
attribute lexerClass occurs on TerminalModifiers;

attribute staticPrec occurs on TerminalModifier;
attribute staticPrec occurs on TerminalModifiers;

attribute dynamicPrec occurs on TerminalModifier;
attribute dynamicPrec occurs on TerminalModifiers;

attribute usePrefix occurs on TerminalModifier;
attribute usePrefix occurs on TerminalModifiers;

synthesized attribute hasDominates :: Boolean occurs on TerminalModifier, TerminalModifiers ;
synthesized attribute hasSubmitsTo :: Boolean occurs on TerminalModifier, TerminalModifiers ;
synthesized attribute dominates :: [String] occurs on TerminalModifier, TerminalModifiers ;
synthesized attribute submitsTo :: [String] occurs on TerminalModifier, TerminalModifiers ;
synthesized attribute precTermList :: [String] occurs on TermPrecList ;

synthesized attribute hasLexerClasses :: Boolean occurs on TerminalModifier, TerminalModifiers ;
synthesized attribute lexerClasses    :: [String]  occurs on TerminalModifier, TerminalModifiers ;


attribute hasActionCode occurs on TerminalModifier, TerminalModifiers;
attribute actionCode occurs on TerminalSpec,TerminalModifiers,TerminalModifier;

aspect production terminalDclModifiers
top::AGDcl ::= t::Terminal_kwd id::Id_t r::RegExpr tm::TerminalModifiers s::Semi_t
{
   top.errors = tm.errors;
   top.typeErrors = tm.typeErrors;
}

aspect production terminalDclAllModifiers
top::AGDcl ::= i::Ignore_kwd t::Terminal_kwd id::Id_t r::RegExpr tm::TerminalModifiers s::Semi_t
{
   top.errors = tm.errors;
   top.typeErrors = tm.typeErrors;
}
 

-- Aspects on core Silver terminal modifier productions
aspect production terminalModifierSingle
top::TerminalModifiers ::= tm::TerminalModifier
{
     top.hasLexerClass = tm.hasLexerClass;
     top.lexerClass = tm.lexerClass;

     top.hasStaticPrec = tm.hasStaticPrec;
     top.staticPrec = tm.staticPrec;

     top.hasDynamicPrec = tm.hasDynamicPrec;
     top.dynamicPrec = tm.dynamicPrec;     

     top.usePrefix = tm.usePrefix;

     top.hasLexerClasses = tm.hasLexerClasses ;
     top.lexerClasses = tm.lexerClasses;

     top.hasDominates = tm.hasDominates ;
     top.dominates = tm.dominates ;
     top.submitsTo = tm.submitsTo ;
     top.hasSubmitsTo = tm.hasSubmitsTo ;

     top.errors = tm.errors;
     top.typeErrors = tm.typeErrors;

     top.hasActionCode = tm.hasActionCode;
     top.actionCode = tm.actionCode;
}

aspect production terminalModifiersCons
top::TerminalModifiers ::= h::TerminalModifier c::Comma_t t::TerminalModifiers
{
     top.hasLexerClass = h.hasLexerClass || t.hasLexerClass;
     top.lexerClass = if   h.hasLexerClass 
                      then h.lexerClass 
                      else if   t.hasLexerClass
                           then t.lexerClass 
                           else "";
     
     top.hasStaticPrec = h.hasStaticPrec || t.hasStaticPrec;
     top.staticPrec = if   h.hasStaticPrec 
                      then h.staticPrec 
                      else if t.hasStaticPrec
                           then t.staticPrec 
                           else 0;

     top.hasDynamicPrec = h.hasDynamicPrec || t.hasDynamicPrec;
     top.dynamicPrec = if   h.hasDynamicPrec 
                       then h.dynamicPrec 
                       else if   t.hasDynamicPrec
                            then t.dynamicPrec 
                            else 0;

     top.usePrefix = h.usePrefix || t.usePrefix;

     top.hasLexerClasses = h.hasLexerClasses || t.hasLexerClasses ;
     top.lexerClasses = h.lexerClasses +++ t.lexerClasses ;

     top.hasDominates = h.hasDominates || t.hasDominates ;
     top.dominates = h.dominates +++ t.dominates ;
     top.hasSubmitsTo = h.hasSubmitsTo || t.hasSubmitsTo ;
     top.submitsTo = h.submitsTo +++ t.submitsTo ;

     top.errors = h.errors ++ t.errors;
     top.typeErrors = h.typeErrors ++ t.typeErrors;

     top.hasActionCode = h.hasActionCode || t.hasActionCode;
     top.actionCode = if h.hasActionCode
                      then h.actionCode
                      else if t.hasActionCode
                           then t.actionCode
                           else actionCodeEmpty();
}


aspect production terminalModifierDefaults
top::TerminalModifier ::= 
{
  top.hasLexerClass = false;
  top.lexerClass = "";

  top.hasStaticPrec = false;
  top.staticPrec = 0;

  top.hasDynamicPrec = false;
  top.dynamicPrec = 0;

  top.hasLexerClasses = false ;
  top.lexerClasses = [] ;

  top.usePrefix = false;

  top.hasDominates = false;
  top.dominates = [] ;
  top.hasSubmitsTo = false;
  top.submitsTo = [];

  top.errors = "";
  top.typeErrors = "";

  top.hasActionCode = false;
  top.actionCode = actionCodeEmpty();
}

concrete production terminalModifierOperatorPrecedence
top::TerminalModifier ::= o::Operator_kwd p::Precedence_kwd e::Equal_t i::Int_t
{
  top.pp = "operator precedence = " ++ i.lexeme;

  top.hasParserPrec = true;
  top.parserPrec = toInt(i.lexeme);

  forwards to terminalModifierDefaults();
}

concrete production terminalModifierLexerClass
top::TerminalModifier ::= l::Lexer_kwd c::Class_kwd e::Equal_t i::Id_t
{
  top.pp = "lexer class = " ++ i.lexeme;

  top.hasLexerClass = true;
  top.lexerClass = i.lexeme;

  forwards to terminalModifierDefaults();
}

concrete production terminalModifierStaticPrecedence
top::TerminalModifier ::= s::Static_kwd p::Precedence_kwd e::Equal_t i::Int_t
{
  top.pp = "static precedence = " ++ i.lexeme;

  top.hasStaticPrec = true;
  top.staticPrec = toInt(i.lexeme);

  forwards to terminalModifierDefaults();
}

concrete production terminalModifierDynamicPrecedence
top::TerminalModifier ::= d::Dynamic_kwd p::Precedence_kwd e::Equal_t i::Int_t
{
  top.pp = "dynamic precedence = " ++ i.lexeme;

  top.hasDynamicPrec = true;
  top.dynamicPrec = toInt(i.lexeme);  

  forwards to terminalModifierDefaults();
}

concrete production terminalModifierDoNotUsePrefix
top::TerminalModifier ::= p::Prefix_kwd e::Equal_t f::False_kwd
{
  top.pp = "prefix = false";

  top.usePrefix = false;

  forwards to terminalModifierDefaults();
}

concrete production terminalModifierUsePrefix
top::TerminalModifier ::= p::Prefix_kwd e::Equal_t t::True_kwd
{
  top.pp = "prefix = true";

  top.usePrefix = true;

  forwards to terminalModifierDefaults();
}

-- Concrete Syntax for new Copper --


concrete production terminalModifierDominates
tm::TerminalModifier ::= d_kwd::Dominates_t lc::LCurly_t terms::TermPrecList  rc::RCurly_t
{
 tm.pp = d_kwd.lexeme ++ " { " ++ terms.pp ++ " } " ;

 tm.hasDominates = true ;
 tm.dominates = terms.precTermList ;

 tm.errors = terms.errors;
 tm.typeErrors = terms.typeErrors;

 forwards to terminalModifierDefaults();
}

concrete production terminalModifierSubmitsTo
tm::TerminalModifier ::= s_kwd::SubmitsTo_t lc::LCurly_t terms::TermPrecList  rc::RCurly_t
{
 tm.pp = s_kwd.lexeme ++ " { " ++ terms.pp ++ " } " ;

 tm.hasSubmitsTo = true ;
 tm.submitsTo = terms.precTermList ;

 tm.errors = terms.errors;
 tm.typeErrors = terms.typeErrors;

 forwards to terminalModifierDefaults();
}

nonterminal TermPrecList with {pp,location} ;

abstract production termPrecList
terms::TermPrecList ::= t::QName terms_tail::TermPrecList
{
  terms.pp = if terms_tail.pp == ""
             then t.pp
             else t.pp ++ ", " ++ terms_tail.pp;
  terms.location = t.location;

 production attribute fullNameItem :: [Decorated EnvItem];
 fullNameItem = getFullNameDclOne(t.name,terms.env);

 production attribute typeItem :: [Decorated EnvItem];
 typeItem = getTypeDclOne(t.name,terms.env);

 terms.precTermList = if null(fullNameItem)
                      then terms_tail.precTermList
                      else [head(fullNameItem).fullName] +++ terms_tail.precTermList ;

 terms.defs = addValueDcl(head(fullNameItem).fullName, termTypeRep(t.name), 
	      addFullNameDcl(t.name,head(fullNameItem).fullName,
              terms_tail.defs));
--     	      addOccursDcl("lexeme", t.name,
-- 	      addOccursDcl("line", t.name,
--  	      addOccursDcl("column", t.name, terms_tail.defs)))));

 terms.errors = if null(fullNameItem)
                then t.location ++ " Error: Name '" ++ t.name ++ "' is not declared.\n" ++ terms_tail.errors
                else terms_tail.errors;

-- Does not seem to be any type-checking on this sort of thing.
 terms.typeErrors = --if null(typeItem) || (!head(typeItem).typerep.isTerminal && !head(typeItem).typerep.isLexerClass)
                     --then t.location ++ " Error: '" ++ t.name ++ "' is not a terminal or a lexer class.\n" ++ terms_tail.typeErrors
                     --else
                     terms_tail.typeErrors ;
  
}

function addTerminalAttrDefs
Decorated Defs ::= tailDefs::Decorated Defs
{
   return addFullNameDcl("lexeme","lexeme",
           addValueDcl("lexeme",stringTypeRep(),
            addFullNameDcl("line","line",
             addValueDcl("line",integerTypeRep(),
              addFullNameDcl("column","column",
               addValueDcl("column",integerTypeRep(),
                tailDefs))))));
}

concrete production terminalModifierActionCode
top::TerminalModifier ::= act::Action_kwd acode::ActionCode_c
{
   top.pp = "action " ++ acode.actionCode.pp;

   top.hasActionCode = true;
   top.actionCode = acode.actionCode;

   acode.actionCodeType = terminalActionType();
   acode.env = appendDefsEnv(addTerminalAttrDefs(acode.defs),top.env);

   acode.signatureEnv = emptyEnv();
   acode.localsEnv = toEnv(acode.defs);

--   acode.signatureNames = [];
--   acode.signatureTypes = [];
--appendDefsEnv(addFullNameDcl("lexeme","lexeme",addValueDcl("lexeme",stringTypeRep(),acode.defs)),top.env);

   forwards to terminalModifierDefaults();
}


abstract production termPrecListNull
terms::TermPrecList ::=
{
   terms.precTermList = [];
   terms.defs = emptyDefs();
   terms.pp = "";
   terms.location = "";
   terms.errors = "";
   terms.typeErrors = "";
}

concrete production termPrecListOne
terms::TermPrecList ::= t::QName
{
   forwards to termPrecList(t'',termPrecListNull());
}

concrete production termPrecListCons
terms::TermPrecList ::= t::QName c::Comma_t terms_tail::TermPrecList
{
   forwards to termPrecList(t'',terms_tail'');
}

concrete production terminalModifierClassSpec
top::TerminalModifier ::= l::Lexer_kwd c::Classes_kwd e::Equal_t lc::LCurly_t classes::ClassList  rc::RCurly_t
{
 top.pp = "lexer classes = { " ++ classes.pp ++ " } " ;

 top.hasLexerClasses = true;
 top.lexerClasses = classes.lexerClasses;

 top.errors = classes.errors;

 forwards to terminalModifierDefaults();
}

nonterminal ClassList with pp, hasLexerClasses, lexerClasses ;

abstract production lexerClassesMain
cl::ClassList ::= n::QName cl_tail::ClassList
{
 cl.pp = if cl_tail.pp == ""
         then n.pp
         else n.pp ++ ", " ++ cl_tail.pp;

 local attribute fullNameItem :: [Decorated EnvItem];
 fullNameItem = getFullNameDclOne(n.name,cl.env);

 local attribute typeItem :: [Decorated EnvItem];
 typeItem = getTypeDclOne(n.name,cl.env);

 cl.lexerClasses = if null(fullNameItem)
                   then cl_tail.lexerClasses
                   else [head(fullNameItem).fullName] +++ cl_tail.lexerClasses;

 cl.errors = if null(fullNameItem)
             then n.location ++ " Error: Name '" ++ n.name ++ "' is not declared.\n" ++ cl_tail.errors
             else cl_tail.errors;

-- Does not appear to be any type-checking on this sort of thing.
 cl.typeErrors = --if null(typeItem) || !head(typeItem).typerep.isLexerClass
--                  then n.location ++ " Error: '" ++ n.name ++ "' is not a lexer class.\n" ++ cl_tail.typeErrors
                 -- else 
                    cl_tail.typeErrors;
}

abstract production lexerClassesNull
cl::ClassList ::=
{
  cl.pp = "";
  cl.errors = "";
  cl.typeErrors = "";
  cl.lexerClasses = [];
}

concrete production lexerClassesOne
cl::ClassList ::= n::QName
{
  forwards to lexerClassesMain(n'',lexerClassesNull());
}

concrete production lexerClassesCons
cl::ClassList ::= n::QName c::Comma_t cl_tail::ClassList
{
  forwards to lexerClassesMain(n'',cl_tail'');
}

