grammar silver:translation:haskelljava:concretesyntax;
import silver:base:core;
import silver:base:cs:default;

synthesized attribute hasGrammarNameSyn :: Boolean;

nonterminal GrammarInfo with {hasGrammarNameSyn,grammarNameSyn};

abstract production defaultGrammarInfo
top::GrammarInfo ::=
{
    top.hasGrammarNameSyn = false;
    top.grammarNameSyn = decorate makeQName("") with {};
}

abstract production grammarInfoName
top::GrammarInfo ::= gName::Decorated QName
{
    top.hasGrammarNameSyn = true;
    top.grammarNameSyn = gName;
    forwards to defaultGrammarInfo();
}

abstract production mashGrammarInfos
top::GrammarInfo ::= gi1::GrammarInfo gi2::GrammarInfo
{
    top.hasGrammarNameSyn = if gi1.hasGrammarNameSyn && gi2.hasGrammarNameSyn
                            then error("Attempting to mash GrammarInfo objects each defining their own grammar name")
                            else gi1.hasGrammarNameSyn || gi2.hasGrammarNameSyn;

    top.grammarNameSyn = if gi1.hasGrammarNameSyn
                         then gi1.grammarNameSyn
                         else if gi2.hasGrammarNameSyn
                              then gi2.grammarNameSyn
                              else decorate makeQName("") with {};
}

synthesized attribute grammarInfo :: GrammarInfo;
synthesized attribute grammarsInfo :: [GrammarInfo];

attribute grammarsInfo occurs on Root,RootSpec,ImportExportStmts,SyntaxStmt;
attribute grammarInfo occurs on GrammarDcl,AGDcls,AGDcl;

aspect production root
top::Root ::= gdcl::GrammarDcl ims::ImportExportStmts  ags::AGDcls
{
   top.grammarsInfo = [mashGrammarInfos(gdcl.grammarInfo,ags.grammarInfo)] +++
                      ims.grammarsInfo;
}

aspect production emptyRootSpec
top::RootSpec ::= 
{
   top.grammarsInfo = [];
}

aspect production consRootSpec
top::RootSpec ::= c1::Decorated Root c2::Decorated RootSpec
{
   top.grammarsInfo = c1.grammarsInfo +++ c2.grammarsInfo;
}

aspect production appendRootSpec
top::RootSpec ::= c1::Decorated RootSpec c2::Decorated RootSpec
{
   top.grammarsInfo = c1.grammarsInfo +++ c2.grammarsInfo;
}


aspect production grammarDcl
top::GrammarDcl ::= g::Grammar_kwd  qn::QName  s::Semi_t
{
   top.grammarInfo = grammarInfoName(qn);
}

aspect production importStmtsNone
top::ImportExportStmts ::=
{
   top.grammarsInfo = [];
}

aspect production importStmtsOne 
top::ImportExportStmts ::= im::ImportStmt
{
   top.grammarsInfo = [];
}

aspect production exportStmtsOne 
top::ImportExportStmts ::= ex::ExportStmt
{
   top.grammarsInfo = [];
}

aspect production importStmtsCons
top::ImportExportStmts ::= h::ImportStmt t::ImportExportStmts
{
   top.grammarsInfo = t.grammarsInfo;
}

aspect production exportStmtsCons
top::ImportExportStmts ::= h::ExportStmt t::ImportExportStmts
{
   top.grammarsInfo = t.grammarsInfo;
}

aspect production appendImportsExports
top::ImportExportStmts ::= fst::ImportExportStmts snd::ImportExportStmts
{
   top.grammarsInfo = fst.grammarsInfo +++ snd.grammarsInfo;
}

aspect production syntaxStmtsOne
top::ImportExportStmts ::= syn::SyntaxStmt
{
   top.grammarsInfo = syn.grammarsInfo;
}

aspect production syntaxStmtsCons
top::ImportExportStmts ::= h::SyntaxStmt t::ImportExportStmts
{
   top.grammarsInfo = h.grammarsInfo +++ t.grammarsInfo;
}

aspect production agDclsOne
top::AGDcls ::= ag::AGDcl
{
   top.grammarInfo = ag.grammarInfo;
}

aspect production agDclsCons
top::AGDcls ::= h::AGDcl t::AGDcls
{
   top.grammarInfo = mashGrammarInfos(h.grammarInfo,t.grammarInfo);
}

aspect production agDclSeq
top::AGDcl ::= ag1::AGDcl ag2::AGDcl
{
   top.grammarInfo = mashGrammarInfos(ag1.grammarInfo,ag2.grammarInfo);
}

aspect production agDclNone
top::AGDcl ::=
{
   top.grammarInfo = defaultGrammarInfo();
}

aspect production attributionDcl
top::AGDcl ::= at::Attribute_kwd a::QName oc::Occurs_kwd o::On_kwd nt::QName s::Semi_t
{
  top.grammarInfo = defaultGrammarInfo();
}

aspect production productionDcl
 top::AGDcl ::=
   a::Abstract_kwd p::Production_kwd id::Id_t ns::ProductionSignature
   body::ProductionBody
{
  top.grammarInfo = defaultGrammarInfo();
}

aspect production aspectAGDcl
 top::AGDcl ::=
   a::Aspect_kwd p::Production_kwd id::QName ns::AspectSignature
   body::ProductionBody
{
  top.grammarInfo = defaultGrammarInfo();
}

aspect production attributeDclInh
top::AGDcl ::= i::Inherited_kwd  at::Attribute_kwd a::Id_t ht::HasType_t te::Type s::Semi_t
{
  top.grammarInfo = defaultGrammarInfo();
}

aspect production attributeDclSyn
top::AGDcl ::= sn::Synthesized_kwd  at::Attribute_kwd a::Id_t ht::HasType_t te::Type  s::Semi_t
{
  top.grammarInfo = defaultGrammarInfo();
}

aspect production terminalDcl
top::AGDcl ::= t::Terminal_kwd id::Id_t r::RegExpr s::Semi_t
{
  top.grammarInfo = defaultGrammarInfo();
}

aspect production disambiguationGroupDcl
top::AGDcl ::= 'disambiguation' 'group' id::Id_t w::With_kwd lc::LCurly_t terms::TermPrecList  rc::RCurly_t act::Action_kwd acode::ActionCode_c s::Semi_t
{
  top.grammarInfo = defaultGrammarInfo();
}

aspect production lexerClassDecl
top::AGDcl ::= l::Lexer_kwd cl::Class_kwd id::Id_t s::Semi_t
{
  top.grammarInfo = defaultGrammarInfo();
} 

aspect production attributeDclParser
top::AGDcl ::= pr::Parser_kwd at::Attribute_kwd a::Id_t ht::HasType_t te::Type act::Action_kwd acode::ActionCode_c s::Semi_t
{
  top.grammarInfo = defaultGrammarInfo();
} 

aspect production defaultNonterminalDcl
top::AGDcl ::= id::Id_t
{
  top.grammarInfo = defaultGrammarInfo();
}

aspect production defaultClosedNonterminalDcl
top::AGDcl ::= id::Id_t
{
  top.grammarInfo = defaultGrammarInfo();
}

aspect production closeNonterminalDcl
top::AGDcl ::= c::Close_kwd q::QName  s::Semi_t
{
  top.grammarInfo = defaultGrammarInfo();
}

aspect production syntaxAll
top::SyntaxStmt ::= s::Syntax_kwd pkg::QName se::Semi_t
{
   production attribute gInfo :: [GrammarInfo] with ++;
   gInfo := [];

   gInfo <- spec.grammarsInfo;

   top.grammarsInfo = gInfo;
}

aspect production syntaxHiding
top::SyntaxStmt ::= s::Syntax_kwd  pkg::QName h::Hiding_kwd ns::NameList se::Semi_t
{
   production attribute gInfo :: [GrammarInfo] with ++;
   gInfo := [];

   gInfo <- spec.grammarsInfo;

   top.grammarsInfo = gInfo;
}
