grammar silver:translation:haskelljava:copper;
import silver:translation:haskelljava:core;
import silver:base;
import silver:utils;


function makeBridge
String ::= gname::String spec::Decorated RootSpec topName::String
{  
  local attribute s :: String;
  s = "NT_" ++ makeName(spec.startName);

  local attribute filtered :: FilterResult;
  filtered = filterSyntax(spec.hiddenSyntax, spec.terminalDcls, spec.ruleDcls);

  local attribute merged :: [RuleSpec];
  merged = mergeRules(filtered.ruleDcls);

  return
	"module Module_" ++ makeName(gname) ++ "_Parser where\n" ++
	"import Terminal\n" ++
	"import Directory\n" ++
	"import System\n" ++
        "import Module_" ++ makeName(topName) ++ "\n" ++
	"import Control.Monad\n" ++
	"import System.IO.Unsafe\n\n" ++

	makeSpecBridge(merged, filtered.terminalDcls) ++ "\n" ++ 
	makeSpecFunction(merged, filtered.terminalDcls) ++ "\n\n" ++

	"rTmp :: String -> Bridge_" ++ makeName(spec.startName) ++ "\n" ++
 	"rTmp = read\n\n" ++

 	"doparse :: String -> (INHS_" ++ s ++ " -> " ++ s ++ ")\n" ++
	"doparse stIn = unsafePerformIO (parse_help stIn)\n\n" ++

        "parse_help :: String -> IO (INHS_" ++ s ++ " -> " ++ s ++ ")\n" ++
        "parse_help stIn =\n" ++
        " if file_name == \"\"\n" ++
        " then do\n" ++
        "      System.system (\"arch > theArch\")\n" ++
        "      arch <- readFile \"theArch\"\n" ++
        "      p <- findParser \"" ++ gname ++ "\" (\".parsers/\" ++ (stripWS arch) ++ \"/" ++ makeName(gname) ++ "_parser\")\n" ++
        "      writeFile out_name stIn\n" ++
        "      System.system (\"cat \" ++ out_name ++ \" | \" ++ p ++ \" > \" ++  cached_file_name )\n" ++
        "      textIn <- readFile cached_file_name\n" ++
        "      System.system (\"rm -rf \" ++ out_name )\n" ++
        "      System.system (\"rm -rf theArch\" )\n" ++
        "      System.system (\"rm -rf \" ++ cached_file_name )\n" ++
        "      return (r_" ++ makeName(spec.startName) ++ "(rTmp textIn))\n" ++
        " else do\n" ++
        "      cached_file_exists <- doesFileExist cached_file_name\n" ++
        "      when (not cached_file_exists)\n" ++
        "	       (do\n" ++
        "		System.system (\"arch > theArch\")\n" ++
        "		arch <- readFile \"theArch\"\n" ++
        "		p <- findParser \"" ++ gname ++ "\" (\".parsers/\" ++ (stripWS arch) ++ \"/" ++ makeName(gname) ++ "_parser\")\n" ++
        "		System.system (\"cat \" ++ file_name ++ \" | \" ++ p ++ \" > \" ++  cached_file_name )\n" ++
        "		temp <- ( System.system \"rm -rf theArch\" )\n" ++
        "	        return ())\n" ++
        "      textIn <- readFile cached_file_name\n" ++
        "      return (r_" ++ makeName(spec.startName) ++ "(rTmp textIn))\n" ++
        "	  where\n" ++
        "	  (file_name, file_text) = separate (stIn)\n" ++
        "	  cached_file_name = file_name ++ \".copper\"\n" ++
        "	  out_name = \"__OUT\" ++ show (genint ())\n" ++
        "	  separate s = if header == \"***\" && f_text_header == \"***\"\n" ++
        "		       then (f_name, f_text)\n" ++
        "		       else (\"\",s)\n" ++
        "	      where\n" ++
        "	      header = take 3 s\n" ++
        "	      header_rest = drop 3 s\n" ++
        "	      f_name = takeWhile (/='*') header_rest\n" ++
        "	      f_name_rest = dropWhile (/='*') header_rest\n" ++
        "	      f_text_header = take 3 f_name_rest\n" ++
        "	      f_text = drop 3 f_name_rest\n\n" ++

	"findParser g parserLoc = do\n" ++ 
	"			 grammar <- return (makeName g)\n" ++ 
	"			 gpath <- System.getEnv (\"GRAMMAR_PATH\")\n" ++ 
	"			 paths <- return (split \"\" gpath)\n" ++ 
	"			 locs <- return (map (\\p -> p ++ \"/\" ++ grammar ++ \"/\" ++ parserLoc) paths)\n" ++ 
	"			 p <- checkPath(locs)\n" ++ 
	"			 case p of {[] -> error (\"Parser '\" ++ parserLoc ++ \"' for grammar '\" ++ g ++ \"' cannot be found.\");" ++ 
					   " _ -> return p}\n\n" ++ 		

 	"split :: [Char] -> [Char] -> [[Char]]\n" ++ 
	"split current [] = [removeSlash current]\n" ++ 
	"split current (':':str) = [removeSlash current] ++ (split \"\" str)\n" ++ 
	"split current (a:str) = split (current ++ [a]) str\n\n" ++ 

	"makeName [] = \"\"\n" ++ 
	"makeName (':':str) = \"/\" ++ makeName(str)\n" ++ 
	"makeName (a:str) = [a] ++ makeName(str)\n\n" ++ 

	"stripWS [] = \"\"\n" ++ 
	"stripWS ('\\n':str) = stripWS(str)\n" ++ 
	"stripWS (a:str) = [a] ++ stripWS(str)\n\n" ++ 

	"removeSlash [] = []\n" ++ 
	"removeSlash ['/'] = []\n" ++ 
	"removeSlash (a:str) = [a] ++ (removeSlash str)\n\n" ++ 

	"checkPath [] = return []\n" ++ 
	"checkPath (p:rest) = do\n" ++ 
	"		     e <- doesFileExist p\n" ++ 
	"		     if e then return p else (checkPath rest);\n\n" ++ 

	"data Bridge_TERMINAL = Bridge_TERMINAL1 String Int Int deriving (Show, Read)\n\n"	++

	"getTerminalLexeme (Bridge_TERMINAL1 lex l c) = lex\n\n"  ++

	"getTerminalLine (Bridge_TERMINAL1 lex l c) = l\n\n" ++	

	"getTerminalColumn (Bridge_TERMINAL1 lex l c) = c\n\n";
}

function makeSpecBridge
String ::= specs::[RuleSpec] t::[TerminalSpec]
{
  return if null(specs)
	 then ""
	 else "data Bridge_" ++ makeName(head(specs).ruleLHS) ++ " = " ++ makeSpecBridgeHelp(head(specs).ruleRHSSpec, t) ++ 
	      "\n\tderiving (Show, Read)\n" ++ 
	      makeSpecBridge(tail(specs), t); 
}

function makeSpecBridgeHelp
String ::= specs::[RHSSpec] t::[TerminalSpec]
{
  return if null(specs)
	 then ""
	 else "Production_" ++ makeName(head(specs).ruleName) ++ " " ++  makeSpecRHS(head(specs).ruleRHS, t) ++
              (if null(tail(specs)) then "" else "\n\t| ") ++
	      makeSpecBridgeHelp(tail(specs), t);
}

function makeSpecRHS
String ::= rhs::[String] t::[TerminalSpec]
{
  local attribute hs :: String;
  hs = 	if terminalCheck(head(rhs), t)
	then "Bridge_TERMINAL"
	else "Bridge_" ++ makeName(head(rhs));

  return if null(rhs) then "" 
	 else hs ++ " " ++ makeSpecRHS(tail(rhs), t);
}


function makeSpecFunction
String ::= specs::[RuleSpec] t::[TerminalSpec]
{
  return if null(specs)
	 then ""
	 else makeSpecFunctionHelp(head(specs).ruleLHS, head(specs).ruleRHSSpec, t) ++ makeSpecFunction(tail(specs), t); 
}

function makeSpecFunctionHelp
String ::= lhs::String specs::[RHSSpec] t::[TerminalSpec]
{
  return if null(specs)
	 then ""
	 else "r_" ++ makeName(lhs) ++ " (Production_" ++ makeName(head(specs).ruleName) ++ " " ++ 
			makeSig(length(head(specs).ruleRHS), 0) ++ ") = " 
	     ++ "production_" ++ makeName(head(specs).ruleName) ++ " " ++ makeSig2(head(specs).ruleRHS, 0, t) ++ "\n"
	     ++ makeSpecFunctionHelp(lhs, tail(specs), t); 
}

function makeSig
String ::= i1::Integer i2::Integer
{
  return if i1 == 0 then "" else "p" ++ toString(i2) ++ " " ++ makeSig(i1-1, i2+1);
}

function makeSig2
String ::= l::[String] i::Integer t::[TerminalSpec]
{
  return if null(l) 
	 then "" 
	 else (if terminalCheck(head(l), t)
	       then "(TERMINAL { syns_TERMINAL = SYNS_TERMINAL { attr_TERMINAL__lexeme = (getTerminalLexeme p" ++ toString(i) ++ "), " ++ 
								"attr_TERMINAL__line = (getTerminalLine p" ++ toString(i) ++ "), " ++ 
								"attr_TERMINAL__column = (getTerminalColumn p" ++ toString(i) ++ ") }  }) "
	       else "(r_" ++ makeName(head(l)) ++ " p" ++ toString(i) ++ ") ") ++ 
	      makeSig2(tail(l), i+1, t);
}