grammar silver:translation:haskelljava:copper;
import silver:translation:haskelljava:command;
import silver:translation:haskelljava:core;
import silver:translation:ag:haskell:core;

import silver:base hiding grammarName;

import core;
import silver:utils;
import silver:utils:command;

aspect production run
top::RunUnit ::= iIn::IO args::String p::Production(Root ::= String)
{
  postOps <- if a.noCSGen then [] else [generateCS(a'', grammars)]; 
}

abstract production generateCS
top::Unit ::= a::Command specs::[GrammarSpec]
{
  forwards to generateCSHelp(a'', specs)
	with {
		ioIn = print("Generating Parsers and Scanners.\n", top.ioIn);
	};
}

abstract production generateCSHelp
top::Unit ::= a::Command specs::[GrammarSpec]
{
  top.io = writeAllCS(top.ioIn, a, specs, mergeParsers(specs));
  top.code = 0;
  top.order = 2;
}

function writeAllCS
IO ::= i::IO a::Decorated Command l::[GrammarSpec] s::[String]
{
  local attribute now :: IO;
  now = writeCSSpec(i, head(l).name, head(l).rootSpec, a, s);

  local attribute recurse :: IO;
  recurse = writeAllCS(now, a, tail(l), s);

  return if null(l) then i else recurse;
}


function writeCSSpec
IO ::= i::IO gName::String r::Decorated RootSpec a::Decorated Command s::[String]
{
  --the top level grammar name
  local attribute thn :: String;
  thn = makeName(a.grammarName);

  -- the haskell name of the grammar.
  local attribute hn :: String;
  hn = makeName(gName);
  
  local attribute has_start :: Boolean;
  has_start = r.hasStartSymbol && containsString(gName ++ ":parse", s);

  local attribute shn :: String;
  shn = "NT_" ++ makeName(r.startName);

  local attribute parser_boot :: String;
  parser_boot = 
      "module Module_" ++ hn ++ "_Parser where\n" ++
      "import {-# SOURCE #-} Module_" ++ thn ++ "\n" ++
      "doparse :: String ->  (INHS_" ++ shn ++ " -> " ++ shn ++ ")\n";

  local attribute copperFile :: String;
  copperFile = a.generatedPath ++ "Module_" ++ hn ++ "_Grammar.copper";

  local attribute copperFileTemp :: String;
  copperFileTemp = a.generatedPath ++ "Module_" ++ hn ++ "_copper.temp";

  local attribute existsCopper :: IOBoolean;
  existsCopper = isFile(copperFile, i);

  local attribute cpCopper :: IOInteger;
  cpCopper = if existsCopper.bValue then system("cp " ++ copperFile ++ " " ++ copperFileTemp, existsCopper.io) else ioInteger(existsCopper.io, 0);

  local attribute prevCopper :: IOString;
  prevCopper = if existsCopper.bValue then readFile(copperFileTemp, cpCopper.io) else ioString(cpCopper.io, "");

  local attribute newCopper :: String;
  newCopper = "grammar name Module_" ++ hn ++"_Grammar spectype LALR1 1.1\n\n" ++ 
          makeCopperGrammarSpec(hn, r, a.grammarName);

  local attribute needsCopper ::Boolean;
  needsCopper = prevCopper.sValue != newCopper;

  local attribute io4 :: IO;
  io4 = if !has_start || !needsCopper || a.onlyCSGen then prevCopper.io
	else writeFile(a.generatedPath ++ "Module_" ++ hn ++ "_Parser.hs-boot", parser_boot, prevCopper.io);

  local attribute io5 :: IO;
  io5 = if (!a.onlyCSGen && (!has_start || !needsCopper)) ||
           (a.onlyCSGen && (!r.hasStartSymbol && hn != thn))  then io4
	else writeFile(copperFile, newCopper, io4);

  local attribute java_gen_command :: String;
  java_gen_command = "java " ++ a.jvmParameters ++ " -classpath " ++
                    a.javaClassPath ++ ":" ++ a.generatedPath ++
                    " edu.umn.cs.melt.copper.main.ParserCompiler NONE Module_" ++ hn ++ "_Grammar " ++ copperFile ++ " > " ++ a.generatedPath ++ "Module_" ++ hn ++ "_Grammar.java";

  local attribute io6 :: IO;
  io6 = if !has_start || !needsCopper || a.onlyCSGen then io5
	else system(java_gen_command, io5).io;

  local attribute io7 :: IO;
  io7 = if !has_start || !needsCopper || a.onlyCSGen then io6
	else writeFile(a.generatedPath ++ "Module_" ++ hn ++ "_Parser.hs", makeBridge(gName, r, a.grammarName), io6);

  return io7;
}
