grammar silver:translation:haskelljava:copper;
import silver:translation:haskelljava:concretesyntax;
import silver:base;
import silver:base:core;
import silver:utils;

-- The root-function: makes the complete grammar specification
-- for the quasi-scannerless LALR(1) system.
function makeCopperGrammarSpec
String ::= grammar_name::String spec::Decorated RootSpec topName::String 
{
  local attribute s :: String;
  s = makeName(spec.startName);

  local attribute filtered :: FilterResult;
  filtered = filterSyntax(spec.hiddenSyntax, spec.terminalDcls, spec.ruleDcls);

  local attribute precsRules :: [ParserPrecSpec];
  precsRules = getParserPrecSpecRules(filtered.ruleDcls);

  local attribute merged :: [RuleSpec];
  merged = mergeRules(filtered.ruleDcls);

  local attribute univLayout :: String;
  univLayout = generateUniversalLayout(grammar_name,spec.terminalDcls) ++ " EmptyString";

  local attribute rv :: String;

  rv =       
--             printGrammarInfoList(spec.grammarsInfo) ++ "\n" ++
             makePrefixSpecString(filtered.terminalDcls) ++
             makeParserAttrSpecString(spec.parserAttrDcls) ++ "\n" ++
             makeDisambigGroupSpecString(spec.disambiguationGroupDcls) ++ "\n" ++
             makeTermTokenSpecString(filtered.terminalDcls,grammar_name) ++
             makeStartDclString(grammar_name,filterInStart(spec.startName,spec.ntDcls),univLayout) ++ "\n" ++
             makeNonTermDclString(filterOutStart(spec.startName,spec.ntDcls)) ++ "\n" ++
             makePrecSpecCopperSpec(filtered.terminalDcls) ++ "\n\n" ++
             makeProdDclString(spec,filtered.terminalDcls,filtered.ruleDcls,univLayout) ++
    "";

  return rv;
}

--function printGrammarInfoList
--String ::= grammars::[GrammarInfo]
--{
--   return if null(grammars)
--          then ""
--          else "# " ++ head(grammars).grammarNameSyn.name ++ ": " ++ toString(head(grammars).hasGrammarNameSyn) ++ " " ++ toString(head(grammars).hasCustomLayout) ++ "\n" ++ printGrammarInfoList(tail(grammars));
--}

function printPrecList
String ::= precs::[String]
{
    return if null(precs)
	   then ""
	   else makeName(head(precs)) ++ " " ++ printPrecList(tail(precs));
}

function printClassList
String ::= classes::[String]
{
    return if null(classes)
	   then ""
	   else makeName(head(classes)) ++ " " ++ printClassList(tail(classes));
}

function makePrefixSpecString
String ::= specs::[TerminalSpec]
{
    return if null(specs)
           then ""
           else if head(specs).usePrefix
                then "term prefix_" ++ makeName(head(specs).prefixGrammar) ++" ::" ++ head(specs).prefixGrammar ++ ":\n" ++
              "token prefix_" ++ makeName(head(specs).prefixGrammar) ++ " class { _prefix } precedence submits to { } dominates { } prefix {} code @" ++ head(specs).actionCode.aCode ++ "@\n\n" ++
                      makePrefixSpecString(tail(specs))
                else makePrefixSpecString(tail(specs));
}

function makeParserAttrSpecString
String ::= specs::[ParserAttrSpec]
{
    return if null(specs) || !head(specs).attrType.typerep.isParserAttrType
           then ""
           else "attribute " ++ makeName(head(specs).attrName) ++
                " type " ++ head(specs).attrType.typerep.parserAttrType ++
                " code @" ++ head(specs).actionCode.aCode ++ "@\n" ++
                makeParserAttrSpecString(tail(specs));
}

function makeDisambigGroupSpecString
String ::= specs::[DisambiguationGroupSpec]
{
   return if null(specs)
          then ""
          else "ambiguous term group " ++ makeName(head(specs).groupName)
               ++ " code @" ++ head(specs).actionCode.aCode ++ "\n@ members "
               ++ printPrecList(head(specs).groupMembers) ++ "\n" ++
               makeDisambigGroupSpecString(tail(specs));
}
               
-- Concatenates terminal specification (name and regex) and token specification
-- (precedence and prefix) for a given terminal.
function makeTermTokenSpecString
String ::= specs::[TerminalSpec] grammar_name::String
--String ::= specs::[TerminalSpec] flags::EnvFlags grammar_name::String
{
  return if null(specs)
	 then ""
	 else "term " ++ makeName(head(specs).terminalName) ++ " :" ++
              substring(1, length(head(specs).terminalRegExpr) -1, head(specs).terminalRegExpr) ++
              "\n" ++
              token_dcl_v2 ++
              makeTermTokenSpecString(tail(specs),grammar_name);

  local attribute token_dcl_v2 :: String ;  -- for use with new higher|lower precedence lists
  token_dcl_v2 = "token " ++ makeName(head(specs).terminalName) ++ 
                 " class { " ++ printClassList(head(specs).lexerClasses) ++ " } " ++
                 " precedence submits to { " ++ printPrecList(head(specs).submitsTo) ++ " } " ++
                 " dominates { " ++ printPrecList(head(specs).dominates) ++ " } " ++
                 " prefix { " ++
                 (if head(specs).usePrefix then " prefix_" ++ makeName(head(specs).prefixGrammar) ++ " " else "") ++ "} code @" ++ head(specs).actionCode.aCode ++ "@\n\n" ;  

  local attribute token_dcl_v1 :: String ;  -- for use with previous integer based precedence settings
  token_dcl_v1 = if !head(specs).isTokenTerminal
                 then "token " ++ makeName(head(specs).terminalName) ++ 
                      " class " ++ head(specs).lexerClass ++
                      " precedence static 0 dynamic 0 bridging 0 prefix {" ++ 
                      (if head(specs).usePrefix then " prefix_" ++ grammar_name ++ " " else "") ++ "} code @@\n\n"

                 else "token " ++ makeName(head(specs).terminalName) ++ 
                      " class " ++ head(specs).lexerClass ++ 
                      " precedence static " ++ toString(head(specs).staticPrec) ++ 
                      " dynamic " ++ toString(head(specs).dynamicPrec) ++
                      " bridging 1 prefix {" ++ 
                      (if head(specs).usePrefix then " prefix_" ++ grammar_name ++ " " else "") ++ "} code @" ++ head(specs).actionCode.aCode ++ "@\n\n"  ;
                 

}

function makeStartDclString
String ::= grammar_name::String sym::[NTSpec] univLayout::String
{
  return if null(sym)
         then ""
           else "nonterm " ++ makeName(head(sym).ntName) ++ "\n" ++
                "start " ++ makeName(head(sym).ntName) ++ " layout { " ++ univLayout ++ " }\n" ++
         "term EmptyString :\n" ++
         --"token EmptyString class _ignore precedence static 0 dynamic 0 bridging 0 prefix {} code @@\n\n";
         "token EmptyString class { _ignore } precedence submits to { } dominates { }  prefix {} code @@\n\n";
}

function makeNonTermDclString
String ::= syms::[NTSpec]
{
    return if null(syms)
           then ""
           else "nonterm " ++ makeName(head(syms).ntName) ++ "\n" ++
                makeNonTermDclString(tail(syms));
}

function makePrecSpecCopperSpec
String ::= specs::[TerminalSpec]
{
  local attribute precedenceClass :: String;
  precedenceClass = "main";

  return if null(specs) 
	 then ""
	 else (if head(specs).isOperator
               then ("operator " ++
              makeName(head(specs).terminalName) ++
              " class " ++
              precedenceClass ++
              " precedence " ++
              toString(head(specs).parserPrecedence) ++
              " associativity " ++
              head(specs).parserAssociation ++
              "\n")
               else "") ++
              makePrecSpecCopperSpec(tail(specs));
}

-- Creates the complete production declaration list by concatenating the
-- above with a declaration of a new whitespace token representing all
-- the layout declared in a grammar.
function makeProdDclString
String ::= spec::Decorated RootSpec terminals::[TerminalSpec] rules::[RuleSpec] univLayout::String
{
  return makeProdSpecString(univLayout'',rules,spec.grammarsInfo);
}

function makeProdSpecString
String ::= univLayout::String rules::[RuleSpec] grammars::[GrammarInfo]
{
  return if null(rules)
         then ""
         else makeProdSpecsNonterm(univLayout,makeName(head(rules).ruleLHS),head(rules).ruleRHSSpec,grammars) ++
              makeProdSpecString(univLayout'',tail(rules),grammars);
}

function filterGrammarInfo
[GrammarInfo] ::= gName::String lft::[GrammarInfo] rght::[GrammarInfo]
{
   return if null(rght)
          then lft +++ rght
          else if head(rght).grammarNameSyn.name == gName && head(rght).hasCustomLayout
               then [head(rght)] +++ lft +++ tail(rght)
               else filterGrammarInfo(gName,lft +++ [head(rght)],tail(rght));
               
}

function makeProdSpecsNonterm
String ::= univLayout::String lhs::String rhs::[RHSSpec] grammars::[GrammarInfo]
{
  local attribute precedenceClass :: String;
  precedenceClass = "main";
  local attribute prodActionCode :: String;
  prodActionCode = "\n     ";

  local attribute reorganizedGrammars :: [GrammarInfo];

  reorganizedGrammars = if null(rhs)
                        then grammars
                        else filterGrammarInfo(head(rhs).grammarNameSyn.name,[],grammars);

  local attribute grammarSpecificLayout :: String;

  grammarSpecificLayout = if head(reorganizedGrammars).grammarNameSyn.name == head(rhs).grammarNameSyn.name && head(reorganizedGrammars).hasCustomLayout
                          then printPrecList(head(reorganizedGrammars).customLayout)
                          else univLayout'';

  return if null(rhs)
         then ""
         else "# " ++ head(rhs).grammarNameSyn.name ++ " --- " ++ head(reorganizedGrammars).grammarNameSyn.name ++ "\n" ++
              "prod " ++
              "Production_" ++ makeName(head(rhs).ruleName) ++
              " class " ++
              "main" ++
              " precedence " ++
              toString(head(rhs).parserPrecedence) ++
              " layout { " ++
              (if head(rhs).hasCustomLayout
               then printPrecList(head(rhs).customLayout) ++ " EmptyString"
               else grammarSpecificLayout ++ " EmptyString") ++
              " } code " ++
              "@ " ++
              prodActionCode ++ head(rhs).actionCode.aCode ++
              " @" ++
              " bnf " ++
              lhs ++
              " -> " ++
              (if null(head(rhs).ruleRHS)
                  then "eps"
                  else makeProdRHS(head(rhs).ruleRHS)) ++
              "\n" ++
              makeProdSpecsNonterm(univLayout'',lhs,tail(rhs),reorganizedGrammars);
              
}

function makeProdRHS
String ::= syms::[String]
{
  return if null(syms)
         then ""
         else makeName(head(syms)) ++ (if null(tail(syms)) then "" else " ") ++ makeProdRHS(tail(syms));
}

function generateUniversalLayout
String ::= grammar_name::String terminals::[TerminalSpec]
{
  local attribute layouts :: [TerminalSpec];
  layouts = filterIgnores(terminals);

 return generateLayoutList(layouts);
}

function generateLayoutList
String ::= layouts::[TerminalSpec]
{
  return if null(layouts)
         then ""
         else makeName(head(layouts).terminalName) ++ " " ++
              generateLayoutList(tail(layouts));
}

function generateLayoutRegex
String ::= layouts::[TerminalSpec]
{
  return if null(layouts)
         then ""
         else if null(tail(layouts)) then
              "[:" ++ makeName(head(layouts).terminalName) ++ ":]"
              else "[:" ++ makeName(head(layouts).terminalName) ++ ":]|" ++ generateLayoutRegex(tail(layouts));
}


function findMinimumIgnore
TerminalSpec ::= ignores::[TerminalSpec]
{
  local attribute tailMin :: TerminalSpec;
  
  tailMin = if null(ignores)
            then terminalSpec("",false,"")
            else if null(tail(ignores))
                 then head(ignores)
                 else findMinimumIgnore(tail(ignores));

  return if null(ignores)
         then tailMin''
         else if head(ignores).staticPrec < tailMin.staticPrec ||
                 (head(ignores).staticPrec == tailMin.staticPrec &&
                 (head(ignores).dynamicPrec < tailMin.dynamicPrec ||
                 (head(ignores).dynamicPrec == tailMin.dynamicPrec &&
                 head(ignores).lexerClass > tailMin.lexerClass)
                 ))
              then head(ignores)
              else tailMin'';
}

function filterIgnores
[TerminalSpec] ::= terminals::[TerminalSpec]
{
  return if null(terminals)
         then []
         else if head(terminals).ignoreTerminal
              then cons(head(terminals),filterIgnores(tail(terminals)))
              else filterIgnores(tail(terminals));
}

function filterInStart
[NTSpec] ::= s::String c2::[NTSpec]
{
  return if null(c2)
         then []
         else if head(c2).ntName == s
              then [head(c2)]
              else filterInStart(s,tail(c2));
}

function filterOutStart
[NTSpec] ::= s::String c2::[NTSpec]
{
  return if   null(c2)
         then []
         else if head(c2).ntName == s
	      then filterOutStart(s,tail(c2))
	      else cons(head(c2),filterOutStart(s,tail(c2)));
}

function filterOutProdSpecs
[ParserPrecSpec] ::= precs::[ParserPrecSpec]
{
  return precs;
}
